/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.datasource.rdb.lib;

import org.objectweb.jorm.runtime.mapper.rdb.adapter.api.RdbAdapter;
import org.objectweb.jorm.runtime.mapper.rdb.adapter.lib.BasicRdbAdapter;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.medor.api.EvaluationException;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TupleStructure;
import org.objectweb.medor.datasource.api.Wrapper;
import org.objectweb.medor.filter.api.CalculatedParameterOperand;
import org.objectweb.medor.filter.api.Expression;
import org.objectweb.medor.filter.api.Operator;
import org.objectweb.medor.filter.api.ParameterOperand;
import org.objectweb.medor.filter.api.VariableOperand;
import org.objectweb.medor.filter.lib.BasicVariableOperand;
import org.objectweb.medor.lib.Log;
import org.objectweb.medor.query.api.FilteredQueryTree;
import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.query.rdb.api.RdbExpQueryLeaf;
import org.objectweb.medor.query.rdb.api.RdbQueryLeaf;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.tuple.api.TupleCollection;
import org.objectweb.medor.tuple.lib.EmptyTupleCollection;
import org.objectweb.medor.tuple.lib.MemoryTuple;
import org.objectweb.medor.type.lib.QType;
import org.objectweb.medor.type.lib.QTypeTuple;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.io.IOException;
import java.io.Serializable;
import java.math.BigDecimal;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Date;

/**
 * This class implements the Wrapper interface for the jdbc resources. This
 * implementation uses Statements to perform the request.
 *
 * @author Mourad Alia
 */
public class JDBCWrapper implements Wrapper {

	/**
	 * this field referencs a Logger which is never null.
	 */
	Logger log = null;
	boolean debug;

	public JDBCWrapper() {
		log = Log.loggerFactory.getLogger(
				Log.MEDOR_PREFIX + ".eval.rdb.JDBCWrapper");
	}

	public TupleCollection fetchData(QueryLeaf lf,
									 ParameterOperand[] parameters,
									 Object conn) throws EvaluationException {
		debug = log.isLoggable(BasicLevel.DEBUG);
		RdbQueryLeaf leaf = (RdbQueryLeaf) lf;
		ArrayList usedParameters = null;
		ResultSet resultSet = null;
		ResultSetMetaData rsmd = null;
		Tuple buffer;
		VariableOperand[] operandBuffer;
		Statement st = null;
		PreparedStatement ps = null;
		if (parameters != null
				&& parameters.length > 0
				&& leaf instanceof FilteredQueryTree) {
			Expression filter = ((FilteredQueryTree) leaf).getQueryFilter();
			if (filter != null)
				usedParameters = getUsedParameters(filter, parameters, null);
		}
		RdbAdapter rdbAdapt = null;
		if (leaf instanceof RdbExpQueryLeaf) {
			rdbAdapt = ((RdbExpQueryLeaf) leaf).getRdbAdapter();
		}
		if (rdbAdapt == null) {
			rdbAdapt = new BasicRdbAdapter();
		}
		try {
			if (debug) log.log(BasicLevel.DEBUG, "Evaluate the query: " + leaf.getSqlRequest(null));
			if (usedParameters == null || usedParameters.size() == 0) {
				if (debug) log.log(BasicLevel.DEBUG, "No parameter => use of a Statement");
				/*
				st = ((Connection) conn).createStatement(
					ResultSet.TYPE_SCROLL_SENSITIVE,
					ResultSet.CONCUR_READ_ONLY);
				resultSet = st.executeQuery(leaf.getSqlRequest(null));
				*/
				ps = ((Connection) conn).prepareStatement(
						leaf.getSqlRequest(null),
						ResultSet.TYPE_SCROLL_SENSITIVE,
						ResultSet.CONCUR_READ_ONLY);
				resultSet = ps.executeQuery();
			}
			else {
				if (debug) log.log(BasicLevel.DEBUG, usedParameters.size() + " parameter(s) => use of a PreparedStatement");
				ps = ((Connection) conn).prepareStatement(
						leaf.getSqlRequest((ParameterOperand[])
						usedParameters.toArray(new ParameterOperand[0])),
						ResultSet.TYPE_SCROLL_SENSITIVE,
						ResultSet.CONCUR_READ_ONLY);
				for (int i = 0; i < usedParameters.size(); i++) {
					ParameterOperand po =
							(ParameterOperand) usedParameters.get(i);
					if (debug) log.log(BasicLevel.DEBUG, "Assign parameter " + po.getName());
					switch (po.getType().getTypeCode()) {
					case QTypeTuple.TYPECODE_INT:
						rdbAdapt.setInt(ps, i + 1, po.getInt());
						break;
					case QTypeTuple.TYPECODE_OBJINT:
						rdbAdapt.setOint(ps, i + 1, (Integer) po.getObject());
						break;
					case QTypeTuple.TYPECODE_SHORT:
						rdbAdapt.setShort(ps, i + 1, po.getShort());
						break;
					case QTypeTuple.TYPECODE_OBJSHORT:
						rdbAdapt.setOshort(ps, i + 1, (Short) po.getObject());
						break;
					case QTypeTuple.TYPECODE_BYTE:
						rdbAdapt.setByte(ps, i + 1, po.getByte());
						break;
					case QTypeTuple.TYPECODE_OBJBYTE:
						rdbAdapt.setObyte(ps, i + 1, (Byte) po.getObject());
						break;
					case QTypeTuple.TYPECODE_LONG:
						rdbAdapt.setLong(ps, i + 1, po.getLong());
						break;
					case QTypeTuple.TYPECODE_OBJLONG:
						rdbAdapt.setOlong(ps, i + 1, (Long) po.getObject());
						break;
					case QTypeTuple.TYPECODE_DOUBLE:
						rdbAdapt.setDouble(ps, i + 1, po.getDouble());
						break;
					case QTypeTuple.TYPECODE_OBJDOUBLE:
						rdbAdapt.setOdouble(ps, i + 1, (Double) po.getObject());
						break;
					case QTypeTuple.TYPECODE_BOOLEAN:
						rdbAdapt.setBoolean(ps, i + 1, po.getBoolean());
						break;
					case QTypeTuple.TYPECODE_OBJBOOLEAN:
						rdbAdapt.setOboolean(ps, i + 1, (Boolean) po.getObject());
						break;
					case QTypeTuple.TYPECODE_FLOAT:
						rdbAdapt.setFloat(ps, i + 1, po.getFloat());
						break;
					case QTypeTuple.TYPECODE_OBJFLOAT:
						rdbAdapt.setOfloat(ps, i + 1, (Float) po.getObject());
						break;
					case QTypeTuple.TYPECODE_BIGDECIMAL:
						rdbAdapt.setBigDecimal(ps, i + 1, (BigDecimal) po.getObject());
						break;
					case QTypeTuple.TYPECODE_STRING:
						rdbAdapt.setString(ps, i + 1, (String) po.getObject());
						break;
					case QTypeTuple.TYPECODE_CHAR:
						rdbAdapt.setChar(ps, i + 1, po.getChar());
						break;
					case QTypeTuple.TYPECODE_OBJCHAR:
						rdbAdapt.setOchar(ps, i + 1, (Character) po.getObject());
						break;
					case QTypeTuple.TYPECODE_SERIALIZED:
						rdbAdapt.setSerialized(ps, i + 1, (Serializable) po.getObject());
						break;
					case QTypeTuple.TYPECODE_DATE:
						rdbAdapt.setDate(ps, i + 1, (Date) po.getObject(), null);
						break;
					case QTypeTuple.TYPECODE_BYTEARRAY:
						rdbAdapt.setByteArray(ps, i + 1, (byte[]) po.getObject());
						break;
					case QTypeTuple.TYPECODE_CHARARRAY:
						rdbAdapt.setCharArray(ps, i + 1, (char[]) po.getObject());
						break;
					case QTypeTuple.TYPECODE_COLLECTION:
						//Already assign in the sql request building
						break;
					default:
						throw new EvaluationException("The type "
								+ po.getType().getJormName()
								+ " is not supported '"
								+ po.getType().getTypeCode() + "'"
						);
					}
				}
				if (debug) log.log(BasicLevel.DEBUG, "Evaluate the query");
				resultSet = ps.executeQuery();
			}
			if (!resultSet.next()) {
				if (debug) log.log(BasicLevel.DEBUG, "No result");
				if (st != null)
					st.close();
				if (ps != null)
					ps.close();
				return new EmptyTupleCollection(lf.getTupleStructure());
			}
			else {
				if (debug) log.log(BasicLevel.DEBUG, "There is a result");
				rsmd = resultSet.getMetaData();
				operandBuffer = new VariableOperand[rsmd.getColumnCount()];
				for (int cpt = 0;
					 (cpt < lf.getTupleStructure().getSize());
					 cpt++) {
					// create and init the buffer
					operandBuffer[cpt] =
							new BasicVariableOperand(
									leaf.getTupleStructure().getField(cpt + 1).getType());
				}
				buffer = new MemoryTuple(operandBuffer);
				return new JDBCTupleCollection(leaf, resultSet, operandBuffer,
						buffer, rdbAdapt);
			}
		}
		catch (SQLException sqlexp) {
			throw new EvaluationException(sqlexp);
		}
		catch (EvaluationException ee) {
			throw ee;
		}
		catch (MedorException mexp) {
			throw new EvaluationException(mexp);
		}
		catch (IOException exp) {
			throw new EvaluationException(exp);
		}
	}

	private class JDBCTupleCollection implements TupleCollection {

		private ResultSet resultSet;
		private RdbQueryLeaf leaf;
		private VariableOperand[] operandBuffer;
		private Tuple buffer;
		private RdbAdapter adapter;

		public JDBCTupleCollection(RdbQueryLeaf lf, ResultSet rs,
								   VariableOperand[] ob,
								   Tuple b,
								   RdbAdapter adapter)
				throws MedorException, SQLException {
			if (debug) log.log(BasicLevel.DEBUG, "TupleCollection creation");
			leaf = lf;
			resultSet = rs;
			operandBuffer = ob;
			buffer = b;
			this.adapter = adapter;
			loadTuple();
		}

		public TupleStructure getMetaData() throws MedorException {
			return leaf.getTupleStructure();
		}

		public boolean isLast() throws MedorException {
			if (debug) log.log(BasicLevel.DEBUG, "");
			try {
				return resultSet.isLast();
			}
			catch (SQLException sqlex) {
				throw new MedorException(sqlex);
			}
		}

		/**
		 * Moves the cursor down one row from its
		 * current position to the next tuple
		 */
		public boolean next() throws MedorException {
			if (debug) log.log(BasicLevel.DEBUG, "");
			try {
				if (resultSet.next()) {
					if (debug) log.log(BasicLevel.DEBUG, "resultset.next()=true");
					loadTuple();
					return true;
				}
				else {
					if (debug) log.log(BasicLevel.DEBUG, "resultset.next()=false");
					return false;
				}
			}
			catch (SQLException sqlex) {
				throw new MedorException(sqlex);
			}
		}

		// Moves the cursor to the previous Tuple of this TupleCollection
		public boolean previous() throws MedorException {
			if (debug) log.log(BasicLevel.DEBUG, "");
			try {
				if (resultSet.previous()) {
					if (debug) log.log(BasicLevel.DEBUG, "resultset.previous()=true");
					loadTuple();
					return true;
				}
				else {
					if (debug) log.log(BasicLevel.DEBUG, "resultset.previous()=false");
					return false;
				}

			}
			catch (SQLException sqlex) {
				throw new MedorException(sqlex);
			}
		}

		/**
		 * Moves the cursor to the first Tuple of this TupleCollection
		 */
		public void first() throws MedorException {
			if (debug) log.log(BasicLevel.DEBUG, "");
			try {
				if (resultSet.first()) {
					if (debug) log.log(BasicLevel.DEBUG, "resultset.previous()=true");
					loadTuple();
				}
				else
					throw new MedorException("can't move the cursor to the first row");
			}
			catch (SQLException ex) {
				throw new MedorException(ex);
			}
		}

		/**
		 * Retrieves the current row Number of this TupleCollection
		 */
		public int getRow() throws MedorException {
			if (debug) log.log(BasicLevel.DEBUG, "");
			try {
				return resultSet.getRow();
			}
			catch (SQLException ex) {
				throw new MedorException(ex);
			}
		}

		public Object getObject(int numColumn) throws MedorException {
			if (debug) log.log(BasicLevel.DEBUG, "numColumn=" + numColumn);
			try {
				Object o = resultSet.getObject(numColumn);
				return (resultSet.wasNull() ? null : o);
			}
			catch (SQLException sqlex) {
				throw new MedorException(sqlex);
			}
			catch (Exception e) {
				throw new MedorException("Invalid field index " + numColumn, e);
			}
		}

		public Tuple getTuple() throws MedorException {
			if (debug) log.log(BasicLevel.DEBUG, "");
			return buffer;
		}

		private void loadTuple() throws MedorException, SQLException {
			if (debug) log.log(BasicLevel.DEBUG, "");
			PType ptype = null;
			for (int cpt = 1;
				 (cpt <= leaf.getTupleStructure().getSize());
				 cpt++) {
				ptype = leaf.getTupleStructure().getField(cpt).getType();
				switch (ptype.getTypeCode()) {
				case QType.TYPECODE_INT:
					operandBuffer[cpt - 1].setValue(adapter.getInt(resultSet, cpt));
					operandBuffer[cpt - 1].setIsDefined(!resultSet.wasNull());
					break;
				case QType.TYPECODE_SHORT:
					operandBuffer[cpt - 1].setValue(adapter.getShort(resultSet, cpt));
					operandBuffer[cpt - 1].setIsDefined(!resultSet.wasNull());
					break;
				case QType.TYPECODE_BYTE:
					operandBuffer[cpt - 1].setValue(adapter.getByte(resultSet, cpt));
					operandBuffer[cpt - 1].setIsDefined(!resultSet.wasNull());
					break;
				case QType.TYPECODE_LONG:
					operandBuffer[cpt - 1].setValue(adapter.getLong(resultSet, cpt));
					operandBuffer[cpt - 1].setIsDefined(!resultSet.wasNull());
					break;
				case QType.TYPECODE_DOUBLE:
					operandBuffer[cpt - 1].setValue(adapter.getDouble(resultSet, cpt));
					operandBuffer[cpt - 1].setIsDefined(!resultSet.wasNull());
					break;
				case QType.TYPECODE_BOOLEAN:
					operandBuffer[cpt - 1].setValue(adapter.getBoolean(resultSet, cpt));
					operandBuffer[cpt - 1].setIsDefined(!resultSet.wasNull());
					break;
				case QType.TYPECODE_FLOAT:
					operandBuffer[cpt - 1].setValue(adapter.getFloat(resultSet, cpt));
					operandBuffer[cpt - 1].setIsDefined(!resultSet.wasNull());
					break;
				case QType.TYPECODE_STRING:
					operandBuffer[cpt - 1].setValue(adapter.getString(resultSet, cpt));
					break;
				case QType.TYPECODE_OBJBOOLEAN:
					operandBuffer[cpt - 1].setValue(adapter.getOboolean(resultSet, cpt));
					break;
				case QType.TYPECODE_OBJCHAR:
					operandBuffer[cpt - 1].setValue(adapter.getOchar(resultSet, cpt));
					break;
				case QType.TYPECODE_OBJBYTE:
					operandBuffer[cpt - 1].setValue(adapter.getObyte(resultSet, cpt));
					break;
				case QType.TYPECODE_OBJSHORT:
					operandBuffer[cpt - 1].setValue(adapter.getOshort(resultSet, cpt));
					break;
				case QType.TYPECODE_OBJINT:
					operandBuffer[cpt - 1].setValue(adapter.getOint(resultSet, cpt));
					break;
				case QType.TYPECODE_OBJLONG:
					operandBuffer[cpt - 1].setValue(adapter.getOlong(resultSet, cpt));
					break;
				case QType.TYPECODE_OBJFLOAT:
					operandBuffer[cpt - 1].setValue(adapter.getOfloat(resultSet, cpt));
					break;
				case QType.TYPECODE_OBJDOUBLE:
					operandBuffer[cpt - 1].setValue(adapter.getOdouble(resultSet, cpt));
					break;
				case QType.TYPECODE_DATE:
					operandBuffer[cpt - 1].setValue(adapter.getDate(resultSet, cpt));
					break;
				case QType.TYPECODE_CHAR:
					operandBuffer[cpt - 1].setValue(adapter.getChar(resultSet, cpt));
					operandBuffer[cpt - 1].setIsDefined(!resultSet.wasNull());
					break;
				case QType.TYPECODE_BYTEARRAY:
					operandBuffer[cpt - 1].setValue(adapter.getByteArray(resultSet, cpt));
					operandBuffer[cpt - 1].setIsDefined(!resultSet.wasNull());
					break;
				case QType.TYPECODE_CHARARRAY:
					operandBuffer[cpt - 1].setValue(adapter.getCharArray(resultSet, cpt));
					operandBuffer[cpt - 1].setIsDefined(!resultSet.wasNull());
					break;
				case PType.TYPECODE_SERIALIZED:
					try {
						operandBuffer[cpt - 1].setValue(adapter.getSerialized(resultSet, cpt));
					}
					catch (Exception e) {
						throw new MedorException("Impossible to fetch a serialized field a the index " + cpt, e);
					}
					break;
				}
			}
		}

		public Tuple getTuple(int row) throws MedorException {
			if (debug) log.log(BasicLevel.DEBUG, "");
			try {
				if (debug) log.log(BasicLevel.DEBUG, "resultSet.absolute(" + row + ")");
				if (resultSet.absolute(row)) {
					loadTuple();
				}
				else
					throw new MedorException("Can't move to the row: " + row);
			}
			catch (SQLException sqlexp) {
				throw new MedorException(sqlexp);
			}
			return buffer;
		}

		public boolean isEmpty() throws MedorException {
			if (debug) log.log(BasicLevel.DEBUG, "return " + (buffer == null));
			return (buffer == null);
		}

		public boolean row(int i) throws MedorException {
			if (debug) log.log(BasicLevel.DEBUG, "");
			try {
				if (debug) log.log(BasicLevel.DEBUG, "resultSet.absolute(" + i + ")");
				if (resultSet.absolute(i)) {
					loadTuple();
					return true;
				}
				else
					return false;
			}
			catch (SQLException sqlex) {
				throw new MedorException(sqlex);
			}
		}
	}

	private ArrayList getUsedParameters(Expression e,
										ParameterOperand[] parameters,
										ArrayList al)
			throws EvaluationException {
		//TODO take memberof into account for parameters
		if (e instanceof Operator) {
			for (int i = 0; i < ((Operator) e).getOperandNumber(); i++) {
				al = getUsedParameters(((Operator) e).getExpression(i),
						parameters, al);
			}
		}
		else if (e instanceof CalculatedParameterOperand) {
			CalculatedParameterOperand cpo = (CalculatedParameterOperand) e;
			try {
				cpo.evaluate(parameters);
			}
			catch (MedorException e1) {
				throw new EvaluationException(e1);
			}
			if (al == null)
				al = new ArrayList();
			al.add(cpo);
		}
		else if (e instanceof ParameterOperand) {
			int i = 0;
			ParameterOperand param = (ParameterOperand) e;
			while (i < parameters.length
					&& !parameters[i].getName().equals(param.getName()))
				i++;
			if (i < parameters.length) {
				if (parameters[i].getType().getTypeCode() !=
						param.getType().getTypeCode()) {
					throw new EvaluationException("Bad parameter type: found "
							+ parameters[i].getType().getJormName()
							+ " / expected " + param.getType().getJormName());
				}
				if (al == null)
					al = new ArrayList();
				al.add(parameters[i]);
			}
			else {
				String msg = "You must specify the value of the parameter "
						+ param.getName() + ". Found: {";
				String sep = "";
				for (int j = 0; j < parameters.length; j++) {
					msg += sep + parameters[j].getName();
					sep = ", ";
				}
				msg += "}";
				throw new EvaluationException(msg);
			}
		}
		return al;

	}
}
