/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.eval.lib;

import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.QueryNodeException;
import org.objectweb.medor.eval.api.NodeEvaluator;
import org.objectweb.medor.eval.cache.lib.TupleCache;
import org.objectweb.medor.eval.cache.lib.WindowCachedTupleCollection;
import org.objectweb.medor.eval.prefetch.api.PrefetchBuffer;
import org.objectweb.medor.filter.api.ParameterOperand;
import org.objectweb.medor.query.api.OperationType;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.tuple.api.TupleCollection;

import java.util.List;

public class BinaryIteratifNodeEvaluator implements NodeEvaluator {

	QueryNode query;
	NodeEvaluator leftNodeEvaluator, rightNodeEvaluator;
	long cacheSize;
	boolean open = false;

	public BinaryIteratifNodeEvaluator(QueryNode query,
									   NodeEvaluator leftNodeEvaluator,
									   NodeEvaluator rightNodeEvaluator) {
		this.query = query;
		this.leftNodeEvaluator = leftNodeEvaluator;
		this.rightNodeEvaluator = rightNodeEvaluator;
	}

	public boolean linkChildren() {
		return true;
	}

	public boolean unlinkChildren() {
		return true;
	}

	public TupleCollection fetchData(ParameterOperand[] parameters)
		throws MedorException {
		switch (query.getType()) {
		case OperationType.CARTESIAN:
			if ((query.getOptimizationMetaData() == null)) // There is no optimization
				return
					new CartesianEvaluatedTC(query, leftNodeEvaluator,
						rightNodeEvaluator, parameters);
			else if (query.getOptimizationMetaData().isCache())
				return new WindowCachedTupleCollection(
					new CartesianEvaluatedTC(query, leftNodeEvaluator,
						rightNodeEvaluator, parameters),
					new TupleCache(cacheSize)
				);
			else
				return
					new CartesianEvaluatedTC(query, leftNodeEvaluator,
						rightNodeEvaluator, parameters);

		case OperationType.INTERSECTION:
			if ((query.getOptimizationMetaData() == null)) // There is no optimization
				return
					new BinaryImplicitTC(query, leftNodeEvaluator,
						rightNodeEvaluator, parameters);
			else if (query.getOptimizationMetaData().isCache())
				return new WindowCachedTupleCollection(
					new BinaryImplicitTC(query, leftNodeEvaluator,
						rightNodeEvaluator, parameters),
					new TupleCache(cacheSize)
				);
			else
				return
					new BinaryImplicitTC(query, leftNodeEvaluator,
						rightNodeEvaluator, parameters);

		case OperationType.JOIN_PROJECT:
			if ((query.getOptimizationMetaData() == null)) // There is no optimization
				return
					new BinaryImplicitTC(query, leftNodeEvaluator,
						rightNodeEvaluator, parameters);
			else if (query.getOptimizationMetaData().isCache())
				return
					new WindowCachedTupleCollection(
						new BinaryImplicitTC(query, leftNodeEvaluator,
							rightNodeEvaluator, parameters),
						new TupleCache(cacheSize)
					);
			else
				return new BinaryImplicitTC(query, leftNodeEvaluator,
					rightNodeEvaluator, parameters);

		case OperationType.UNION:
			if ((query.getOptimizationMetaData() == null)) // There is no optimization
				return new UnionEvaluatedTC(query, leftNodeEvaluator,
					rightNodeEvaluator, parameters);
			else if (query.getOptimizationMetaData().isCache())
				return new WindowCachedTupleCollection(
					new UnionEvaluatedTC(query, leftNodeEvaluator,
						rightNodeEvaluator, parameters),
					new TupleCache(cacheSize)
				);
			else
				return
					new UnionEvaluatedTC(query, leftNodeEvaluator,
						rightNodeEvaluator, parameters);

		default:
			throw new QueryNodeException("Unreconized QueryNode Type");
		}
	}

	public long getCacheSize() {
		return cacheSize;
	}

	public void setCacheSize(long cacheSize) {
		this.cacheSize = cacheSize;
	}

	public QueryNode getQueryNode() {
		return query;
	}

    public void setPrefetchBuffer(PrefetchBuffer pb) throws MedorException {
        // No prefetching for this evaluator: ignore
    }
}
