/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.eval.lib;

import org.objectweb.medor.api.EvaluationException;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.datasource.api.Wrapper;
import org.objectweb.medor.datasource.api.WrapperFactory;
import org.objectweb.medor.datasource.lib.MedorWrapperFactory;
import org.objectweb.medor.eval.api.ConnectionResources;
import org.objectweb.medor.eval.api.NodeEvaluator;
import org.objectweb.medor.eval.api.NodeEvaluatorFactory;
import org.objectweb.medor.eval.api.QueryEvaluator;
import org.objectweb.medor.eval.prefetch.api.PrefetchCache;
import org.objectweb.medor.eval.prefetch.api.PrefetchBuffer;
import org.objectweb.medor.filter.api.ParameterOperand;
import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.rdb.api.RdbQueryLeaf;
import org.objectweb.medor.tuple.api.TupleCollection;

import java.util.HashMap;
import java.util.Stack;
import java.util.ArrayList;

public class MedorEvaluator implements QueryEvaluator {

	public MedorEvaluator(QueryTree query, long cacheSize)
		throws EvaluationException {
		this.query = query;
		this.cacheSize = cacheSize; // The all cahe size
        this.prefetchBuffer = null;
		toStack();
	}

	QueryTree query;
	Stack queryTreeStack = new Stack();
	HashMap nodeEvaluators = new HashMap();
	WrapperFactory wrapperFactory = new MedorWrapperFactory();
	NodeEvaluatorFactory nodeEvaluatorFactory =
		new IteratifNodeEvaluatorFactory();
	long cacheSize;
    PrefetchBuffer prefetchBuffer;

	public WrapperFactory getWrapperFactory() {
		return wrapperFactory;
	}

	/**
	 * The query to evaluate must be evaluable: compiled,...
	 */
	public TupleCollection evaluate(ParameterOperand[] parameters,
									ConnectionResources ressources)
		throws EvaluationException {
		try {
			Stack runTimeStack = new Stack();
			int cpt = queryTreeStack.size() - 1;
			while (cpt >= 0) {
				if (queryTreeStack.elementAt(cpt) instanceof QueryLeaf) {
					QueryLeaf ql = (QueryLeaf) queryTreeStack.elementAt(cpt);
					//System.out.println("Generating wrapppers ...");
					Wrapper w = wrapperFactory.getWrapper(ql.getDataStore());
					if (ressources != null) {
						runTimeStack.push(
							new TupleCollectionNodeEvaluator(
								w.fetchData(ql,
									parameters,
									ressources.getConnection(ql))));
					}
					else {
						// allowed passing null connection
						runTimeStack.push(
							new TupleCollectionNodeEvaluator(
								w.fetchData(ql,
									parameters,
									null)));
					}
				}
				else {
					//System.out.println("Generating NodeEvaluators...");
					QueryNode qn = (QueryNode) queryTreeStack.elementAt(cpt);
					QueryTree[] children = qn.getChildren();
					if (children.length == 2) {
						NodeEvaluator right =
							(NodeEvaluator) runTimeStack.pop();
						right.setCacheSize(cacheSize);
						NodeEvaluator left = (NodeEvaluator) runTimeStack.pop();
						left.setCacheSize(cacheSize);
						runTimeStack.push(
							nodeEvaluatorFactory.createNodeEvaluator(
								qn,
								new NodeEvaluator[]{left, right}));
					}
					else {
						NodeEvaluator subEval =
							(NodeEvaluator) runTimeStack.pop();
						runTimeStack.push(
							nodeEvaluatorFactory.createNodeEvaluator(
								qn,
								new NodeEvaluator[]{subEval}));
					}
				}
				cpt--;
			}

			NodeEvaluator queryTreeHead = ((NodeEvaluator) runTimeStack.pop());
			queryTreeHead.setCacheSize(cacheSize);
            queryTreeHead.setPrefetchBuffer(prefetchBuffer);
			return queryTreeHead.fetchData(parameters);
		}
		catch (MedorException mexp) {
			throw new EvaluationException(mexp);
		}
	}

    /**
     * Requests the evaluator to index the resulting tuple collection by passing
     * along the structure for indexing the prefetched tuples.
     * @param pb   The PrefetchBuffer to be filled by the TupleCollection.
     */
    public void setPrefetchBuffer(PrefetchBuffer pb) throws MedorException {
        prefetchBuffer = pb;
    }

	public QueryTree getQueryTree() {
		return query;
	}

	private void toStack() throws EvaluationException {
		toStack(query, queryTreeStack);
	}

	private void toStack(QueryTree query, Stack pile)
		throws EvaluationException {
		pile.push(query);
		if (query instanceof QueryLeaf) {
		}
		else {
			QueryTree[] children = ((QueryNode) query).getChildren();
			if (children.length == 2) {
				toStack(children[1], pile);
				toStack(children[0], pile);
			}
			else if (children.length == 1) {
				toStack(children[0], pile);
			}
			else
				throw new EvaluationException("Wrong children number (" +
					+children.length
					+ ") impossible to evaluate this QueryTree Type "
					+ ((QueryNode) query).getType());
		}
	}

	public long getCacheSize() {
		return cacheSize;
	}

	public ConnectionResources getRequiredConnectionResources() {
		HashMap qls = new HashMap();
		for (int cpt = 0; (cpt < queryTreeStack.size()); cpt++) {
			if (queryTreeStack.elementAt(cpt) instanceof RdbQueryLeaf) {
				// Requires a JDBC connection
				qls.put(queryTreeStack.elementAt(cpt), null);
				// There is no an Object Connection
			}
		}
		return new MedorConnectionResources(qls);
	}
}
