/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.eval.lib;

import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TupleStructure;
import org.objectweb.medor.api.Field;
import org.objectweb.medor.eval.api.BinaryEvaluatedTC;
import org.objectweb.medor.eval.api.NodeEvaluator;
import org.objectweb.medor.eval.prefetch.api.PrefetchBuffer;
import org.objectweb.medor.filter.api.Operand;
import org.objectweb.medor.filter.api.ParameterOperand;
import org.objectweb.medor.filter.api.VariableOperand;
import org.objectweb.medor.filter.api.Expression;
import org.objectweb.medor.filter.lib.BasicOperand;
import org.objectweb.medor.filter.lib.BasicVariableOperand;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.CalculatedField;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.tuple.api.TupleCollection;
import org.objectweb.medor.tuple.lib.MemoryTuple;
import org.objectweb.jorm.type.api.PType;

import java.util.List;


/**
 * A project Operation QueryNode is performed by this TupleCollection. This
 * operation is not costs and doeas not uses buffers.
 */
public class ProjectEvaluatedTC implements BinaryEvaluatedTC {

    private QueryNode query;
    private TupleCollection subResultTC;
    private ParameterOperand[] parameters;
    private Tuple currentBuffer;
    private VariableOperand[] currentAttBuffers;
    private int cursor,size;
    private boolean empty,isTheLast;
    private PrefetchBuffer prefetchBuffer;

    public ProjectEvaluatedTC(QueryNode query,
                              NodeEvaluator subNodeEvaluator,
                              ParameterOperand[] parameters,
                              PrefetchBuffer pb) throws MedorException {
        this.parameters = parameters;
        this.query = query;
        this.prefetchBuffer = pb;
        if (prefetchBuffer != null) {
            pb.setTupleCollection(this);
        }

        // the size of the result of the projection operation
        size = query.getTupleStructure().getSize();

        currentAttBuffers = new BasicVariableOperand[size];
        PType type;
        for (int cpt = 0; (cpt < query.getTupleStructure().getSize()); cpt++) {
            currentAttBuffers[cpt] = new BasicVariableOperand(
                    query.getTupleStructure().getField(cpt + 1).getType());
        }
        currentBuffer = new MemoryTuple(currentAttBuffers);// This object is used as a buffer
        subResultTC = subNodeEvaluator.fetchData(parameters);
        isTheLast = false;
        empty = (subResultTC.isEmpty());
        if (empty)
            cursor = 0;
        else {
            init();
        }
    }

    public TupleStructure getMetaData() throws MedorException {
        return query.getTupleStructure();
    }

    public boolean isLast() throws MedorException {
        return subResultTC.isLast();
    }

    public int card() throws MedorException {
        return cursor;
    }

    public boolean next() throws MedorException {
        boolean moved;
        if (subResultTC.isLast()) {
            // we have already iterate the last tuple
            cursor = -2;
            moved = false;
        } else {
            subResultTC.next();
            //Upgrade the buffer
            getNextResult();
            if (prefetchBuffer != null) {
                prefetchBuffer.addPrefetchTuple();
            }
            moved = true;
            cursor++;
        }
        return moved;
    }

    private void init() throws MedorException {
        subResultTC.first();
        getNextResult();
        cursor = 1;
        if (prefetchBuffer != null) {
            prefetchBuffer.addPrefetchTuple();
        }
    }

    public void first() throws MedorException {
        //The Initialisation
        init();
    }

    public int getRow() throws MedorException {
        return cursor;
    }

    public Object getObject(int numColumn) throws MedorException {
        if (!isEmpty() && getRow() >= 1)
            return currentBuffer.getObject(numColumn);
        else
            throw new MedorException(" No elements fetched in this TupleCollection");
    }

    public Tuple getTuple() throws MedorException {
        if (!isEmpty() && (getRow() >= 1))
            return currentBuffer;
        else
            throw new MedorException(" No elements fetched in this TupleCollection  " + getRow());
    }

    public synchronized Tuple getTuple(int numTuple) throws MedorException {
        row(numTuple);
        return getTuple();
    }

    public boolean isEmpty() throws MedorException {
        return (subResultTC.isEmpty());
    }

    public boolean row(int numTuple) throws MedorException {
// P. Dechamboux: REMOVE CODE
//        first();
//        int cpt = 1;
//        boolean go = true;
//        while ((cpt <= numTuple) && (cpt > 1) && go) {
//            if (!next()) go = false;
//        }
        boolean go = subResultTC.row(numTuple); // P. Dechamboux: ADD CODE
        if (!go) {
            return false;
        } else {
            getNextResult();                     // P. Dechamboux: ADD CODE
            cursor = numTuple;
            return true;
        }
    }

    public int getLeftTCCursor() throws MedorException {
        return subResultTC.getRow();
    }

    public int getRightTCCursor() throws MedorException {
        return -1;
    }

    private void getNextResult() throws MedorException {
        query.getTupleLoader().loadTuple(
                subResultTC.getTuple(), currentAttBuffers, parameters);
    }
}
