/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.eval.lib;

import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TupleStructure;
import org.objectweb.medor.eval.api.BinaryEvaluatedTC;
import org.objectweb.medor.eval.api.NodeEvaluator;
import org.objectweb.medor.eval.prefetch.api.PrefetchBuffer;
import org.objectweb.medor.filter.api.Expression;
import org.objectweb.medor.filter.api.Operand;
import org.objectweb.medor.filter.api.ParameterOperand;
import org.objectweb.medor.filter.api.VariableOperand;
import org.objectweb.medor.filter.lib.BasicVariableOperand;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.tuple.api.TupleCollection;
import org.objectweb.medor.tuple.lib.MemoryTuple;
import org.objectweb.jorm.type.api.PType;

/**
 * This classe is a TupleCollection result of a Select + Project Operations.
 * It can be seen as a simple filtering of a sigle TupleCollection. It can be
 * implemented with the BinaryImplicitTC but this implementation is more
 * optimized and simplified.
 */
public class SelProjEvaluatedTC implements BinaryEvaluatedTC {

	private QueryNode query;
	private Expression filter;
	private TupleCollection subTupleCollection;
	private ParameterOperand[] parameters;
	private Tuple tupleBuffer1,tupleBuffer2;
	private VariableOperand[] operandBuffer1,operandBuffer2;
	private int cursor;
	private boolean empty, isTheLast = false, isBuffer2Free = false;
	Operand filterResult;
    PrefetchBuffer prefetchBuffer;

	public SelProjEvaluatedTC(QueryNode query,
							  NodeEvaluator subNodeEvaluator,
							  ParameterOperand[] parameters,
                              PrefetchBuffer pb)
		throws MedorException {
		this.query = query;
		this.parameters = parameters;
		this.filter = query.getQueryFilter();
		this.filterResult = filter.getResult();
        this.prefetchBuffer = pb;
        if (prefetchBuffer != null) {
            pb.setTupleCollection(this);
        }

		// Init the two buffers
		int size = query.getTupleStructure().getSize();
		operandBuffer1 = new BasicVariableOperand[size];
		operandBuffer2 = new BasicVariableOperand[size];
		for (int cpt = 0; (cpt < query.getTupleStructure().getSize()); cpt++) {
			PType type = query.getTupleStructure().getField(cpt + 1).getType();
			operandBuffer1[cpt] = new BasicVariableOperand(type);
			operandBuffer2[cpt] = new BasicVariableOperand(type);
		}
		tupleBuffer1 = new MemoryTuple(operandBuffer1); // Creating the Buffer1
		tupleBuffer2 = new MemoryTuple(operandBuffer2);
		subTupleCollection = subNodeEvaluator.fetchData(parameters);
		empty = (subTupleCollection.isEmpty());
		if (empty)
			cursor = 0;
		else
			this.init();
	}

	public TupleStructure getMetaData() throws MedorException {
		return query.getTupleStructure();
	}

	public boolean isLast() throws MedorException {
		return isTheLast;
	}

	public int card() throws MedorException {
		return cursor;
	}

	public boolean next() throws MedorException {
		Tuple currentTuple = null;
		boolean moved = false;

		if (isEmpty() || (cursor == -2))
			moved = false;
		else if (isLast()) {
			// we have iterated all tuples
			cursor = -2;
			moved = false;
		}
		else {
			// we move to the next Result
			moved = true;
			cursor++;
			do {
				//Verify if there is others results
				if (subTupleCollection.next()) {
					currentTuple = subTupleCollection.getTuple();
					filter.evaluate(parameters, currentTuple);
				}
				else {
					isTheLast = true;
				}
			} while (!(isTheLast) && !(filterResult.getBoolean()));
			if (!isTheLast) {
				if (isBuffer2Free) {
					//Load the next Result in the buffer2
					query.getTupleLoader().loadTuple(currentTuple,
						operandBuffer2, parameters);
					isBuffer2Free = false;
				}
				else {
					//Load the next Result in the buffer1
					query.getTupleLoader().loadTuple(currentTuple,
						operandBuffer1, parameters);
					isBuffer2Free = true;
				}
                if (prefetchBuffer != null) {
                    prefetchBuffer.addPrefetchTuple();
                }
			}
			else
				isBuffer2Free = !isBuffer2Free;
		}
		return moved;
	}

	/**
	 * This method Move the cursor to the first result if there is one, load it,
	 * then it verfy if it is the last one. In the case of the last one it
	 * returns false whether it returns true.
	 */
	private boolean init() throws MedorException {
		subTupleCollection.first();
		isTheLast = false;
		Tuple currentTuple;
		isBuffer2Free = false;
		currentTuple = subTupleCollection.getTuple();
		filter.evaluate(parameters, currentTuple);
		while (!empty && !filterResult.getBoolean()) {
			if (subTupleCollection.next()) {
				currentTuple = subTupleCollection.getTuple();
				filter.evaluate(parameters, currentTuple);
			}
			else {
				empty = true;//This TupleCollection result is empty
			}
		}
		if (empty) {
			cursor = -2;
		}
		else {
			// Loading the first result in the buffer1
			query.getTupleLoader().loadTuple(currentTuple,
				operandBuffer1, parameters);
			cursor = 1;
            if (prefetchBuffer != null) {
                prefetchBuffer.addPrefetchTuple();
            }
			do {
				// Verify if there is other result
				if (subTupleCollection.next()) {
					currentTuple = subTupleCollection.getTuple();
					filter.evaluate(parameters, currentTuple);
				}
				else {
					// There is no more results
					isTheLast = true;
				}
			} while (!(isTheLast) && !(filterResult.getBoolean()));
			if (!isTheLast) {
				// Loading the next Result in the buffer2
				query.getTupleLoader().loadTuple(currentTuple,
					operandBuffer2, parameters);
                if (prefetchBuffer != null) {
                    prefetchBuffer.addPrefetchTuple();
                }
			}
		}
		return (!empty);
	}

	public void first() throws MedorException {
		//The Initialisation
		init();
	}

	public int getRow() throws MedorException {
		return cursor;
	}

	public Object getObject(int numColumn) throws MedorException {
		if (!isEmpty() && getRow() >= 1) {
			if (isBuffer2Free)
				return tupleBuffer2.getObject(numColumn);
			else
				return tupleBuffer1.getObject(numColumn);
		}
		else
			throw new MedorException(" No elements fetched in this TupleCollection");
	}

	public Tuple getTuple() throws MedorException {
		if (!isEmpty() && getRow() >= 1) {
			if (isBuffer2Free)
				return tupleBuffer2;
			else
				return tupleBuffer1;
		}
		else
			throw new MedorException(" No elements fetched in this TupleCollection");
	}

	public synchronized Tuple getTuple(int numTuple) throws MedorException {
		row(numTuple);
		return getTuple();
	}

	public boolean isEmpty() throws MedorException {
		return empty;
	}

	public boolean row(int numTuple) throws MedorException {
		first();
		int cpt = 1;
		boolean stop = false;
		while ((cpt < numTuple) && !stop) {
			if (!next())
				stop = true;
			else
				cpt++;
		}
		if (cpt < numTuple)
			return false;
		else {
			cursor = numTuple;
			return true;
		}

	}

	public int getLeftTCCursor() throws MedorException {
		return subTupleCollection.getRow();
	}

	public int getRightTCCursor() throws MedorException {
		return -1;
	}
}
