/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 * 
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.eval.lib;

import org.objectweb.jorm.api.PExceptionIO;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.naming.api.PNameGetter;
import org.objectweb.jorm.naming.api.PNameCoder;
import org.objectweb.jorm.naming.api.PExceptionNaming;
import org.objectweb.jorm.type.api.PExceptionTyping;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.filter.api.Operand;

import java.util.Date;
import java.util.Map;

/**
 * This class implements a JORM PnameGetter. It relies on a Tuple, which
 * contains values, and a Map which gives the index in the Tuple for a given
 * field name.
 *
 * @author	<A HREF="mailto:alia.mourad@rd.francetelecom.com><b>
 *			Mourad Alia
 *			</b></A>
 *		<A HREF="mailto:alexandre.lefebvre@rd.francetelecom.com><b>
 * 			Alexandre Lefebvre
 </b></A>
 */
//TODO move this class to a eval.jorm package since it is JORM specific.
public class TuplePNameGetter implements PNameGetter {

	/**
     * This Tuple contains request fields and in particular fields which
     * composed a PName.
     */
	private Tuple currentTuple;

	/**
     * This map contains the association between the PName field name and
     * the index in the tuple. This field describes the mapping between the
     * PName structure and the fields availlable in the tuple.
     *
     * key: String PName field name
     * value: Integer index in the Tuple
     */
	private Map fields;

    private PNameGetter png;

	private Operand pnc;

	/**
	 * Constructs an empty TuplePNameGetter.
	 */
	TuplePNameGetter() {
	}

	/**
	 * Constructs a TuplePNameGetter from a	set of associations (field name,
	 * index in tuple).
	 * @param fieldName2index is the Map which contains the association field
	 * name -> index in tuple.
	 */
	public TuplePNameGetter(Map fieldName2index, Operand nc)
            throws MedorException {
		this.fields = fieldName2index;
        pnc = nc;
    }

	/**
	 * Sets the current tuple to a tuple passed as a parameter.
	 * @param t is the tuple which becomes the current tuple.
	 */
	public void setTuple(Tuple t) {
		this.currentTuple = t;
	}

	/**
	 * Sets the field name to index association map.
	 * @param fieldName2index is the Map which contains the association field
	 * name -> index in tuple.
	 */
	public void setMap(Map fieldName2index) {
		this.fields = fieldName2index;
	}

	//Methods from PNameGetter
	/**
	 * Retrieves the value of a byte[] field.
	 * @param fn	A String containing the name of the field.
	 * @param context A context used by PNameGetter to extract the related data.
	 *                In in the case of RDB, it can be the result set.
	 * @return	The byte[] value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							potentially accessing the DSI.
	 */
	public byte[] pngetByteArrayField(String fn, Object context)
		throws PException {
		try {
            Integer in = (Integer) fields.get(fn);
            if (in == null)
                throw new PExceptionTyping("Field " + fn + " has not been found");
			return currentTuple.getByteArray(in.intValue());
		}
		catch (MedorException e) {
			throw new PException(e.getMessage());
		}
	}

	/**
	 * It retrieves the value of a byte field.
	 * @param fn	A String containing the name of the field.
	 * @param context A context used by PNameGetter to extract the related data.
	 *                In in the case of RDB, it can be the result set.
	 * @return	The byte value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							potentially accessing the DSI.
	 */
	public byte pngetByteField(String fn, Object context)
		throws PException {
		try {
            Integer in = (Integer) fields.get(fn);
            if (in == null)
                throw new PExceptionTyping("Field " + fn + " has not been found");
			if (currentTuple.isDefined(in.intValue())) {
                return currentTuple.getByte(in.intValue());
            } else {
				if (png == null) {
					png = (PNameGetter)
						((PNameCoder)pnc.getObject()).getNull().encodeAbstract();
				}
                return png.pngetByteField(fn, null);
            }
		}
		catch (MedorException e) {
			throw new PException(e.getMessage());
		}
	}

	/**
	 * It retrieves the value of a char field.
	 * @param fn	A String containing the name of the field.
	 * @param context A context used by PNameGetter to extract the related data.
	 *                In in the case of RDB, it can be the result set.
	 * @return	The char value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							potentially accessing the DSI.
	 */
	public char pngetCharField(String fn, Object context)
		throws PException {
		try {
            Integer in = (Integer) fields.get(fn);
            if (in == null)
                throw new PExceptionTyping("Field " + fn + " has not been found");
            if (currentTuple.isDefined(in.intValue())) {
                return currentTuple.getChar(in.intValue());
            } else {
                return png.pngetCharField(fn, null);
            }
		}
		catch (MedorException e) {
			throw new PException(e.getMessage());
		}
	}

	/**
	 * It retrieves the value of a char[] field.
	 * @param fn	A String containing the name of the field.
	 * @param context A context used by PNameGetter to extract the related data.
	 *                In in the case of RDB, it can be the result set.
	 * @return	The char[] value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							potentially accessing the DSI.
	 */
	public char[] pngetCharArrayField(String fn, Object context)
		throws PException {
		try {
            Integer in = (Integer) fields.get(fn);
            if (in == null)
                throw new PExceptionTyping("Field " + fn + " has not been found");
			return currentTuple.getCharArray(in.intValue());
		}
		catch (MedorException e) {
			throw new PException(e.getMessage());
		}
	}

	/**
	 * It retrieves the value of a short field.
	 * @param fn	A String containing the name of the field.
	 * @param context A context used by PNameGetter to extract the related data.
	 *                In in the case of RDB, it can be the result set.
	 * @return	The short value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							potentially accessing the DSI.
	 */
	public short pngetShortField(String fn, Object context)
		throws PException {
		try {
            Integer in = (Integer) fields.get(fn);
            if (in == null)
                throw new PExceptionTyping("Field " + fn + " has not been found");
            if (currentTuple.isDefined(in.intValue())) {
                return currentTuple.getShort(in.intValue());
            } else {
				if (png == null) {
					png = (PNameGetter)
						((PNameCoder)pnc.getObject()).getNull().encodeAbstract();
				}
                return png.pngetShortField(fn, null);
            }
		}
		catch (MedorException e) {
			throw new PException(e.getMessage());
		}
	}

	/**
	 * It retrieves the value of a Byte field.
	 * @param fn	A String containing the name of the field.
	 * @param ctxt	A context used by PNameGetter to extract the related data.
	 * @return	The Byte value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							potentially accessing the DSI.
	 */
	public Byte pngetObyteField(String fn, Object ctxt)
		throws PException {
		try {
            Integer in = (Integer) fields.get(fn);
            if (in == null)
                throw new PExceptionTyping("Field " + fn + " has not been found");
			return new Byte(currentTuple.getByte(in.intValue()));
		}
		catch (MedorException e) {
			throw new PException(e.getMessage());
		}

	}

	/**
	 * It retrieves the value of a Character field.
	 * @param fn	A String containing the name of the field.
	 * @param ctxt	A context used by PNameGetter to extract the related data.
	 * @return	The Character value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							potentially accessing the DSI.
	 */
	public Character pngetOcharField(String fn, Object ctxt)
		throws PException {
		try {
            Integer in = (Integer) fields.get(fn);
            if (in == null)
                throw new PExceptionTyping("Field " + fn + " has not been found");
			return new Character(currentTuple.getChar(in.intValue()));
		}
		catch (MedorException e) {
			throw new PException(e.getMessage());
		}
	}

	/**
	 * It retrieves the value of a Short field.
	 * @param fn	A String containing the name of the field.
	 * @param ctxt	A context used by PNameGetter to extract the related data.
	 * @return	The Short value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							potentially accessing the DSI.
	 */
	public Short pngetOshortField(String fn, Object ctxt)
		throws PException {
		try {
            Integer in = (Integer) fields.get(fn);
            if (in == null)
                throw new PExceptionTyping("Field " + fn + " has not been found");
			return new Short(currentTuple.getShort(in.intValue()));
		}
		catch (MedorException e) {
			throw new PException(e.getMessage());
		}
	}

	/**
	 * It retrieves the value of an Integer field.
	 * @param fn	A String containing the name of the field.
	 * @param ctxt	A context used by PNameGetter to extract the related data.
	 * @return	The Integer value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							potentially accessing the DSI.
	 */
	public Integer pngetOintField(String fn, Object ctxt)
		throws PException {
		try {
            Integer in = (Integer) fields.get(fn);
            if (in == null)
                throw new PExceptionTyping("Field " + fn + " has not been found");
			return new Integer(currentTuple.getInt(in.intValue()));
		}
		catch (MedorException e) {
			throw new PException(e.getMessage());
		}
	}

	/**
	 * It retrieves the value of a Long field.
	 * @param fn	A String containing the name of the field.
	 * @param ctxt	A context used by PNameGetter to extract the related data.
	 * @return	The Long value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							potentially accessing the DSI.
	 */
	public Long pngetOlongField(String fn, Object ctxt)
		throws PExceptionTyping, PExceptionIO {
		try {
            Integer in = (Integer) fields.get(fn);
            if (in == null)
                throw new PExceptionTyping("Field " + fn + " has not been found");
			return new Long(currentTuple.getLong(in.intValue()));
		}
		catch (MedorException e) {
			throw new PExceptionTyping(e.getMessage());
		}
	}

	/**
	 * It retrieves the value of a date field.
	 * @param fn	A String containing the name of the field.
	 * @param ctxt	A context used by PNameGetter to extract the related data.
	 * @return	The java.util.Date value of the field as stored within the
	 * 			composite name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							potentially accessing the DSI.
	 */
	public Date pngetDateField(String fn, Object ctxt)
		throws PExceptionTyping, PExceptionIO {
		try {
            Integer in = (Integer) fields.get(fn);
            if (in == null)
                throw new PExceptionTyping("Field " + fn + " has not been found");
			return currentTuple.getDate(in.intValue());
		}
		catch (MedorException e) {
			throw new PExceptionTyping(e.getMessage());
		}
	}

	/**
	 * It retrieves the value of an int field.
	 * @param context A context used by PNameGetter to extract the related data.
	 *                In in the case of RDB, it can be the result set.
	 * @param fn	A String containing the name of the field.
	 * @return	The int value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							potentially accessing the DSI.
	 */
	public int pngetIntField(String fn, Object context)
		throws PException {
		try {
            Integer in = (Integer) fields.get(fn);
            if (in == null)
                throw new PExceptionTyping("Field " + fn + " has not been found");
            if (currentTuple.isDefined(in.intValue())) {
                return currentTuple.getInt(in.intValue());
            } else {
				if (png == null) {
					png = (PNameGetter)
						((PNameCoder)pnc.getObject()).getNull().encodeAbstract();
				}
                return png.pngetIntField(fn, null);
            }
		}
		catch (MedorException e) {
			throw new PException(e.getMessage());
		}
	}

	/**
	 * It retrieves the value of a long field.
	 * @param context A context used by PNameGetter to extract the related data.
	 *                In in the case of RDB, it can be the result set.
	 * @param fn	A String containing the name of the field.
	 * @return	The long value of the field as stored within the composite
	 * 			name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							potentially accessing the DSI.
	 */
	public long pngetLongField(String fn, Object context)
		throws PException {
		try {
            Integer in = (Integer) fields.get(fn);
            if (in == null)
                throw new PExceptionTyping("Field " + fn + " has not been found");
            if (currentTuple.isDefined(in.intValue())) {
                return currentTuple.getLong(in.intValue());
            } else {
				if (png == null) {
					png = (PNameGetter)
						((PNameCoder)pnc.getObject()).getNull().encodeAbstract();
				}
                return png.pngetLongField(fn, null);
            }
		}
		catch (MedorException e) {
			throw new PException(e.getMessage());
		}
	}

	/**
	 * It retrieves the value of a string field.
	 * @param context A context used by PNameGetter to extract the related data.
	 *                In in the case of RDB, it can be the result set.
	 * @param fn	A String containing the name of the field.
	 * @return	The java.lang.String value of the field as stored within the
	 * 			composite name.
	 * @exception	PExceptionTyping	There is no field with this name and
	 * 									this type within this composite name.
	 * @exception PExceptionIO	It is raised when a problem occurs while
	 * 							potentially accessing the DSI.
	 */
	public String pngetStringField(String fn, Object context)
		throws PExceptionTyping, PExceptionIO {
		try {
            Integer in = (Integer) fields.get(fn);
            if (in == null)
                throw new PExceptionTyping("Field " + fn + " has not been found");
			return currentTuple.getString(in.intValue());
		}
		catch (MedorException e) {
			throw new PExceptionTyping(e.getMessage());
		}
	}
}
