/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 * 
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

/**
 * Package definition.
 */
package org.objectweb.medor.eval.lib;

import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TupleStructure;
import org.objectweb.medor.eval.api.BinaryEvaluatedTC;
import org.objectweb.medor.eval.api.NodeEvaluator;
import org.objectweb.medor.filter.api.ParameterOperand;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.tuple.api.TupleCollection;

/**
 * This class represent a result of a Union operation. The result is equal to the left TupleCollection + the right
 * TupleCollection in this Order.
 */
class UnionEvaluatedTC implements BinaryEvaluatedTC {
    public UnionEvaluatedTC(QueryNode query, NodeEvaluator leftNodeEvaluator, NodeEvaluator rightNodeEvaluator, ParameterOperand[] parameters) throws MedorException {
        leftTCResult = leftNodeEvaluator.fetchData(parameters);
        rightTCResult = rightNodeEvaluator.fetchData(parameters);
        if (leftTCResult.isEmpty()) second = true;
    }

    private boolean second = false;// This variable tells us if we are iterating over the
                                   // left TupleCollection or the right one

    private TupleCollection leftTCResult, rightTCResult;
    private QueryNode query;

    public TupleStructure getMetaData() throws MedorException {
        return query.getTupleStructure();
    }

    public boolean isLast() throws MedorException {
        boolean end = false;
        if (!rightTCResult.isEmpty()) {
            end = rightTCResult.isLast();
        } else {
            end = leftTCResult.isLast();
        }
        return end;
    }

    public boolean next() throws MedorException {
        if (!second) {
            if (!leftTCResult.next()) {
                second = true;
                return (!rightTCResult.isEmpty());
            } else
                return true;
        } else
            return rightTCResult.next();
    }

    public void first() throws MedorException {
        leftTCResult.first();
        rightTCResult.first();
        second = false;
    }

    public int getRow() throws MedorException {
        if (!second)
            return leftTCResult.getRow();
        else
            return leftTCResult.getRow() + rightTCResult.getRow();
    }

    public Object getObject(int column) throws MedorException {
        if (second)
            return rightTCResult.getObject(column);
        else
            return leftTCResult.getObject(column);
    }

    public Tuple getTuple() throws MedorException {
        if (second)
            return rightTCResult.getTuple();
        else
            return leftTCResult.getTuple();
    }

    public Tuple getTuple(int row) throws MedorException {
        if (row <=0){
            throw new MedorException("Invalid Tuple index: "+ row);
        }else{
            try{
                return leftTCResult.getTuple(row);
            }catch (MedorException me){
                return rightTCResult.getTuple(row - getLeftCard());    // Test in the row value is to be added
            }
        }
    }

    /**
     * This function will calculate the cardinality of the left TupleCollection
     * @return int size of the TupleCollection
     */
    private int getLeftCard(){
        return 0; // TO DO
    }

    public boolean isEmpty() throws MedorException {
        return (leftTCResult.isEmpty() & rightTCResult.isEmpty());
    }

    public boolean row(int numTuple) throws MedorException {
        first();
        int cpt = 1;
        boolean go = true;
        while ((cpt <= numTuple) && (cpt > 1) && go) {
            if (!next()) go = false;
        }
        if (!go)
            return false;
        else {
            return true;
        }
    }

    public synchronized int getLeftTCCursor() throws MedorException {
        return leftTCResult.getRow();
    }

    public synchronized int getRightTCCursor() throws MedorException {
        return rightTCResult.getRow();
    }

}
