/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.filter.api;

import org.objectweb.jorm.type.api.PType;
import org.objectweb.medor.api.Cloneable;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TypingException;
import org.objectweb.medor.api.MalformedExpressionException;
import org.objectweb.medor.tuple.api.Tuple;

/**
 * This interface represents an arithmetic and/or a
 * boolean expression.
 * <p> It represents operations over operands. It is
 * represented by a tree where nodes represents operations and leafs operands.
 * An expression have two states:
 * <p>
 * Not compiled in wich case it can be modified.
 * <p>
 * Compiled, this state is possible after compilation method
 * CompileExpression().
 * <p>
 * Then it will be evaluable and its structure cannot be further modified.
 * <p> Here is a simple example:<br>
 * <tt> Expression exp = ...<br>
 * exp.CompileExpression();<br>
 * Operand op = exp.getResult();<br>
 * exp.evaluate(......);<br>
 * System.out.println("the result = "+ op.getInt());<br>
 * </tt>
 */
public interface Expression extends Cloneable {

	/**
	 * It gets the result type of this expression
	 * @return a PType object
	 */
	public PType getType();

	/**
	 * Gets wether this expression is compiled or not
	 * @return true if the expression is compiled and false if it is not
	 */
	public boolean isCompiled();

	/**
	 * Checks the semantic integrity of an expression.
	 * It checks that all types are compatible and prepare the expression to be
	 * evaluable.It also creates buffers where stores the result. Notes that
	 * when evaluating there is no creation of new objects. This method change
	 * the state of this expression object, it will be evaluable and not
	 * modifiable.
	 * @throws TypingException when incompatible types error
	 * occures.
	 * @throws MalformedExpressionException if syntax error
	 */
	void compileExpression()
		throws TypingException, MalformedExpressionException;

	/**
	 * Gets the operand in wich the result will be put in
	 * @return an Operand Object
	 * @throws IllegalStateException if this expression is not
	 * compiled.
	 */
	public Operand getResult() throws IllegalStateException;

	/**
	 * It evaluates the expression tree and puts the result into the operand
	 * result.
	 * @param  values a list of ParameterOperand
	 * @param t Tuple
	 * @throws MedorException if evaluation error
	 * @throws IllegalStateException if this expression is not
	 * compiled.
	 */
	public void evaluate(ParameterOperand[] values, Tuple t)
		throws MedorException, IllegalStateException;
}
