/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 * 
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.filter.api;

import org.objectweb.medor.api.TypingException;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.tuple.api.TupleCollection;

import java.util.Date;
import java.sql.Time;
import java.sql.Timestamp;
import java.math.BigDecimal;

/**
 * This interface represents a byte, char, int,...or an object value.
 * <p>When
 * of primitive type, widing conversions allowed by the Java programming
 * language are used to get the value in different types.
 * <p>
 * Each getXXX() method returns the operand value of the XXX type.
 */
public interface Operand extends Expression, org.objectweb.medor.api.Cloneable {

	/**
	 * get the operand result as a boolean value
	 * @return a boolean value
	 * @throws TypingException if it is impossible to return a boolean value
	 */
	boolean getBoolean() throws TypingException;

	/**
	 * get the operand result as a byte value
	 * @return an int value
	 * @throws TypingException if it is impossible to return an int value
	 */
	int getInt() throws TypingException;

	/**
	 * get the operand result as a byte value
	 * @return a byte value
	 * @throws TypingException if it is impossible to return an int value
	 */
	byte getByte() throws TypingException;

	/**
	 * get the operand result as a byte[] value
	 * @return a byte[] value
	 * @throws TypingException if it is impossible to return an int value
	 */
	byte[] getByteArray() throws TypingException;

	/**
	 * get the operand result as a short value
	 * @return a short value
	 * @throws TypingException if it is impossible to return a short value
	 */
	short getShort() throws TypingException;

	/**
	 * get the operand result as a long value
	 * @return a long value
	 * @throws TypingException if it is impossible to return a long value
	 */
	long getLong() throws TypingException;

	/**
	 * get the operand result as a float value
	 * @return a float value
	 * @throws TypingException if it is impossible to return a float value;
	 */
	float getFloat() throws TypingException;

	/**
	 * get the operand result as a double value
	 * @return a double value
	 * @throws TypingException if it is impossible to return a double value
	 */
	double getDouble() throws TypingException;

	/**
	 * get the operand result as a char value
	 * @return a char value
	 * @throws TypingException if it is impossible to return a char value;
	 */
	char getChar() throws TypingException;

	/**
	 * get the operand result as a char[] value
	 * @return a char[] value
	 * @throws TypingException if it is impossible to return a char value;
	 */
	char[] getCharArray() throws TypingException;

	/**
	 * evalute the expression result and return a string value
	 * @return string result
	 * @throws TypingException if it is impossible to return a string value
	 */
	String getString() throws TypingException;

	/**
	 * evalute the expression result and return a string value
	 * @return string result
	 * @throws TypingException if it is impossible to return a string value
	 */
	BigDecimal getBigDecimal() throws TypingException;

	/**
	 * get the operand result a java.util.Date value
	 * @return a date value
	 * @throws TypingException if it is impossible to return java.util.Date
	 */
	Date getDate() throws TypingException;

	/**
	 * get the operand result a java.sql.Time value
	 * @return a Time value
	 * @throws TypingException if it is impossible to return java.sql.Time
	 */
	Time getTime() throws TypingException;

	/**
	 * get the operand result a java.sql.Timestamp value
	 * @return a Timestamp value
	 * @throws TypingException if it is impossible to return java.sql.Timestamp
	 */
	Timestamp getTimestamp() throws TypingException;

	/**
	 * get the operand result as a Tuple value
	 * @return a Tuple value
	 * @throws TypingException if it is impossible to return a Tuple
	 */
	Tuple getTuple() throws TypingException;

	/**
	 * get the operand result as a TupleCollection value
	 * @return a TupleCollection value
	 * @throws TypingException if it is impossible to return a TupleCollection
	 */
	TupleCollection getTupleCollection() throws TypingException;

	/**
	 * get the operand result as a java.lang.Object value
	 * @return an object
	 */
	Object getObject();

    /**
     * Indicates if the operand has been defined, ie is the value returned if
     * pertinent
     */
    boolean isDefined();

    void setIsDefined(boolean isdefined);
}
