/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre, S. Chassande-Barrioz
 */
package org.objectweb.medor.filter.jorm.lib;

import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.naming.api.PNameGetter;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.medor.api.EvaluationException;
import org.objectweb.medor.api.TypingException;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.filter.api.CalculatedParameterOperand;
import org.objectweb.medor.filter.api.ParameterOperand;
import org.objectweb.medor.filter.api.Operand;
import org.objectweb.medor.filter.api.VariableOperand;
import org.objectweb.medor.filter.lib.BasicParameterOperand;
import org.objectweb.medor.type.lib.QType;

/**
 * This implementation of the CalculatedParameterOperand interface encode a
 * a PName. This PName must be given in parameter among the ParameterOperand
 * of the evaluate method.
 * There are two constructor weither the PName structure
 *
 * @author S.Chassande-Barrioz
 */
public class EncodePName
	extends BasicParameterOperand
	implements CalculatedParameterOperand {

	String fieldName = null;

	public EncodePName(PType type, String parameterName) {
		super(type, parameterName);
		fieldName = null;
	}

	public EncodePName(PType type, String parameterName, String fieldName) {
		super(type, parameterName);
		this.fieldName = fieldName;
	}

	// IMPLEMENTATION OF THE CalculatedParameterOperand INTERFACE //
	//------------------------------------------------------------//

	public void evaluate(ParameterOperand[] pos) throws MedorException {
		PName pn = null;
		for (int i = 0; i < pos.length; i++) {
			if (pos[i] != null
                    && name.equals(pos[i].getName())
				    && pos[i].getType().getTypeCode() == QType.TYPECODE_PNAME) {
				pn = (PName) pos[i].getObject();
			}
		}
		if (pn == null)
			throw new EvaluationException("Impossible to evaluate this parameter wihtout the parameter " + name);
        try {
            assignEncodedValue(pn, this, fieldName);
        } catch (PException e) {
            throw new MedorException("Impossible to encode the PName", e);
        }
    }

    public final static void assignEncodedValue(PName pn,
                                                VariableOperand op,
                                                String fieldname)
            throws PException, MedorException {
        if (fieldname == null) {
            switch (op.getType().getTypeCode()) {
            case QType.TYPECODE_BYTE:
                op.setValue(pn.encodeByte());
                break;
            case QType.TYPECODE_CHAR:
                op.setValue(pn.encodeChar());
                break;
            case QType.TYPECODE_INT:
                op.setValue(pn.encodeInt());
                break;
            case QType.TYPECODE_LONG:
                op.setValue(pn.encodeLong());
                break;
            case QType.TYPECODE_SHORT:
                op.setValue(pn.encodeShort());
                break;
            case QType.TYPECODE_OBJBYTE:
                op.setValue(pn.encodeObyte());
                break;
            case QType.TYPECODE_OBJCHAR:
                op.setValue(pn.encodeOchar());
                break;
            case QType.TYPECODE_OBJINT:
                op.setValue(pn.encodeOint());
                break;
            case QType.TYPECODE_OBJLONG:
                op.setValue(pn.encodeOlong());
                break;
            case QType.TYPECODE_OBJSHORT:
                op.setValue(pn.encodeOshort());
                break;
            case QType.TYPECODE_DATE:
                op.setValue(pn.encodeDate());
                break;
            case QType.TYPECODE_STRING:
                op.setValue(pn.encodeString());
                break;
            default:
                throw new EvaluationException("Unsupported naming type: (jormName:"
                        + op.getType().getJormName()
                        +", code" + op.getType().getTypeCode() + ")");
            }
        } else {
            // Composite PName
            Object conn = null;
            PNameGetter png = (PNameGetter) pn.encodeAbstract();
            switch (op.getType().getTypeCode()) {
            case QType.TYPECODE_BYTE:
                op.setValue(png.pngetByteField(fieldname, conn));
                break;
            case QType.TYPECODE_CHAR:
                op.setValue(png.pngetCharField(fieldname, conn));
                break;
            case QType.TYPECODE_INT:
                op.setValue(png.pngetIntField(fieldname, conn));
                break;
            case QType.TYPECODE_LONG:
                op.setValue(png.pngetLongField(fieldname, conn));
                break;
            case QType.TYPECODE_SHORT:
                op.setValue(png.pngetShortField(fieldname, conn));
                break;
            case QType.TYPECODE_OBJBYTE:
                op.setValue(png.pngetObyteField(fieldname, conn));
                break;
            case QType.TYPECODE_OBJCHAR:
                op.setValue(png.pngetOcharField(fieldname, conn));
                break;
            case QType.TYPECODE_OBJINT:
                op.setValue(png.pngetOintField(fieldname, conn));
                break;
            case QType.TYPECODE_OBJLONG:
                op.setValue(png.pngetOlongField(fieldname, conn));
                break;
            case QType.TYPECODE_OBJSHORT:
                op.setValue(png.pngetOshortField(fieldname, conn));
                break;
            case QType.TYPECODE_DATE:
                op.setValue(png.pngetDateField(fieldname, conn));
                break;
            case QType.TYPECODE_STRING:
                op.setValue(png.pngetStringField(fieldname, conn));
                break;
            default:
                throw new EvaluationException("Unsupported naming type" + op.getType());
            }
        }

    }
}
