/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 * 
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.filter.jorm.lib;

import org.objectweb.jorm.naming.api.PNamingContext;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.medor.api.MalformedExpressionException;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TypingException;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.filter.api.ParameterOperand;
import org.objectweb.medor.filter.api.Operand;
import org.objectweb.medor.filter.lib.BasicBinaryOperator;
import org.objectweb.medor.filter.lib.BasicVariableOperand;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.type.lib.QTypeTuple;

/**
 * A SinglePName is an Operator representing the construction of a PName
 * from a NamingContext and a single Field.
 * <p>
 * The NamingContext is provided as a ParameterOperand. The name of the
 * ParameterOperand is defined to be:
 * <ol><li>The fully qualified class name for the PName of the class itself</li>
 * <li>The fully qualified class name, to which the attribute name is
 * concatenated for a reference PName</li>
 * </ol>
 * The Field is provided as a FieldOperand.
 *
 * @author	<A HREF="mailto:alia.mourad@rd.francetelecom.com><b>
 *			Mourad Alia
 *			</b></A>
 *		<A HREF="mailto:alexandre.lefebvre@rd.francetelecom.com><b>
 * 			Alexandre Lefebvre
 </b></A>
 */
public class SinglePName extends BasicBinaryOperator {

	/**
	 * Constructs a SinglePName from a ParameterOperand representing the
	 * naming context, a FieldOperand, and the PType of the corresponding
	 * PName.
	 * @param f is the FieldOperand from which the value is taken.
	 * @param p is the Operand representing the naming context.
	 * @param nameType is the PType of the PName created by this operator.
	 */
	public SinglePName(FieldOperand f, Operand p, PType nameType) {
		this.left = f;
		this.right = p;
		this.type = nameType;
	}

	/**
	 *
	 */
	public void evaluate(ParameterOperand[] pos, Tuple t)
		throws MedorException {
		//assign the value of the NamingContext
		right.evaluate(pos, t);
		//gets the naming context
		PNamingContext nc =
			(PNamingContext) ((Operand) right).getObject();
		if (result == null)
			result = new BasicVariableOperand(type);
		PType fieldType = ((FieldOperand) left).getField().getType();
		int i = ((FieldOperand) left).getIndex();
		try {
			switch (fieldType.getTypeCode()) {
			case QTypeTuple.TYPECODE_BYTE:
                if (!t.isDefined(i)) {
                    result.setValue(nc.getNull());
                } else {
				    result.setValue(nc.decodeByte(t.getByte(i)));
                }
				break;
			case QTypeTuple.TYPECODE_CHAR:
                if (!t.isDefined(i)) {
                    result.setValue(nc.getNull());
                } else {
				    result.setValue(nc.decodeChar(t.getChar(i)));
                }
				break;
			case QTypeTuple.TYPECODE_DATE:
				result.setValue(nc.decodeDate(t.getDate(i)));
				break;
			case QTypeTuple.TYPECODE_INT:
                if (!t.isDefined(i)) {
                    result.setValue(nc.getNull());
                } else {
				    result.setValue(nc.decodeInt(t.getInt(i)));
                }
				break;
			case QTypeTuple.TYPECODE_LONG:
                if (!t.isDefined(i)) {
                    result.setValue(nc.getNull());
                } else {
				    result.setValue(nc.decodeLong(t.getLong(i)));
                }
				break;
			case QTypeTuple.TYPECODE_SHORT:
                if (!t.isDefined(i)) {
                    result.setValue(nc.getNull());
                } else {
				    result.setValue(nc.decodeShort(t.getShort(i)));
                }
				break;
			case QTypeTuple.TYPECODE_STRING:
				result.setValue(nc.decodeString(t.getString(i)));
				break;
			case QTypeTuple.TYPECODE_OBJBYTE:
				result.setValue(nc.decodeObyte((Byte) t.getObject(i)));
				break;
			case QTypeTuple.TYPECODE_OBJCHAR:
				result.setValue(nc.decodeOchar((Character) t.getObject(i)));
				break;
			case QTypeTuple.TYPECODE_OBJINT:
				result.setValue(nc.decodeOint((Integer) t.getObject(i)));
				break;
			case QTypeTuple.TYPECODE_OBJLONG:
				result.setValue(nc.decodeOlong((Long) t.getObject(i)));
				break;
			case QTypeTuple.TYPECODE_OBJSHORT:
				result.setValue(nc.decodeOshort((Short) t.getObject(i)));
				break;
			default:
				throw new MedorException("Impossible to construct PName from field"); // This case don't accure if we compile at first
			}
		}
		catch (MedorException e) {
			throw e;
		}
		catch (Exception e) {
			throw new MedorException("Error while building SinglePName", e);
		}
	}

	/**
	 * Checks the semantic integrity of an expression.
	 * It checks that all types are compatible and prepare the expression to be
	 * evaluable. It also creates buffers where stores the result. Notes that
	 * when evaluating there is no creation of new objects. This method changes
	 * the state of this expression object, it will be evaluable and not
	 * modifiable.
	 * @throws TypingException when incompatible types error
	 * occures.
	 * @throws MalformedExpressionException if syntax error
	 */
	public void compileExpression()
		throws TypingException, MalformedExpressionException {
		if ((left != null) && (right != null)) {
			result = new BasicVariableOperand(type);
			verified = true;
		}
		else
		// Operands not yet assigned
			throw new MalformedExpressionException("null children value");
	}

	public FieldOperand getField() {
		return (FieldOperand) left;
	}

    public ParameterOperand getPNameManagerParameter() {
        return (right instanceof ParameterOperand ? (ParameterOperand) right : null);
    }

	//Implementation of Operator interface
	/**
	 * There is no Java operator associated to a PName
	 */
	public String getOperatorString() {
		return "singlePName";
	}
}
