/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 * 
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.filter.lib;

import org.objectweb.medor.api.MalformedExpressionException;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TypingException;
import org.objectweb.medor.filter.api.BinaryArithmeticOperator;
import org.objectweb.medor.filter.api.ParameterOperand;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.type.lib.QTypeTuple;
import org.objectweb.medor.type.lib.PTypeSpaceMedor;

import java.util.Date;
import java.sql.Time;
import java.sql.Timestamp;

/**
 *
 * @author Sebastien Chassande-Barrioz
 */
public abstract class BasicBinaryArithmeticOperator
        extends BasicBinaryOperator
        implements BinaryArithmeticOperator {


    // IMPLEMENTATION OF THE Expression INTERFACE //
    //--------------------------------------------//

    public void evaluate(ParameterOperand[] pos, Tuple t)
            throws MedorException, IllegalStateException {
        try {
            left.evaluate(pos, t);
            right.evaluate(pos, t);
            switch (type.getTypeCode()) {
                case QTypeTuple.TYPECODE_BYTE:
                case QTypeTuple.TYPECODE_OBJBYTE:
                    result.setValue(evaluate(left.getResult().getByte(),
									right.getResult().getByte()));
                    break;
                case QTypeTuple.TYPECODE_CHAR:
                case QTypeTuple.TYPECODE_OBJCHAR:
                    result.setValue(evaluate(left.getResult().getChar(),
									right.getResult().getChar()));
                    break;
                case QTypeTuple.TYPECODE_SHORT:
                case QTypeTuple.TYPECODE_OBJSHORT:
                    result.setValue(evaluate(left.getResult().getShort(),
									right.getResult().getShort()));
                    break;
                case QTypeTuple.TYPECODE_INT:
                case QTypeTuple.TYPECODE_OBJINT:
                    result.setValue(evaluate(left.getResult().getInt(),
									right.getResult().getInt()));
                    break;
                case QTypeTuple.TYPECODE_LONG:
                case QTypeTuple.TYPECODE_OBJLONG:
                    result.setValue(evaluate(left.getResult().getLong(),
									right.getResult().getLong()));
                    break;
                case QTypeTuple.TYPECODE_FLOAT:
                case QTypeTuple.TYPECODE_OBJFLOAT:
                    result.setValue(evaluate(left.getResult().getFloat(),
									right.getResult().getFloat()));
                    break;
                case QTypeTuple.TYPECODE_DOUBLE:
                case QTypeTuple.TYPECODE_OBJDOUBLE:
                    result.setValue(evaluate(left.getResult().getDouble(),
									right.getResult().getDouble()));
                    break;
                case QTypeTuple.TYPECODE_STRING:
                    result.setValue(evaluate(left.getResult().getString(),
									right.getResult().getString()));
                    break;
                default:
				    // This case don't occurre if we compile at first
                    throw new TypingException("Incompatible type");
            }
        } catch (NullPointerException e) {
			// result == null or missing children objects
            throw new IllegalStateException("Unevaluable Expression: Not compiled");
        }
    }

    // IMPLEMENTATION OF THE BinaryArithmeticOperator INTERFACE //
    //----------------------------------------------------------//
    public abstract int evaluate(int op1, int op2);

    public abstract int evaluate(int op1, short op2);

    public abstract long evaluate(int op1, long op2);

    public abstract float evaluate(int op1, float op2);

    public abstract double evaluate(int op1, double op2);

    public abstract float evaluate(float op1, float op2);

    public abstract float evaluate(float op1, short op2);

    public abstract float evaluate(float op1, int op2);

    public abstract float evaluate(float op1, long op2);

    public abstract double evaluate(float op1, double op2);

    public abstract int evaluate(char op1, char op2);

    public abstract String evaluate(char op1, String op2)
            throws TypingException;

    public abstract long evaluate(long op1, long op2);

    public abstract long evaluate(long op1, short op2);

    public abstract long evaluate(long op1, int op2);

    public abstract float evaluate(long op1, float op2);

    public abstract double evaluate(long op1, double op2);

    public abstract double evaluate(double op1, double op2);

    public abstract double evaluate(double op1, short op2);

    public abstract double evaluate(double op1, int op2);

    public abstract double evaluate(double op1, float op2);

    public abstract double evaluate(double op1, long op2);

    public abstract String evaluate(String op1, String op2)
            throws TypingException;

    public abstract String evaluate(String op1, char op2)
            throws TypingException;

    public abstract Timestamp evaluate(Timestamp op1, Timestamp op2)
            throws TypingException;

    public abstract Time evaluate(Time op1, Time op2) throws TypingException;

    public abstract Date evaluate(Date op1, Date op2) throws TypingException;

    public void compileExpression()
		throws TypingException, MalformedExpressionException {
        if ((left != null) && (right != null)) {
            left.compileExpression();
            leftResult = left.getResult();
            right.compileExpression();
            rightResult = right.getResult();
            if (!(left.getType() == null || right.getType() == null)) {
                if ((PTypeSpaceMedor.isArithmeticType(left.getType()))
					&& (PTypeSpaceMedor.isArithmeticType(right.getType()))) {
                    if (left.getType().isa(right.getType())) {
                        type = right.getType();
                    } else {
                        type = left.getType();
                    }
                    result = new BasicVariableOperand(type);
                    verified = true;
                } else
                    throw new TypingException("Attempt an aritmetic type " + left.getType().getJavaName() + "  " + right.getType().getJavaName()); // it is not a numeric type
            } else
                throw new MalformedExpressionException("null children value");
        } else
		// Operands not yet assigned
            throw new MalformedExpressionException("null children value");
    }
}
