/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 * 
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.filter.lib;

import org.objectweb.jorm.type.api.PType;
import org.objectweb.medor.api.MalformedExpressionException;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TypingException;
import org.objectweb.medor.filter.api.Comparator;
import org.objectweb.medor.filter.api.ParameterOperand;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.type.lib.PTypeSpaceMedor;
import org.objectweb.medor.type.lib.QTypeTuple;

/**
 *
 * @author Sebastien Chassande-Barrioz
 */
public abstract class BasicComparator
        extends BasicBinaryOperator
        implements Comparator {

    // IMPLEMENTATION OF THE Expression INTERFACE //
    //--------------------------------------------//

    public PType getType() {
        return PTypeSpaceMedor.BOOLEAN;
    }

    /**
     * We supose that the comparisons can be performed between all types.
     */
    public void evaluate(ParameterOperand[] pos, Tuple t)
            throws MedorException, IllegalStateException {
        left.evaluate(pos, t);
        right.evaluate(pos, t);
        PType rtype = PTypeSpaceMedor.getResultType(left.getType(),
                right.getType());
        switch (rtype.getTypeCode()) {
		case QTypeTuple.TYPECODE_BOOLEAN:
			result.setValue(evaluate(left.getResult().getBoolean(),
				right.getResult().getBoolean()));
			break;
		case QTypeTuple.TYPECODE_BYTE:
			result.setValue(evaluate(left.getResult().getByte(),
				right.getResult().getByte()));
			break;
		case QTypeTuple.TYPECODE_CHAR:
			result.setValue(evaluate(left.getResult().getChar(),
				right.getResult().getChar()));
			break;
		case QTypeTuple.TYPECODE_SHORT:
			result.setValue(evaluate(left.getResult().getShort(),
				right.getResult().getShort()));
			break;
		case QTypeTuple.TYPECODE_INT:
			result.setValue(evaluate(left.getResult().getInt(),
				right.getResult().getInt()));
			break;
		case QTypeTuple.TYPECODE_LONG:
			result.setValue(evaluate(left.getResult().getLong(),
				right.getResult().getLong()));
			break;
		case QTypeTuple.TYPECODE_FLOAT:
			result.setValue(evaluate(left.getResult().getFloat(),
				right.getResult().getFloat()));
			break;
		case QTypeTuple.TYPECODE_DOUBLE:
			result.setValue(evaluate(left.getResult().getDouble(),
				right.getResult().getDouble()));
			break;
		case QTypeTuple.TYPECODE_STRING:
			result.setValue(evaluate(left.getResult().getString(),
				right.getResult().getString()));
			break;
		case QTypeTuple.TYPECODE_OBJBOOLEAN:
		case QTypeTuple.TYPECODE_OBJBYTE:
		case QTypeTuple.TYPECODE_OBJCHAR:
		case QTypeTuple.TYPECODE_OBJSHORT:
		case QTypeTuple.TYPECODE_OBJINT:
		case QTypeTuple.TYPECODE_OBJLONG:
		case QTypeTuple.TYPECODE_OBJFLOAT:
		case QTypeTuple.TYPECODE_OBJDOUBLE:
			result.setValue(evaluate(left.getResult().getObject(),
				right.getResult().getObject()));
			break;
		default:
			throw new TypingException("Unmanaged the type: ("
				+ " typeCode: " + rtype.getTypeCode()
				+ " jormName: " + rtype.getJormName()
				+ " javaName: " + rtype.getJavaName()
				+ ") this=" + this
			);
        }
    }


// IMPLEMENTATION OF THE Comparator INTERFACE //
//--------------------------------------------//

    public abstract boolean evaluate(Object op1, Object op2);

    public abstract boolean evaluate(int op1, int op2);

    public abstract boolean evaluate(int op1, short op2);

    public abstract boolean evaluate(int op1, long op2);

    public abstract boolean evaluate(int op1, float op2);

    public abstract boolean evaluate(int op1, double op2);

    public abstract boolean evaluate(float op1, float op2);

    public abstract boolean evaluate(float op1, short op2);

    public abstract boolean evaluate(float op1, int op2);

    public abstract boolean evaluate(float op1, long op2);

    public abstract boolean evaluate(float op1, double op2);

    public abstract boolean evaluate(char op1, char op2);

    public abstract boolean evaluate(long op1, long op2);

    public abstract boolean evaluate(long op1, float op2);

    public abstract boolean evaluate(long op1, double op2);

    public abstract boolean evaluate(double op1, double op2);

    public abstract boolean evaluate(double op1, short op2);

    public abstract boolean evaluate(double op1, int op2);

    public abstract boolean evaluate(double op1, float op2);

    public abstract boolean evaluate(double op1, long op2);

    public abstract boolean evaluate(String op1, String op2);

    public abstract boolean evaluate(boolean op1, boolean op2);

    public void compileExpression()
            throws TypingException, MalformedExpressionException {
        if ((left != null) && (right != null)) {
            left.compileExpression();
            right.compileExpression();
            if (!(left.getType() == null || right.getType() == null)) {
                if ((PTypeSpaceMedor.isArithmeticType(left.getType()) &&
                        (PTypeSpaceMedor.isArithmeticType(right.getType())))
                        //Equal operations OK between arithmetic types
                        ||
                        (left.getType().getTypeCode() ==
                        (right.getType().getTypeCode()))) {
                    //between operands which have the same types
                    if (!(left.getType().isa(right.getType())) &&
                            !(right.getType().isa(left.getType()))) {
                        throw new TypingException("incompatible types");
                    } else {
                        result = new BasicVariableOperand(PTypeSpaceMedor.BOOLEAN);
                        leftResult = left.getResult();
                        rightResult = right.getResult();
                        verified = true;
                    }
                } else
                // it is not a numeric type
                    throw new TypingException("Attempt an aritmetic type on a non arithmetic type");
            } else
            //syntax error
                throw new MalformedExpressionException("Null children value in Expression");
        } else
        // Operands not yet assigned
            throw new MalformedExpressionException("Null children value in Expression");
    }
}
