/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 * 
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.filter.lib;

import org.objectweb.jorm.type.api.PType;
import org.objectweb.medor.api.Cloneable;
import org.objectweb.medor.api.TypingException;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.filter.api.Operand;
import org.objectweb.medor.filter.api.ParameterOperand;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.tuple.api.TupleCollection;
import org.objectweb.medor.type.lib.QTypeTuple;
import org.objectweb.medor.type.lib.PTypeSpaceMedor;
import org.objectweb.medor.type.lib.QType;

import java.util.Date;
import java.sql.Time;
import java.sql.Timestamp;
import java.math.BigDecimal;

/**
 *
 * @author Sebastien Chassande-Barrioz
 */
public class BasicOperand implements Operand {

    protected long longValue;
    protected double doubleValue;
    protected Object objectValue;

    protected PType type = null;

    protected boolean isDefined = true;

    protected BasicOperand(PType p) {
        type = p;
    }

    public BasicOperand(boolean p) {
        type = PTypeSpaceMedor.BOOLEAN;
        longValue = (p ? 1 : 0);
        doubleValue = (p ? 1 : 0);
    }

    public BasicOperand(char p) {
        type = PTypeSpaceMedor.CHAR;
        longValue = p;
        doubleValue = p;
    }

	public BasicOperand(char[] p) {
		type = PTypeSpaceMedor.CHARARRAY;
		objectValue = p;
	}

     public BasicOperand(byte p) {
        type = PTypeSpaceMedor.BYTE;
        longValue = p;
        doubleValue = p;
    }

	public BasicOperand(byte[] p) {
		type = PTypeSpaceMedor.BYTEARRAY;
		objectValue = p;
	}

    public BasicOperand(short p) {
        type = PTypeSpaceMedor.SHORT;
        longValue = p;
        doubleValue = p;
    }

    public BasicOperand(int p) {
        type = PTypeSpaceMedor.INT;
        longValue = p;
        doubleValue = p;
    }

    public BasicOperand(long p) {
        type = PTypeSpaceMedor.LONG;
        longValue = p;
        doubleValue = p;
    }

    public BasicOperand(float p) {
        type = PTypeSpaceMedor.FLOAT;
        doubleValue = p;
    }

    public BasicOperand(double p) {
        type = PTypeSpaceMedor.DOUBLE;
        doubleValue = p;
    }

    public BasicOperand(String p) {
        type = PTypeSpaceMedor.STRING;
        objectValue = p;
    }

    public BasicOperand(Date p) {
        type = PTypeSpaceMedor.DATE;
        objectValue = p;
    }

    public BasicOperand(Time p) {
		//type = PTypeSpaceMedor.TIME;
        objectValue = p;
    }

    public BasicOperand(Timestamp p) {
		//type = PTypeSpaceMedor.TIMESTAMP;
        objectValue = p;
    }

    public BasicOperand(Object p, PType type) {
		objectValue = p;
		this.type = type;
    }

	/**
	 * It retrieves the value of the operand as a String.
	 */
	public String getValueAsString() {
		switch (type.getTypeCode()) {
		case QTypeTuple.TYPECODE_BOOLEAN:
			return "" + (longValue == 1);
		case QTypeTuple.TYPECODE_BYTE:
		case QTypeTuple.TYPECODE_SHORT:
		case QTypeTuple.TYPECODE_INT:
		case QTypeTuple.TYPECODE_LONG:
			return "" + longValue;
		case QTypeTuple.TYPECODE_CHAR:
			return new Character((char) longValue).toString();
		case QTypeTuple.TYPECODE_FLOAT:
		case QTypeTuple.TYPECODE_DOUBLE:
			return "" + doubleValue;
		default:
			return (objectValue==null ? null : objectValue.toString());
		}
	}

	public String toString() {
		return getValueAsString();
	}

    // IMPLEMENTATION OF THE Cloneable INTERFACE //
    //-------------------------------------------//
    public Object clone() throws CloneNotSupportedException {
        BasicOperand bo = (BasicOperand) super.clone();
        if (bo.objectValue != null && bo.objectValue instanceof Cloneable) {
            bo.objectValue = ((Cloneable) bo.objectValue).clone();
        }
        return bo;
    }
    // IMPLEMENTATION OF THE Expression INTERFACE //
    //--------------------------------------------//

    public void evaluate(ParameterOperand[] pos, Tuple t) throws MedorException{
    }

    public PType getType() {
        return type;
    }

    // IMPLEMENTATION OF THE Operand INTERFACE //
    //-----------------------------------------//

    public boolean isDefined() {
        return isDefined;
    }

    public void setIsDefined(boolean isdefined) {
        isDefined = isdefined;
    }

    public boolean getBoolean() throws TypingException {
        if (type.getTypeCode() == QType.TYPECODE_BOOLEAN) {
            return longValue == 1;
        } if (type.getTypeCode() == QType.TYPECODE_OBJBOOLEAN
			&& objectValue instanceof Boolean) {
            return ((Boolean) objectValue).booleanValue();
        } else
            throw new TypingException("Can not get a boolean value");
    }

    public int getInt() throws TypingException {
        if (type.getTypeCode() == QType.TYPECODE_INT) {
            return (int) longValue;
		} else if (type.getTypeCode() == QType.TYPECODE_OBJINT
			&& objectValue instanceof Integer) {
            return ((Integer) objectValue).intValue();
        } else
            throw new TypingException("Can not get an integer value " + type.getTypeCode());
    }

    public byte getByte() throws TypingException {
        if (type.getTypeCode() == QType.TYPECODE_BYTE) {
            return (byte) longValue;
		} else if (type.getTypeCode() == QType.TYPECODE_OBJBYTE
			&& objectValue instanceof Byte) {
            return ((Byte) objectValue).byteValue();
        } else
            throw new TypingException("Can not get a byte value");
    }

	public byte[] getByteArray() throws TypingException {
		 if (type.getTypeCode() == QType.TYPECODE_BYTEARRAY) {
			 return (byte[]) objectValue;
		 } else
			 throw new TypingException("Can not get a byteArray value");
	 }

     public short getShort() throws TypingException {
        if (type.getTypeCode() == QType.TYPECODE_SHORT) {
            return (short) longValue;
		} else if (type.getTypeCode() == QType.TYPECODE_OBJSHORT
			&& (objectValue == null || objectValue instanceof Short)) {
            return ((Short) objectValue).shortValue();
        } else
            throw new TypingException("Can not get a short value");
    }

    public long getLong() throws TypingException {
        if (type.getTypeCode() == QType.TYPECODE_LONG) {
            return longValue;
		} else if (type.getTypeCode() == QType.TYPECODE_OBJLONG
			&& (objectValue == null || objectValue instanceof Long)) {
            return ((Long) objectValue).longValue();
        } else
            throw new TypingException("Can not get a long value");
    }

    public float getFloat() throws TypingException {
        if (type.getTypeCode() == QType.TYPECODE_FLOAT) {
            return (float) doubleValue;
		} else if (type.getTypeCode() == QType.TYPECODE_OBJFLOAT
			&& (objectValue == null || objectValue instanceof Float)) {
            return ((Float) objectValue).floatValue();
        } else
            throw new TypingException("Can not get a float value");
    }

    public double getDouble() throws TypingException {
        if (type.getTypeCode() == QType.TYPECODE_DOUBLE) {
            return doubleValue;
		} else if (type.getTypeCode() == QType.TYPECODE_OBJDOUBLE
			&& (objectValue == null || objectValue instanceof Double)) {
            return ((Double) objectValue).doubleValue();
        } else
            throw new TypingException("Can not get a double value " + type.getJavaName());
    }

    public char getChar() throws TypingException {
        if (type.getTypeCode() == QType.TYPECODE_CHAR) {
            return (char) longValue;
		} else if (type.getTypeCode() == QType.TYPECODE_OBJCHAR
			&& (objectValue == null || objectValue instanceof Character)) {
            return ((Character) objectValue).charValue();
        } else
            throw new TypingException("Can not get a char value");
    }

	public char[] getCharArray() throws TypingException {
		if (type.getTypeCode() == QType.TYPECODE_CHARARRAY) {
			return (char[]) objectValue;
		}
		else
			throw new TypingException("Can not get a char value");
	}

      public String getString() throws TypingException {
        if (type.getTypeCode() == QType.TYPECODE_STRING
			&& (objectValue == null || objectValue instanceof String)) {
            return (String) objectValue;
        } else
            throw new TypingException("Can not get a string value: " + objectValue);
    }

	public BigDecimal getBigDecimal() throws TypingException {
		if (type.getTypeCode() == QType.TYPECODE_BIGDECIMAL
			&& (objectValue == null || objectValue instanceof BigDecimal)) {
			return (BigDecimal) objectValue;
		} else
			throw new TypingException("Can not get a BigDecimal value: " + objectValue);
	}

    public Date getDate() throws TypingException {
        // We will use the cast method. It will be changed soon
        if (objectValue == null || objectValue instanceof Date) {
            return (Date) objectValue;
        } else
            throw new TypingException("Can not get a Date value: "
                + (objectValue == null ? "null" : objectValue.getClass().getName()));
    }

    public Time getTime() throws TypingException {
        // We will use the cast method. It will be changed soon
        if (objectValue == null || objectValue instanceof Time) {
            return (Time) objectValue;
        } else
            throw new TypingException("Can not get a Time value: " + objectValue);
    }

    public Timestamp getTimestamp() throws TypingException {
        // We will use the cast method. It will be changed soon
        if (objectValue == null || objectValue instanceof Timestamp) {
            return (Timestamp) objectValue;
        }
        throw new TypingException("Can not get a Timestamp value: " + objectValue);
    }

    public Tuple getTuple() throws TypingException {
        // We will use the cast method. It will be changed soon
        if (objectValue == null || objectValue instanceof Tuple) {
            return (Tuple) objectValue;
        }
        throw new TypingException("Can not get a Tuple value");
    }

    public TupleCollection getTupleCollection() throws TypingException {
        // We will use the cast method. It will be changed soon
        if (objectValue == null || objectValue instanceof TupleCollection) {
            return (TupleCollection) objectValue;
        }
        throw new TypingException("Can not get a TupleCollection value");
    }

    public Object getObject() {
        if (type == null)
            return objectValue;
		switch (type.getTypeCode()) {
		case QTypeTuple.TYPECODE_BOOLEAN:
			return new Boolean(longValue == 1);
		case QTypeTuple.TYPECODE_BYTE:
			return new Byte((byte) longValue);
		case QTypeTuple.TYPECODE_CHAR:
			return new Character((char) longValue);
		case QTypeTuple.TYPECODE_DOUBLE:
			return new Double((double) longValue);
		case QTypeTuple.TYPECODE_FLOAT:
			return new Float((float) longValue);
		case QTypeTuple.TYPECODE_INT:
			return new Integer((int) longValue);
		case QTypeTuple.TYPECODE_LONG:
			return new Long(longValue);
		case QTypeTuple.TYPECODE_SHORT:
			return new Short((short) longValue);
		default:
			return objectValue;
		}
    }

    public void compileExpression() {
    }

    public boolean isCompiled(){
        return true;
    }

    public Operand getResult(){
        return this;
    }
}
