/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.filter.lib;

import org.objectweb.medor.filter.api.ThreeArgsOperator;
import org.objectweb.medor.filter.api.Expression;
import org.objectweb.medor.filter.api.ParameterOperand;
import org.objectweb.medor.filter.api.Operator;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TypingException;
import org.objectweb.medor.api.MalformedExpressionException;
import org.objectweb.medor.type.lib.QTypeTuple;
import org.objectweb.medor.type.lib.PTypeSpaceMedor;

/**
 * Operator representing the search for a substring within a given input string,
 * starting at a given index.
 */
public class IndexedLocate extends BasicThreeArgsOperator
	implements ThreeArgsOperator {

	/**
	 * Constructs an IndexedLocate operator, representing the search for a
	 * substring in an input string, starting at a given index.
	 * <p>It is the equivalent of the Java indexOf String manipulation method.
	 * <p>If the substring argument occurs as a substring within the
	 * inputString at a starting index no smaller than
	 * <code>fromIndex</code>, then the index of the first character
	 * of the first such substring is returned. If it does not occur
	 * as a substring starting at <code>fromIndex</code> or beyond,
	 * <code>-1</code> is returned.
	 * <p>Unlike in Java, the index starts with 1 and not 0.
	 * @param   inputString the String in which to look.
	 * @param   substring         the substring to search for.
	 * @param   fromIndex   the index to start the search from.

	 */
	public IndexedLocate(Expression inputString, Expression substring,
						 Expression fromIndex) {
		super(inputString, substring, fromIndex);
	}

	public IndexedLocate() {
		super();
	}

	public void evaluate(ParameterOperand[] pos, Tuple t)
		throws MedorException, IllegalStateException {
		try {
			first.evaluate(pos, t);
			second.evaluate(pos, t);
			third.evaluate(pos, t);
			result.setValue(evaluate(firstResult.getString(),
				secondResult.getString(), thirdResult.getInt()));
		}
		catch (NullPointerException e) {
			// result == null or missing children objects
			throw new IllegalStateException("Unevaluable Expression: May not be compiled");
		}
	}

	public void compileExpression()
		throws TypingException, MalformedExpressionException {
		if ((first != null) && (second != null) && (third != null)) {
			first.compileExpression();
			second.compileExpression();
			third.compileExpression();
			if ((first.getType().getTypeCode() ==
				QTypeTuple.TYPECODE_STRING) &&
				(second.getType().getTypeCode() ==
				QTypeTuple.TYPECODE_STRING) &&
				(third.getType().isa(PTypeSpaceMedor.INT))) {
				firstResult = first.getResult();
				secondResult = second.getResult();
				thirdResult = third.getResult();
				type = PTypeSpaceMedor.INT;
				result = new BasicVariableOperand(type);
				verified = true;
			}
			else
			//it is not a boolean type
				throw new TypingException("Attempted argument types: String, String, int");
		}
		else
		//Operands not yet assigned
			throw new MalformedExpressionException("Null children value");
	}

	public int evaluate(String inputString, String subString, int fromIndex) {
		return inputString.indexOf(subString, fromIndex - 1);
	}

	public String getOperatorString() {
		return Operator.INDEXEDLOCATE;
	}
}