/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 * 
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.filter.postfix;

import org.objectweb.jorm.type.api.PType;
import org.objectweb.medor.api.ExpressionException;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TypingException;
import org.objectweb.medor.filter.api.BelongOperator;
import org.objectweb.medor.filter.api.BinaryArithmeticOperator;
import org.objectweb.medor.filter.api.BinaryLogicalOperator;
import org.objectweb.medor.filter.api.BinaryOperator;
import org.objectweb.medor.filter.api.Comparator;
import org.objectweb.medor.filter.api.Expression;
import org.objectweb.medor.filter.api.ExpressionHelper;
import org.objectweb.medor.filter.api.Operand;
import org.objectweb.medor.filter.api.Operator;
import org.objectweb.medor.filter.api.ParameterOperand;
import org.objectweb.medor.filter.api.ThreeArgsOperator;
import org.objectweb.medor.filter.api.UnaryArithmeticOperator;
import org.objectweb.medor.filter.api.UnaryLogicalOperator;
import org.objectweb.medor.filter.api.UnaryOperator;
import org.objectweb.medor.filter.api.VariableOperand;
import org.objectweb.medor.filter.lib.FirstLocate;
import org.objectweb.medor.filter.lib.IndexedLocate;
import org.objectweb.medor.filter.lib.Length;
import org.objectweb.medor.filter.lib.Not;
import org.objectweb.medor.filter.lib.Sqrt;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.type.lib.QTypeTuple;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.Stack;

/**
 *
 * An implmentation of the ExpressionHeleper Interface
 */
public class PostfixExpressionHelper implements ExpressionHelper {

    int MAX_OPERAND_ON_OPERATOR = 3;

    public Expression[] toStack(Expression e)
            throws ExpressionException, TypingException {
        return (Expression[])
                toStack(e, new ArrayList()).toArray(new Expression[0]);
    }

    public Expression join(Expression[] stack)
            throws ExpressionException, TypingException {

        Expression[] temp = new Expression[MAX_OPERAND_ON_OPERATOR];
        int operandIdx = 0;
        for (int i = stack.length; i >= 0; i--) {
            if (stack[i] instanceof Operand) {
                if (operandIdx == MAX_OPERAND_ON_OPERATOR) {
                    throw new ExpressionException(
                            "Malformed stack: too many operand:" + (operandIdx + 1));
                }
                temp[operandIdx] = stack[i];
                operandIdx++;
            } else if (stack[i] instanceof Operator) {
				Operator op = (Operator) stack[i];
				if (operandIdx != op.getOperandNumber()) {
                    throw new ExpressionException(
                            "Malformed stack: wrong operand number:" + stack[i]);
                }
				for(int j=0; i<op.getOperandNumber(); i++) {
					op.setExpression(j, temp[j]);
				}
				temp[0] = op;
                operandIdx = 1;
            } else {
                throw new ExpressionException(
                        "Malformed stack: Unknown expression element: " + stack[i]);
            }
        }
        return temp[0];
    }

    public Expression substituteParameters(ParameterOperand[] pos, Expression e)
            throws ExpressionException, TypingException {

        if (e instanceof ParameterOperand) {
            ParameterOperand p = (ParameterOperand) e;
            int j = 0;
            while (j < pos.length) {
                if (pos[j].getName().equalsIgnoreCase(p.getName())) {
                    return pos[j];
                }
                j++;
            }
            throw new ExpressionException("Parameter not defined");
        } else if (e instanceof Operator) {
			Operator op = (Operator) e;
			for(int i=0; i<op.getOperandNumber(); i++) {
				op.setExpression(i, substituteParameters(pos, op.getExpression(i)));
			}
			return op;
		}
        return e;
    }

    public Expression[] substituteParameters(ParameterOperand[] pos,
                                             Expression[] stack)
            throws ExpressionException, TypingException {

        for (int i = 0; (i < stack.length); i++) {
            if (stack[i] instanceof ParameterOperand) {
                ParameterOperand p = (ParameterOperand) stack[i];
                int j = 0;
                while (j < pos.length) {
                    if (pos[j].getName().equalsIgnoreCase(p.getName())) {
                        stack[i] = pos[j];
                        break;
                    }
                    j++;
                }
                if (j == pos.length) {
                    throw new ExpressionException("Parameter not defined");
                }
            }
        }
        return stack;
    }

    public Object clone() throws CloneNotSupportedException {
        return super.clone();
    }

    public void fixIndexes(QueryNode node, Expression e) {
    }

    public Operand evaluate(Expression[] stack, ParameterOperand[] parameters, Tuple t)
            throws MedorException, IllegalStateException {
        try {
            Stack expressionResult = new Stack();
            int cpt = 0,cptResult = 0;
            boolean result;
            PType type;
            Operator bop;
            Comparator c;
            BinaryLogicalOperator bLogicop;
            BinaryArithmeticOperator bAritOp;
            Not notop;
            Sqrt sqrt;
            Length length;
            UnaryArithmeticOperator uAritOp;
            Operand ve1,ve2,ve3;
            // the current buffer wich contains result of the expression evaluation
            VariableOperand vop;
            while (cpt < stack.length) {
                if (stack[cpt] instanceof Operand) {
                    // it sets parameters value and FieldOperand value
                    stack[cpt].evaluate(parameters, t);
                    expressionResult.push(stack[cpt]);
                } else if (stack[cpt] instanceof Operator) {
                    // we have VariableOperand in the current implementation
                    // this operand is created when compiling
                    vop = (VariableOperand) stack[cpt].getResult();
                    if (stack[cpt] instanceof ThreeArgsOperator) {
                        ve3 = (Operand) expressionResult.pop();
                        ve2 = (Operand) expressionResult.pop();
                        ve1 = (Operand) expressionResult.pop();
                        if (stack[cpt] instanceof IndexedLocate) {
                            IndexedLocate locate = (IndexedLocate) stack[cpt];
                            vop.setValue(locate.evaluate(ve1.getString(), ve2.getString(), ve2.getInt()));
                        } else
                            throw new MedorException("Uknowen Expression element");
                    } else if (stack[cpt] instanceof BinaryOperator) {
                        ve2 = (Operand) expressionResult.pop();
                        ve1 = (Operand) expressionResult.pop();
                        if (stack[cpt] instanceof Comparator) {
                            c = (Comparator) stack[cpt];
                            type = stack[cpt].getType();
                            switch (type.getTypeCode()) {
                                case QTypeTuple.TYPECODE_INT:
                                case QTypeTuple.TYPECODE_OBJINT:
                                    vop.setValue(c.evaluate(ve1.getInt(), ve2.getInt()));
                                    break;
                                case QTypeTuple.TYPECODE_SHORT:
                                case QTypeTuple.TYPECODE_OBJSHORT:
                                    vop.setValue(c.evaluate(ve1.getShort(), ve2.getShort()));
                                    break;
                                case QTypeTuple.TYPECODE_BYTE:
                                case QTypeTuple.TYPECODE_OBJBYTE:
                                    vop.setValue(c.evaluate(ve1.getByte(), ve2.getByte()));
                                    break;
                                case QTypeTuple.TYPECODE_LONG:
                                case QTypeTuple.TYPECODE_OBJLONG:
                                    vop.setValue(c.evaluate(ve1.getLong(), ve2.getLong()));
                                    break;
                                case QTypeTuple.TYPECODE_DOUBLE:
                                case QTypeTuple.TYPECODE_OBJDOUBLE:
                                    vop.setValue(c.evaluate(ve1.getDouble(), ve2.getDouble()));
                                    break;
                                case QTypeTuple.TYPECODE_FLOAT:
                                case QTypeTuple.TYPECODE_OBJFLOAT:
                                    vop.setValue(c.evaluate(ve1.getFloat(), ve2.getFloat()));
                                    break;
                                case QTypeTuple.TYPECODE_STRING:
                                    vop.setValue(c.evaluate(ve1.getString(), ve2.getString()));
                                    break;
                                case QTypeTuple.TYPECODE_CHAR:
                                case QTypeTuple.TYPECODE_OBJCHAR:
                                    vop.setValue(c.evaluate(ve1.getChar(), ve2.getChar()));
                                    break;
                                case QTypeTuple.TYPECODE_DATE:
                                    vop.setValue(c.evaluate(ve1.getDate(), ve2.getDate()));
                                    break;
/*case QTypeTuple.TYPECODE_TIMESTAMP:
                                        vop.setValue(c.evaluate(ve1.getTimestamp(), ve2.getTimestamp())));
                                        break;
                                    case QTypeTuple.TYPECODE_TIME:
                                        vop.setValue(c.evaluate(ve1.getTime(), ve2.getTime())));
                                        break;
                                    case QTypeTuple.TYPECODE_TUPLECOLLECTION:
                                        vop.setValue(c.evaluate(ve1.getTupleCollection(),
                                                ve2.getTupleCollection())));
                                        break;

                                    case QTypeTuple.TYPECODE_TUPLE:
                                        vop.setValue(c.evaluate(ve1.getTuple(), ve2.getTuple())));
                                        break;*/
                            }
                        } else if (stack[cpt] instanceof BinaryLogicalOperator) {
                            bLogicop = (BinaryLogicalOperator) stack[cpt];
                            result = bLogicop.evaluate(ve1.getBoolean(),
                                    ve2.getBoolean());
                            vop.setValue(result);

                        } else if (stack[cpt] instanceof BinaryArithmeticOperator) {
                            bAritOp = (BinaryArithmeticOperator) stack[cpt];
                            //Gets the result buffer operand
                            //this operand is created wehen compiling expression
                            //We never create objects when evaluating
                            type = bAritOp.getType();
                            switch (type.getTypeCode()) {
                                case QTypeTuple.TYPECODE_INT:
                                case QTypeTuple.TYPECODE_OBJINT:
                                    vop.setValue(bAritOp.evaluate(ve1.getInt(), ve2.getInt()));
                                    break;
                                case QTypeTuple.TYPECODE_SHORT:
                                case QTypeTuple.TYPECODE_OBJSHORT:
                                    vop.setValue(bAritOp.evaluate(ve1.getShort(), ve2.getShort()));
                                    break;
                                case QTypeTuple.TYPECODE_BYTE:
                                case QTypeTuple.TYPECODE_OBJBYTE:
                                    vop.setValue(bAritOp.evaluate(ve1.getByte(), ve2.getByte()));
                                    break;
                                case QTypeTuple.TYPECODE_LONG:
                                case QTypeTuple.TYPECODE_OBJLONG:
                                    vop.setValue(bAritOp.evaluate(ve1.getLong(), ve2.getLong()));
                                    break;
                                case QTypeTuple.TYPECODE_DOUBLE:
                                case QTypeTuple.TYPECODE_OBJDOUBLE:
                                    vop.setValue(bAritOp.evaluate(ve1.getDouble(), ve2.getDouble()));
                                    break;
                                case QTypeTuple.TYPECODE_FLOAT:
                                case QTypeTuple.TYPECODE_OBJFLOAT:
                                    vop.setValue(bAritOp.evaluate(ve1.getFloat(), ve2.getFloat()));
                                    break;
                                case QTypeTuple.TYPECODE_STRING:
                                    vop.setValue(bAritOp.evaluate(ve1.getString(), ve2.getString()));
                                    break;
                                case QTypeTuple.TYPECODE_CHAR:
                                case QTypeTuple.TYPECODE_OBJCHAR:
                                    vop.setValue(bAritOp.evaluate(ve1.getChar(), ve2.getChar()));
                                    break;
                                case QTypeTuple.TYPECODE_DATE:
                                    vop.setValue(bAritOp.evaluate(ve1.getDate(), ve2.getDate()));
                                    break;
                            }
                        } else if (stack[cpt] instanceof BelongOperator) {
                            BelongOperator in = (BelongOperator) stack[cpt];
                            vop.setValue(in.evaluate(ve1.getTuple(), ve2.getTupleCollection()));
                        } else if (stack[cpt] instanceof FirstLocate) {
                            FirstLocate locate = (FirstLocate) stack[cpt];
                            vop.setValue(locate.evaluate(ve1.getString(), ve2.getString()));
                        } else
                            throw new MedorException("Uknowen Expression element");
                        cpt++;
                    } else if (stack[cpt] instanceof UnaryOperator) {
                        ve1 = (Operand) expressionResult.pop();
                        if (stack[cpt] instanceof UnaryArithmeticOperator) {
                            uAritOp = (UnaryArithmeticOperator) stack[cpt];
                            type = uAritOp.getType();
                            switch (type.getTypeCode()) {
                                case QTypeTuple.TYPECODE_INT:
                                case QTypeTuple.TYPECODE_OBJINT:
                                    vop.setValue(uAritOp.evaluate(ve1.getInt()));
                                    break;
                                case QTypeTuple.TYPECODE_SHORT:
                                case QTypeTuple.TYPECODE_OBJSHORT:
                                    vop.setValue(uAritOp.evaluate(ve1.getShort()));
                                    break;
                                case QTypeTuple.TYPECODE_BYTE:
                                case QTypeTuple.TYPECODE_OBJBYTE:
                                    vop.setValue(uAritOp.evaluate(ve1.getByte()));
                                    break;
                                case QTypeTuple.TYPECODE_LONG:
                                case QTypeTuple.TYPECODE_OBJLONG:
                                    vop.setValue(uAritOp.evaluate(ve1.getLong()));
                                    break;
                                case QTypeTuple.TYPECODE_DOUBLE:
                                case QTypeTuple.TYPECODE_OBJDOUBLE:
                                    vop.setValue(uAritOp.evaluate(ve1.getDouble()));
                                    break;
                                case QTypeTuple.TYPECODE_FLOAT:
                                case QTypeTuple.TYPECODE_OBJFLOAT:
                                    vop.setValue(uAritOp.evaluate(ve1.getFloat()));
                                    break;
                            }
                        } else if (stack[cpt] instanceof UnaryLogicalOperator) {
                            notop = (Not) stack[cpt];
                            vop.setValue(notop.evaluate(ve1.getBoolean()));
                        } else if (stack[cpt] instanceof Sqrt) {
                            sqrt = (Sqrt) stack[cpt];
                            vop.setValue(sqrt.evaluate(ve1.getDouble()));
                        } else if (stack[cpt] instanceof Length) {
                            length = (Length) stack[cpt];
                            vop.setValue(length.evaluate(ve1.getString()));
                        } else
                            throw new MedorException("Uknowen Expression element");
                    } else
                        throw new MedorException("Uknowen Expression element");
                    expressionResult.push(vop);
                }
                cpt++;
            }
            return (Operand) expressionResult.pop();
        } catch (NullPointerException e) {
            throw new IllegalStateException("Expression Error: may be not Compiled");
        }
    }

    public ArrayList toStack(Expression e, ArrayList al)
            throws ExpressionException, TypingException {
        if (e instanceof Operator) {
			for(int i=0; i< ((Operator) e).getOperandNumber(); i++) {
				toStack(((Operator) e).getExpression(i), al);
			}
        } else if (!(e instanceof Operand)) {
            throw new ExpressionException(
                    "Malformed stack: Unknown expression element: " + e);
        }
        al.add(e);
        return al;
    }

    public int getNumberOfNode(Expression e) throws ExpressionException {
        if (e instanceof Operand) {
            return 1;
        } else if (e instanceof Operator) {
			int result = 1;
			for (int i= 0; i< ((Operator) e).getOperandNumber(); i++)
				result += getNumberOfNode(((Operator) e).getExpression(i));
			return result;
        } else {
            throw new ExpressionException(
                    "Malformed stack: Unknown expression element: " + e);
        }
    }

    /**
     * Checks no branch of the expression is null.
     */
    public boolean verifyNoNull(Expression e) {

        if (e instanceof Operator) {
			boolean res = true;
			for (int i=0 ; i<((Operator) e).getOperandNumber() && res; i++) {
				res = verifyNoNull(((Operator) e).getExpression(i));
			}
			return res;
        } else {
            return e != null;
        }
    }

    public String[] getParameterNames(Expression

            e) throws ExpressionException, TypingException {
        LinkedList nameList = new LinkedList();
        Expression[] stack = this.toStack(e);
        for (int i = 0; (i < stack.length); i++) {
            if (stack[i] instanceof ParameterOperand) {
                nameList.add(((ParameterOperand) stack[i]).getName());
            }
        }
        return (String[]) nameList.toArray(new String[nameList.size()]);
    }
}
