/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */
package org.objectweb.medor.lib;

import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.util.monolog.wrapper.printwriter.LoggerImpl;
import org.objectweb.util.monolog.api.HandlerFactory;
import org.objectweb.util.monolog.api.LevelFactory;
import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.file.monolog.PropertiesConfAccess;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.FileNotFoundException;
import java.util.Properties;

/**
 * This class references the logger factory used in Medor. A default logger
 * factory is instanciated. It is based on a printwriter implementation which by
 * default logs anything.
 *
 * @author S.Chassande-Barrioz
 */
public class Log
        implements LoggerFactory {

	public static final boolean DEBUG = false;

	public static final String MEDOR_PREFIX = "org.objectweb.medor";

	public static LoggerFactory loggerFactory = new LoggerImpl();

    /**
     * Default name of the logger factory
     */
    public final static String DEFAULT_LOGGER_FACTORY =
            "org.objectweb.util.monolog.wrapper.printwriter.LoggerImpl";

    public static LoggerFactory getLoggerFactory(String lfClassName)
            throws Exception {
        if (lfClassName == null)
            return (LoggerFactory)
                    Class.forName(DEFAULT_LOGGER_FACTORY).newInstance();
        else
            return (LoggerFactory) Class.forName(lfClassName).newInstance();
    }

    /**
     * Gets a new <code>LoggerFactory</code>
     *
     * @param propFileName properties containing <code>logger.factory</code>
	 * and <code>log.conf.file</code>
     * @return the build LoggerFactory
     */
    public static LoggerFactory loadLoggerFactory(String propFileName)
            throws Exception {
        if (propFileName == null)
            return (LoggerFactory)
                    Class.forName(DEFAULT_LOGGER_FACTORY).newInstance();
        // lookup the properties file in the file system
        File f = new File(propFileName);
        if (f.exists()) {
            Properties p = new Properties();
            p.load(new FileInputStream(propFileName));
            return loadLoggerFactory(p);
        }
        //lookup the properties file in the classpath
        InputStream is = ClassLoader.getSystemResourceAsStream(propFileName);
        if (is != null) {
            Properties p = new Properties();
            p.load(is);
            return loadLoggerFactory(p);
        }
        throw new FileNotFoundException(propFileName + " was not found");
    }

    /**
     * Gets a new <code>LoggerFactory</code>
     *
     * @param prop properties containing <code>logger.factory</code> and <code>log.conf.file</code>
     * @return the build LoggerFactory
     */
    public static LoggerFactory loadLoggerFactory(Properties prop) throws Exception {
        if (prop == null)
            return (LoggerFactory)
                    Class.forName(DEFAULT_LOGGER_FACTORY).newInstance();

        // Creates the factory if needed
        String lfn = prop.getProperty("log.config.classname", DEFAULT_LOGGER_FACTORY);
        LoggerFactory factory = (LoggerFactory) Class.forName(lfn).newInstance();

        if (factory instanceof HandlerFactory && factory instanceof LevelFactory) {
            // Configures the factory
            PropertiesConfAccess.load(prop,
                    factory,
                    (HandlerFactory) factory,
                    (LevelFactory) factory);
        }
        return factory;
    }

    public static LoggerFactory getLoggerFactory() {
        if (lastLF == null)
            try {
                lastLF = loadLoggerFactory((String) null);
            } catch (Exception e) {
            }
        return lastLF;
    }

    private static LoggerFactory lastLF = null;

    private LoggerFactory delegate = null;
    private String pfn = null;

    // IMPLEMENTATION OF THE LogAttributes INTERFACE //
    //-----------------------------------------------//

    public String getPropertiesFileName() {
        return pfn;
    }

    public void setPropertiesFileName(String propfn) {
        if (pfn != null && pfn.equals(propfn))
            return;
        pfn = propfn;
        try {
            delegate = Log.loadLoggerFactory(pfn);
        } catch (Exception e) {
            e.printStackTrace();
            try {
                delegate = Log.loadLoggerFactory((String) null);
            } catch (Exception e1) {
                e1.printStackTrace();
            }
        }
        lastLF = this;
    }

    // IMPLEMENTATION OF THE LoggerFactory INTERFACE //
    //-----------------------------------------------//

    public Logger getLogger(String s) {
        return delegate.getLogger(s);
    }

    public Logger getLogger(String s, String s1) {
        return delegate.getLogger(s1);
    }

    public String getResourceBundleName() {
        return delegate.getResourceBundleName();
    }

    public void setResourceBundleName(String s) {
        delegate.setResourceBundleName(s);
    }

    public Logger[] getLoggers() {
        return delegate.getLoggers();
    }
}
