/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre, S. Chassande-Barrioz
 */
package org.objectweb.medor.optim.jorm;

import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.metainfo.api.NameDef;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;
import org.objectweb.jorm.metainfo.rdb.api.RdbClassMultiMapping;
import org.objectweb.jorm.metainfo.rdb.api.RdbExternalTable;
import org.objectweb.jorm.metainfo.rdb.api.RdbGenClassMapping;
import org.objectweb.jorm.metainfo.rdb.api.RdbJoin;
import org.objectweb.jorm.metainfo.rdb.api.RdbPrimitiveElementMapping;
import org.objectweb.jorm.metainfo.rdb.api.RdbTable;
import org.objectweb.jorm.runtime.mapper.rdb.adapter.api.JoinedTable;
import org.objectweb.jorm.runtime.mapper.rdb.adapter.lib.RdbAdapterFactory;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MalformedExpressionException;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.datasource.api.DataStore;
import org.objectweb.medor.datasource.lib.ConnectionFactoryDataStore;
import org.objectweb.medor.filter.api.BinaryLogicalOperator;
import org.objectweb.medor.filter.api.Comparator;
import org.objectweb.medor.filter.api.Expression;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.filter.api.Operand;
import org.objectweb.medor.filter.api.Operator;
import org.objectweb.medor.filter.api.ParameterOperand;
import org.objectweb.medor.filter.api.VariableOperand;
import org.objectweb.medor.filter.jorm.lib.CompositePName;
import org.objectweb.medor.filter.jorm.lib.EncodePName;
import org.objectweb.medor.filter.jorm.lib.SinglePName;
import org.objectweb.medor.filter.lib.And;
import org.objectweb.medor.filter.lib.BasicFieldOperand;
import org.objectweb.medor.filter.lib.Equal;
import org.objectweb.medor.filter.lib.ExpressionPrinter;
import org.objectweb.medor.filter.lib.MemberOf;
import org.objectweb.medor.filter.lib.Not;
import org.objectweb.medor.filter.lib.NotEqual;
import org.objectweb.medor.filter.lib.Or;
import org.objectweb.medor.filter.lib.BasicVariableOperand;
import org.objectweb.medor.filter.lib.IsNull;
import org.objectweb.medor.query.api.CalculatedField;
import org.objectweb.medor.query.api.OrderField;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.jorm.api.JormExtent;
import org.objectweb.medor.query.jorm.lib.PNameField;
import org.objectweb.medor.query.jorm.lib.GenClassExtent;
import org.objectweb.medor.query.lib.BasicCalculatedField;
import org.objectweb.medor.query.lib.BasicOrderField;
import org.objectweb.medor.query.lib.QueryTreePrinter;
import org.objectweb.medor.query.lib.SelectProject;
import org.objectweb.medor.query.rdb.api.QualifiedTable;
import org.objectweb.medor.query.rdb.api.RdbExpField;
import org.objectweb.medor.query.rdb.api.RdbExpQueryLeaf;
import org.objectweb.medor.query.rdb.api.RdbField;
import org.objectweb.medor.query.rdb.lib.BasicQualifiedTable;
import org.objectweb.medor.query.rdb.lib.BasicRdbExpQueryLeaf;
import org.objectweb.medor.optim.lib.PushNotInExpressionRule;
import org.objectweb.medor.optim.api.RewriteRule;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

/**
 * @author S.Chassande-Barrioz
 */
public class JormFlatten2Rdb extends JormRule {

	public JormFlatten2Rdb() {
		super("JormFlatten2Rdb");
	}

	/**
	 * element = JormExtent
	 */
	private List extents = new ArrayList();

	/**
	 * element = RdbJormExtentMapping
	 */
	private List mappings = new ArrayList();

	/**
	 * element = BasicRdbExpQueryLeaf
	 */
	private List subLeaves = new ArrayList();

	/**
	 * This field is used as counter for calculating the table alias name.
	 */
	private int alias = 0;

	private List joinedTables = new ArrayList();

	/**
	 * These two fields are used to map new MemberOf operators (created from the
	 * existing ones) with the extent of
	 * the fields on the right hand side of the MemberOf
	 */
	private List mofs = new ArrayList();
	private List mofExtents = new ArrayList();

	public QueryTree rewrite(QueryTree qt) throws MedorException {
		log.log(BasicLevel.DEBUG, "JormFlatten2Rdb: Input query tree:");
		QueryTreePrinter.printQueryTree(qt, log, BasicLevel.DEBUG);
		extents.clear();
		mappings.clear();
		alias = 0;
		joinedTables.clear();
		mofs.clear();
		mofExtents.clear();
		subLeaves.clear();
		if (qt instanceof JormExtent) {
			SelectProject p = new SelectProject("", null);
			Field[] fs = qt.getTupleStructure().getFields();
			for (int i = 0; i < fs.length; i++) {
				p.addPropagatedField(fs[i].getName(), fs[i].getType(),
						new QueryTreeField[]{(QueryTreeField) fs[i]});
			}
			return rewrite(p);
		}
		QueryNode qn = (QueryNode) qt;
		QueryTree[] children = qn.getChildren();
		for (int i = 0; i < children.length; i++) {
			RdbJormExtentMapping rjem =
					new RdbJormExtentMapping((JormExtent) children[i]);
			extents.add(children[i]);
			mappings.add(rjem);
		}
		Expression filter;
		boolean assignfilter = false;
		try {
			filter = qn.getQueryFilter();
			assignfilter = (filter != null);
		}
		catch (Exception e) {
			filter = null;
		}
		if (debug) {
			log.log(BasicLevel.DEBUG, "The initial structure values");
			for (int i = 0; i < mappings.size(); i++) {
				log.log(BasicLevel.DEBUG, "Extent " + ((RdbJormExtentMapping)mappings.get(i)).extent.getName());
				log.log(BasicLevel.DEBUG, "Extent tables size " + ((RdbJormExtentMapping)mappings.get(i)).tables.size());
				QualifiedTable quali = ((QualifiedTable)((RdbJormExtentMapping)mappings.get(i)).tables.get(0));
				log.log(BasicLevel.DEBUG, "Extent table 0 " + quali.getTableName());
				log.log(BasicLevel.DEBUG, "Extent table alias 0 " + quali.getAliasName());
				log.log(BasicLevel.DEBUG, "Join size " + ((RdbJormExtentMapping)mappings.get(i)).joins.size());
				//log.log(BasicLevel.DEBUG, "Join 0 name " + ((Join)((RdbJormExtentMapping)mappings.get(i)).joins.get(0)).name);
				//log.log(BasicLevel.DEBUG, "Join 0 columns " + ((Join)((RdbJormExtentMapping)mappings.get(i)).joins.get(0)).columns.length);

				log.log(BasicLevel.DEBUG, mappings.get(i));
			}
			//log.log(BasicLevel.DEBUG, "filter: " + ExpressionPrinter.e2str(filter));
			log.log(BasicLevel.DEBUG, "Removing useless pname equality");
		}
		filter = removeUseLessPNameEquality(filter);
		if (debug) {
			for (int i = 0; i < mappings.size(); i++) {
				log.log(BasicLevel.DEBUG, mappings.get(i));
			}
			log.log(BasicLevel.DEBUG, "filter: " + ExpressionPrinter.e2str(filter));
			log.log(BasicLevel.DEBUG, "Updating the projected fields");
		}

		// Create the rdb leaf with its datastore
		DataStore ds = ((JormExtent) children[0]).getDataStore();
		String databasename = ((JormExtent) children[0]).getPMapper().getMapperName();
		int idx = databasename.indexOf('.');
		if (idx != -1) {
			databasename = databasename.substring(idx + 1);
		}
		else {
			databasename = RdbAdapterFactory.DATABASE_NAME_JDBC;
		}
		if (ds == null) {
			Object cf = ((JormExtent) children[0]).getPMapper().getConnectionFactory();
			ds = new ConnectionFactoryDataStore(
					DataStore.JDBC_STORE,
					cf.toString(),
					new short[0],
					cf);
		}
		BasicRdbExpQueryLeaf leaf = new BasicRdbExpQueryLeaf(ds, "");
		leaf.setDistinct(qt.getDistinct());
		try {
			leaf.setRdbAdapter(RdbAdapterFactory.getTypeConverter(databasename));
		}
		catch (PException e) {
			throw new MedorException(
					"Impossible to assign the rdb adapter to the leaf", e);
		}
		Field[] fs = qt.getTupleStructure().getFields();
		for (int i = 0; i < fs.length; i++) {
			if (fs[i] instanceof PropagatedField) {
				QueryTreeField qtf = (QueryTreeField)
						((PropagatedField) fs[i]).getPreviousFields()[0];
				if (qtf instanceof PNameField) {
					CalculatedField cf = new BasicCalculatedField(
							qtf.getName(), qtf.getType(), qtf.getQueryTree(),
							getDecode((PNameField) qtf, leaf, true));
					qn.replace((QueryTreeField) fs[i], cf);
				}
				else {
					QueryTreeField f = getField(qtf, leaf, true);
					qn.updatePropagatedField(
							fs[i].getName(), new QueryTreeField[]{f});
				}
			}
			else if (fs[i] instanceof CalculatedField) {
				Expression e = ((CalculatedField) fs[i]).getExpression();
				e = updateFieldOfExpression(e, leaf, true);
				qn.updateCalculatedField(fs[i].getName(), e);
			}
			else {
				throw new MedorException(
						"Unmanaged field (" + fs[i].getName() + "): " + fs[i]);
			}
		}
		if (debug) {
			log.log(BasicLevel.DEBUG, "Updating the order by clause");
		}
		OrderField[] ofields = qt.getOrderBy();
		if (ofields != null) {
			for (int j = 0; j < ofields.length; j++) {
				QueryTreeField f = getField(ofields[j].getField(), leaf, false);
				ofields[j].setField(f);
			}
		}
		if (debug) {
			log.log(BasicLevel.DEBUG, "Updating the filter");
		}
		filter = updateFieldOfExpression(filter, leaf, false);
		if (debug) {
			log.log(BasicLevel.DEBUG, "filter: " + ExpressionPrinter.e2str(filter));
		}
		if (assignfilter) {
			qn.setQueryFilter(null);
		}
		leaf.setQueryFilter(filter);
		leaf.setRootJoinedTables(joinedTables);
		if (debug) {
			log.log(BasicLevel.DEBUG, "joinedTables: " + joinedTables);
			log.log(BasicLevel.DEBUG, "JormFlatten2Rdb: Query leaf expression:");
			log.log(BasicLevel.DEBUG, ExpressionPrinter.e2str(filter));
		}

        //Push the not in the expression before to split the filter due to the
        // member of.
        RewriteRule rule = new PushNotInExpressionRule();
        rule.rewrite(leaf);
        filter = leaf.getQueryFilter();

		splitLeafIfMemberOf(leaf, filter);
		//propagates the distinct
		leaf.setDistinct(qt.getDistinct());
		//propagates the order by to the leaf
		if (ofields != null) {
			OrderField[] leafOFields =
					new BasicOrderField[ofields.length];
			for (int j = 0; j < ofields.length; j++) {
				OrderField of = new BasicOrderField(
						ofields[j].getField(),
						ofields[j].getDesc()
				);
				leafOFields[j] = of;
			}
			leaf.setOrderBy(leafOFields);
		}
		leaf.setQueryFilter(updateFilterWithNewLeaves(filter, leaf));
		return qt;
	}

	/**
	 * Represents the mapping of a JormExtent over relational tables.
	 * <p>Only the
	 * main table and the external not collocated table are listed in a node.
	 */
	public class RdbJormExtentMapping {

		/**
		 * The extent which is mapped.
		 */
		private JormExtent extent = null;

		/**
		 * List of tables inside which the extent is mapped.
		 * <p>This list contains
		 * only the main table and  the not collocated external tables.
		 * The first element of the list is the main table.
		 * element = QualifiedTable
		 */
		private List tables = new ArrayList(5);

		/**
		 * List of join to reach the table.
		 * <p>This list is linked to the tables
		 * list. The first element is always null because the main table does
		 * not need a join.
		 * element = Join
		 */
		private List joins = new ArrayList(5);

		public RdbJormExtentMapping(JormExtent je) throws MedorException {
			setJormExtent(je);
		}

		/**
		 * Assigns the JormExtent and calculates the fields 'tables'
		 * and 'joins'.
		 * @param je is a JormExtent instance (ClassExtent or GenClassExtent)
		 * @throws MedorException if the jorm meta information is malformed
		 */
		public void setJormExtent(JormExtent je) throws MedorException {
			extent = je;
			if (extent.getMetaObject() instanceof Class) {
				RdbClassMultiMapping cm =
						(RdbClassMultiMapping) getClassMapping(extent);
				RdbTable maintable = cm.getRdbTable();
				QualifiedTable mainqtable = new BasicQualifiedTable(
						maintable.getName(),
						getTableAlias(maintable, null));
				tables.add(mainqtable);
				joins.add(null);
				Collection etables = cm.getRdbExternalTables();
				Iterator tableit = etables.iterator();
				while (tableit.hasNext()) {
					RdbExternalTable etable = (RdbExternalTable) tableit.next();
					Collection rdbJjoins = etable.getRdbJoins();
					Iterator joinit = rdbJjoins.iterator();
					if (!joinit.hasNext()) {
						throw new MedorException(
								"No join defined for the external table "
								+ etable.getName() + " in the extent "
								+ je.getJormName());
					}
					while (joinit.hasNext()) {
						RdbJoin j = (RdbJoin) joinit.next();
						Join join = new Join(j, maintable, etable);
						joins.add(join);
						QualifiedTable qtable =
								new BasicQualifiedTable(etable.getName(),
										getTableAlias(etable, j));
						tables.add(qtable);
					}
				}
			}
			else if (extent.getMetaObject() instanceof GenClassRef) {
				RdbGenClassMapping gcm =
						(RdbGenClassMapping) getGenClassMapping(extent);
				RdbTable maintable = gcm.getRdbTable();
				QualifiedTable mainqtable = new BasicQualifiedTable(
						maintable.getName(),
						getTableAlias(maintable, null));
				tables.add(mainqtable);
				joins.add(null);
			}
		}

		public int getJoinIndex(String joinname) throws MedorException {
			for (int i = 1; i < joins.size(); i++) {
				if (joinname.equals(((Join) joins.get(i)).name)) {
					return i;
				}
			}
			throw new MedorException("Join " + joinname
					+ " not found in the extent of the class "
					+ extent.getJormName());
		}

		public String toString() {
			StringBuffer sb = new StringBuffer();
			sb.append("* ExtentMapping: ");
			sb.append(extent.getJormName());
			sb.append("\n");
			sb.append("Tables:\n");
			for (int i = 0; i < tables.size(); i++) {
				QualifiedTable qt = (QualifiedTable) tables.get(i);
				sb.append(qt.getAliasName());
				sb.append(" join: ");
				if (joins.get(i) != null) {
					sb.append(joins.get(i));
				}
				sb.append("\n");
			}
			return sb.toString();
		}
	}

	/**
	 * A join linking two relational tables.
	 */
	public class Join {

		/**
		 * the join name is used in the table alias in order to avoid name
		 * collision in case of multiple joins toward a same table.
		 */
		public String name = "";

		/**
		 * is a list of JoinColumn instance.
		 */
		public JoinColumn[] columns = null;

		/**
		 * Builds a Join between a main table and an external table.
		 * The Join is equal to the name of the jorm meta object describing the
		 * join.
		 * @param j is the jorm meta object describing the join.
		 * @param maintable is the jorm meta object describing the main table
		 * @param externaltable is the jorm meta object describing the
		 * external table.
		 */
		public Join(RdbJoin j, RdbTable maintable, RdbTable externaltable) {
			name = j.getName();
			List mcols = j.getPTJoinColumnNames();
			List ecols = j.getETJoinColumnNames();
			columns = new JoinColumn[ecols.size()];
			for (int i = 0; i < ecols.size(); i++) {
				JoinColumn jc = new JoinColumn();
				jc.column1 = (String) mcols.get(i);
				jc.column2 = (String) ecols.get(i);
				RdbPrimitiveElementMapping pem =
						(RdbPrimitiveElementMapping) maintable
						.getPrimitiveElementMappingByCol(jc.column1);
				PrimitiveElement pe = (PrimitiveElement) pem.getLinkedMO();
				jc.type = pe.getType();
				jc.field1 = pe.getName();
				pem = (RdbPrimitiveElementMapping) externaltable
						.getPrimitiveElementMappingByCol(jc.column2);
				if (pem != null) {
					pe = pem.lookupPrimitiveElement(j);
					jc.field2 = pe.getName();
				}
				columns[i] = jc;
			}
		}

		public String toString() {
			StringBuffer sb = new StringBuffer();
			sb.append(name);
			String sep = "{";
			for (int i = 0; i < columns.length; i++) {
				sb.append(sep);
				sb.append(columns[i]);
				sep = " && ";
			}
			sb.append('{');
			return sb.toString();
		}
	}

	/**
	 * It describes the join between two columns
	 */
	public class JoinColumn {

		/**
		 * Both fields have the same jorm type.
		 */
		public PType type = null;

		/**
		 * Name of the first jorm field
		 */
		public String field1 = null;

		/**
		 * Name of the first column name
		 */
		public String column1 = null;

		/**
		 * Name of the second jorm field
		 */
		public String field2 = null;

		/**
		 * Name of the second column name
		 */
		public String column2 = null;

		public String toString() {
			return "(" + column1 + " = " + column2 + ")";
		}
	}

	/**
	 * Retrieves the RdbJormExtentMapping at the position i the packet.
	 * @param i is position of the expected RdbJormExtentMapping
	 * @return a RdbJormExtentMapping (never null).
	 */
	public RdbJormExtentMapping getMapping(int i) {
		return (RdbJormExtentMapping) mappings.get(i);
	}

	/**
	 * Retrieves the RdbJormExtentMapping at the position i the packet.
	 * @param je is the jorm extent which the mapping is expected
	 * @return a RdbJormExtentMapping (never null).
	 */
	public RdbJormExtentMapping getMapping(JormExtent je) {
		int i = extents.indexOf(je);
		return (i == -1 ? null : (RdbJormExtentMapping) mappings.get(i));
	}

	/**
	 * Calculates the table alias.
	 * @param table is the tablt which the alias is expected.
	 * @param join is the join to use to reach the table. This field can be null
	 * if the table is the main.
	 * @return a string value which is the alias of the table (never null).
	 */
	private String getTableAlias(RdbTable table, RdbJoin join) {
		return (join == null
				? table.getName() + "_" + (alias++)
				: table.getName() + "_" + (alias++) + "_" + join.getName());
	}

	/**
	 * Merges the useless tables and removes the useless PName equalities. This
	 * method is recursive.
	 * @param e the current expression
	 * @return the new expression. If the value is null then that means the
	 * expression given in paremter is useless. if the value is different from
	 * the parameter then that means the expression has been modified and this
	 * is the new value.
	 * @throws MedorException if an arror occurs during the merging of table
	 * or if the expressin is malformed.
	 */
	private Expression removeUseLessPNameEquality(Expression e)
			throws MedorException {
		if (e == null) {
			return null;
		}
		if (e instanceof Equal) {
			Equal eq = (Equal) e;
			Expression left = eq.getExpression(0);
			Expression right = eq.getExpression(1);
			if (left instanceof FieldOperand && right instanceof FieldOperand) {
				FieldOperand fo1 = (FieldOperand) left;
				FieldOperand fo2 = (FieldOperand) right;
				QueryTreeField qtf1 = (QueryTreeField) fo1.getField();
				QueryTreeField qtf2 = (QueryTreeField) fo2.getField();
				if (qtf1 instanceof PNameField & qtf2 instanceof PNameField) {
					//This is a PName equality

					PNameField pnf1 = (PNameField) qtf1;
					JormExtent je1 = (JormExtent) pnf1.getQueryTree();
					log.log(BasicLevel.DEBUG, "JormExtent: " + je1 + " " + je1.getName());
					log.log(BasicLevel.DEBUG, "PNameField: " + pnf1.getName());
					NameDef nd1 = pnf1.getNameDef(je1);
					log.log(BasicLevel.DEBUG, "NameDef1: " + nd1);

					PNameField pnf2 = (PNameField) qtf2;
					JormExtent je2 = (JormExtent) pnf2.getQueryTree();
					NameDef nd2 = pnf2.getNameDef(je2);
					log.log(BasicLevel.DEBUG, "NameDef2: " + nd2);

					boolean sameTable = true;
					if (nd1.isFieldName() && nd2.isFieldName()) {
						//Replace if the column are the same
						sameTable = sameColumn(je1, nd1.getFieldName(),
								je2, nd2.getFieldName(), true);
					}
					else if (nd1.isNameRef() && nd2.isNameRef()) {
						if (!nd1.getNameRef().getCompositeName().getFQName()
								.equals(
										nd2.getNameRef().getCompositeName().getFQName())) {
							throw new MedorException("Bad PName equality: " +
									"two different composite name");
						}
						Map proj1 = nd1.getNameRef().getProjection();
						Map proj2 = nd2.getNameRef().getProjection();
						if (proj1.size() != proj2.size()) {
							throw new MedorException(
									"Bad PName equality: bad projection size");
						}
						Iterator it = proj1.entrySet().iterator();
						//Replace if all columns are the same
						while (it.hasNext() && sameTable) {
							Map.Entry me = (Map.Entry) it.next();
							String cofn = (String) me.getKey();
							String clfn1 = (String) me.getValue();
							String clfn2 = (String) proj2.get(cofn);
							sameTable = sameColumn(
									je1, clfn1, je2, clfn2, !it.hasNext());
						}
					}
					else {
						throw new MedorException("Bad PName equality: " +
								"PName structures unknwon or not same");
					}
					if (sameTable) {
						//remove this PName equality
						return null;
					}
				}
			}
		}
		else if (e instanceof BinaryLogicalOperator) {
			// some children of this node can be removed.
			Operator op = (Operator) e;
			Expression e0 = removeUseLessPNameEquality(op.getExpression(0));
			Expression e1 = removeUseLessPNameEquality(op.getExpression(1));
			if (e0 == null && e1 != null) {
				// e0 has been removed ==> return only the other child
				return e1;
			}
			else if (e0 != null && e1 == null) {
				// e1 has been removed ==> return only the other child
				return e0;
			}
			else if (e0 == null && e1 == null) {
				//Both children has been removed ==> remove the operator
				return null;
			}
			else {
				// No children has been removed but they could be changed
				if (e0 != null && e0 != op.getExpression(0)) {
					op.setExpression(0, e0);
				}
				if (e0 != null && e0 != op.getExpression(0)) {
					op.setExpression(0, e0);
				}
				return op;
			}
		}
		else if (e instanceof Operator) {
			// Manage operator which are not logical operators.
			// No removing is authorized in this operator.
			Operator op = (Operator) e;
			int size = op.getOperandNumber();
			for (int i = 0; i < size; i++) {
				if (removeUseLessPNameEquality(op.getExpression(i)) == null) {
					throw new MedorException(
							"Impossible to remove the operand " + i
							+ " of operator " + e);
				}
			}
		}
		// nothing done.
		return e;
	}

	/**
	 * Indicates if two field are mapped over the same column, and replaces if
	 * required the use of the external table by the use of other.
	 * <p>If both tables are external then the second is replaced.
	 * Of course the replacement is done
	 * if the column are the same and if the 'replacetable' parameter value is
	 * 'true'.
	 * @param je1 is the extent of the first field.
	 * @param fieldname1 is the field name of the first field.
	 * @param je2 is the extent of the second field.
	 * @param fieldname2 is the field name of the second field.
	 * @param replacetable indicates if the use of the one table must be
	 * replaced by the use of the other.
	 * @return a boolean value indicating if two fields are mapped over the same
	 * column.
	 * @throws MedorException if a problem occurs during the fetching of the
	 * table.
	 */
	private boolean sameColumn(JormExtent je1,
							   String fieldname1,
							   JormExtent je2,
							   String fieldname2,
							   boolean replacetable) throws MedorException {
		RdbPrimitiveElementMapping pem1 =
				(RdbPrimitiveElementMapping) getPEM(je1, fieldname1);
		RdbTable t1 = (RdbTable) pem1.getParent();
		RdbPrimitiveElementMapping pem2 =
				(RdbPrimitiveElementMapping) getPEM(je2, fieldname2);
		RdbTable t2 = (RdbTable) pem2.getParent();
		boolean res = t1.getName().equals(t2.getName())
				&& t1.isColocated()
				&& t2.isColocated()
				&& pem1.getName().equals(pem2.getName());
		if (debug) {
			log.log(BasicLevel.DEBUG, "Comparing field " + fieldname1 + " to " + fieldname2
				+ ": result " + res);
		}
		if (res && replacetable) {
			RdbJormExtentMapping rjem1 = getMapping(je1);
			int idx1 = getTableIndex(rjem1, pem1, fieldname1);
			RdbJormExtentMapping rjem2 = getMapping(je2);
			int idx2 = getTableIndex(rjem2, pem2, fieldname2);
			if (t1 instanceof RdbExternalTable) {
				replaceTable(rjem2, idx2, rjem1, idx1);
			}
			else {
				replaceTable(rjem1, idx1, rjem2, idx2);
			}
		}
		return res;
	}

	/**
	 * Replaces the use of a table by another.
	 * @param rjem1 is the mapping of the extent which one table must be replaced
	 * @param idx1 is the index of the table to replace
	 * @param rjem2 is the mapping of the extent which contains the table to use
	 * @param idx2 is the index of the table to use
	 */
	private void replaceTable(RdbJormExtentMapping rjem1,
							  int idx1,
							  RdbJormExtentMapping rjem2,
							  int idx2) {
		if (debug) {
			log.log(BasicLevel.DEBUG, "Replacing table "
					+ idx1 + " of extent " + mappings.indexOf(rjem1)
					+ " by table "
					+ idx2 + " of extent " + mappings.indexOf(rjem2));
		}
		QualifiedTable old = (QualifiedTable) rjem1.tables.get(idx1);
		QualifiedTable neo = (QualifiedTable) rjem2.tables.get(idx2);
		for (int i = 0; i < mappings.size(); i++) {
			RdbJormExtentMapping r = (RdbJormExtentMapping) mappings.get(i);
			for (int j = 0; j < r.tables.size(); j++) {
				if (r.tables.get(j) == old) {
					r.tables.set(j, neo);
				}
			}
		}
	}

	/**
	 * Retrieves the index of the table which is reached with a given join name
	 * @param rjem is the mapping of the JormExtent which the table index is
	 * expected
	 * @param pem is the column which designates the table index expected
	 * @param fieldname is used to known which join must be used
	 * @throws MedorException if no join is found for the column if it was
	 * needed
	 */
	private int getTableIndex(RdbJormExtentMapping rjem,
							  RdbPrimitiveElementMapping pem,
							  String fieldname) throws MedorException {
		//one table
		if (rjem.tables.size() == 1 && rjem.joins.get(0) == null) {
			return 0;
		}
		Iterator it = pem.getPrimitiveElementByRdbJoin().entrySet().iterator();
		if (!it.hasNext())
			return 0;
		RdbJoin j = null;
		while (it.hasNext()) {
			Map.Entry me = (Map.Entry) it.next();
			if (((PrimitiveElement) me.getValue()).getName().equals(fieldname)) {
				j = (RdbJoin) me.getKey();
				break;
			}
		}
		if (j == null) {
			//field not found in joins
			throw new MedorException(
					"No join found for the mapping of the column '"
					+ pem.getName() + "'");
		}
		return rjem.getJoinIndex(j.getName());
	}


	/**
	 * Replaces the uses of the old fields of extents by fields of a leaf.
	 * The PNameField uses are replaced by the corresponding equality between
	 * fields of the name(single or composite).
	 * @param e the expression to update
	 * @param leaf is the field containing the new field.
	 * @param add indicates whether the Field should be added (it is projected)
	 * or not (it is only in the Filter).
	 * @return the updated expression (Never null).
	 * @throws MedorException if it is not possible to fetch a new field in the
	 * leaf or if the PName equality are malformed
	 */
	public Expression updateFieldOfExpression(Expression e,
											  BasicRdbExpQueryLeaf leaf,
											  boolean add)
			throws MedorException {
		log.log(BasicLevel.DEBUG, "Entering updateFieldOfExpression for " + ExpressionPrinter.e2str(e));
		if (e == null) {
			return null;
		}
		if ((e instanceof Equal)
				|| (e instanceof NotEqual)
				|| (e instanceof MemberOf)) {
			Operator eq = (Operator) e;
			Expression left = eq.getExpression(0);
			Expression right = eq.getExpression(1);
			if (left instanceof FieldOperand && right instanceof FieldOperand) {
				FieldOperand fo1 = (FieldOperand) left;
				FieldOperand fo2 = (FieldOperand) right;
				QueryTreeField qtf1 = (QueryTreeField) fo1.getField();
				QueryTreeField qtf2 = (QueryTreeField) fo2.getField();
				if (qtf1 instanceof PNameField && qtf2 instanceof PNameField) {
					//This is a PName equality

					PNameField pnf1 = (PNameField) qtf1;
					JormExtent je1 = (JormExtent) pnf1.getQueryTree();
					NameDef nd1 = pnf1.getNameDef(je1);

					PNameField pnf2 = (PNameField) qtf2;
					JormExtent je2 = (JormExtent) pnf2.getQueryTree();
					NameDef nd2 = pnf2.getNameDef(je2);

					if (nd1.isFieldName() && nd2.isFieldName()) {
						// only the operand must be changed
						if (debug) {
							log.log(BasicLevel.DEBUG, "Replace the PName comparison: "
									+ ExpressionPrinter.e2str(e));
						}
						fo1.setField(getField(nd1.getFieldName(), je1, leaf, add));
						fo2.setField(getField(nd2.getFieldName(), je2, leaf, add));
						if (debug) {
							log.log(BasicLevel.DEBUG, "by the field comparison: "
									+ ExpressionPrinter.e2str(e));
						}
						if (e instanceof MemberOf) {
							mofs.add(e);
							extents.add(je2);
						}
						return e;
					}
					else if (nd1.isNameRef() && nd2.isNameRef()) {
						if (!nd1.getNameRef().getCompositeName().getFQName()
								.equals(nd2.getNameRef().getCompositeName().getFQName())) {
							throw new MalformedExpressionException("Bad PName " +
									"equality: two different composite name");
						}
						Map proj1 = nd1.getNameRef().getProjection();
						Map proj2 = nd2.getNameRef().getProjection();
						if (proj1.size() != proj2.size()) {
							throw new MalformedExpressionException(
									"Bad PName equality: bad projection size");
						}
						//replace the PName equality by the equalities of the
						// PName fields
						if (debug) {
							log.log(BasicLevel.DEBUG, "Replace the PName comparison: "
									+ ExpressionPrinter.e2str(e));
						}
						Expression res = null;
						Iterator it = proj1.entrySet().iterator();
						ArrayList al1 = null;
						ArrayList al2 = null;
						if (e instanceof MemberOf) {
							al1 = new ArrayList(1);
							al2 = new ArrayList(1);
						}
						while (it.hasNext()) {
							Map.Entry me = (Map.Entry) it.next();
							String cofn = (String) me.getKey();
							String clfn1 = (String) me.getValue();
							String clfn2 = (String) proj2.get(cofn);
							Field f1 = getField(clfn1, je1, leaf, add);
							Field f2 = getField(clfn2, je2, leaf, add);
							if (e instanceof Equal) {
								Equal equal = new Equal(
										new BasicFieldOperand(f1),
										new BasicFieldOperand(f2));
								if (debug) {
									log.log(BasicLevel.DEBUG, "equality: "
											+ ExpressionPrinter.e2str(equal));
								}
								if (res == null) {
									res = equal;
								}
								else {
									res = new And(res, equal);
								}
							}
							else if (e instanceof NotEqual) {
								NotEqual notequal = new NotEqual(
										new BasicFieldOperand(f1),
										new BasicFieldOperand(f2));
								if (debug) {
									log.log(BasicLevel.DEBUG, "not-equality: "
											+ ExpressionPrinter.e2str(notequal));
								}
								if (res == null) {
									res = notequal;
								}
								else {
									res = new And(res, notequal);
								}
							}
							else if (e instanceof MemberOf) {
								al1.add(new BasicFieldOperand(f1));
								al2.add(new BasicFieldOperand(f2));
							}
						}
						if (e instanceof MemberOf) {
							MemberOf memberof = new MemberOf(al1, al2);
							if (debug) {
								log.log(BasicLevel.DEBUG, "memberof: "
										+ ExpressionPrinter.e2str(memberof));
							}
							if (res == null) {
								res = memberof;
							}
							else {
								res = new And(res, memberof);
							}
							if (e instanceof MemberOf) {
								mofs.add(memberof);
								mofExtents.add(je2);
							}
						}
						if (debug) {
							log.log(BasicLevel.DEBUG, "by the fields comparison: "
									+ ExpressionPrinter.e2str(res));
						}
						return res;
					}
					else {
						throw new MalformedExpressionException("Bad PName " +
								"equality: PName structures unknwon or not same");
					}
				}
				//not PName fields
				else if (e instanceof MemberOf) {
					if (debug) {
						log.log(BasicLevel.DEBUG, "MemberOf on normal fields");
					}
					JormExtent je1 = (JormExtent) qtf1.getQueryTree();
					JormExtent je2 = (JormExtent) qtf2.getQueryTree();
					if (debug) {
						log.log(BasicLevel.DEBUG, "Replace the MemberOf: "
								+ ExpressionPrinter.e2str(e));
					}
					ArrayList al1 = new ArrayList(1);
					ArrayList al2 = new ArrayList(1);
					Field f1 = getField(
							qtf1.getName().substring(je1.getName().length() + 1),
							je1, leaf, add);
					Field f2 = getField(
							qtf2.getName().substring(je2.getName().length() + 1),
							je2, leaf, add);
					al1.add(new BasicFieldOperand(f1));
					al2.add(new BasicFieldOperand(f2));
					MemberOf memberof = new MemberOf(al1, al2);

					if (debug) {
						log.log(BasicLevel.DEBUG, "by the field comparison: "
								+ ExpressionPrinter.e2str(memberof));
					}
					mofs.add(memberof);
					extents.add(je2);
					return memberof;
				}
			} else if (left instanceof FieldOperand && right instanceof Operand) {
                FieldOperand fo = (FieldOperand) left;
                QueryTreeField qtf = (QueryTreeField) fo.getField();
                if (qtf instanceof PNameField) {
                    return replacePNameOperandEquality(eq, leaf, add);
                }
            } else if (left instanceof Operand && right instanceof FieldOperand) {
                FieldOperand fo = (FieldOperand) right;
                QueryTreeField qtf = (QueryTreeField) fo.getField();
                if (qtf instanceof PNameField) {
                    return replacePNameOperandEquality(eq, leaf, add);
                }
            }
			//TODO for MemberOf add the case of value comparison (not PName)
		}

		//Not PName equality
		if (e instanceof Operator) {
			Operator op = (Operator) e;
			if (debug) {
				log.log(BasicLevel.DEBUG, "update the operator: " + op.getOperatorString());
			}
			int size = op.getOperandNumber();
			for (int i = 0; i < size; i++) {
				op.setExpression(i,
						updateFieldOfExpression(op.getExpression(i), leaf, add));
			}
		}
		else if (e instanceof FieldOperand) {
			Field f = ((FieldOperand) e).getField();
			if (f instanceof PNameField) {
				Expression res = getDecode((PNameField) f, leaf, add);
				if (debug) {
					log.log(BasicLevel.DEBUG, "Replace the FieldOperand "
							+ f.getName() + " by a decode: "
							+ ExpressionPrinter.e2str(res));
				}
				return res;
			}
			else {
				if (debug) {
					log.log(BasicLevel.DEBUG,
							"Update the FieldOperand " + f.getName());
				}
				((FieldOperand) e).setField(getField(f, leaf, add));
				return e;
			}
		}
		return e;
	}

	/**
	 * Create an expression to decode a PName.
	 * @param pnf is the PNameField linked to JormExtent which is the old
	 * representation of the PName.
	 * @param leaf is the leaf which contains the field of the NameDef. It the
	 * fields do not exist then they are created.
	 * @return an expression to decode a PName.
	 * @throws MedorException if the NameDef is not supported or if the a field
	 * can not be created in the leaf.
	 */
	public Expression getDecode(PNameField pnf,
								BasicRdbExpQueryLeaf leaf,
								boolean add)
			throws MedorException {
		if (debug) {
			log.log(BasicLevel.DEBUG, "(getDecode) Handling PNameField " + pnf.getName());
		}
		JormExtent extent = (JormExtent) pnf.getQueryTree();
		if (debug) {
			log.log(BasicLevel.DEBUG, "in extent " + extent.getJormName());
		}
		Operand o = getPNCOperand(extent, pnf);
		PType type = pnf.getType();
		log.log(BasicLevel.DEBUG, "Looking for NameDef of  " + pnf);
		NameDef nd = pnf.getNameDef(extent);
		if (nd.isFieldName()) {
			return new SinglePName(new BasicFieldOperand(
					getField(nd.getFieldName(), extent, leaf, add)), o, type);
		}
		else if (nd.isNameRef()) {
			Map proj = nd.getNameRef().getProjection();
			FieldOperand[] fos = new BasicFieldOperand[proj.size()];
			String[] cofns = new String[proj.size()];
			int counter = 0;
			QueryTreeField qtf = null;
			for (Iterator it = proj.entrySet().iterator(); it.hasNext();) {
				Map.Entry me = (Map.Entry) it.next();
				cofns[counter] = (String) me.getKey();
				String clafn = (String) me.getValue();
				qtf = getField(clafn, extent, leaf, add);
				fos[counter] = new BasicFieldOperand(qtf);
				counter++;
			}
			//add a CalculatedField into the intermediate QueryTree.
			// This calculatedField has an expression which is a
			// CompositePName.
			return new CompositePName(fos, cofns, o, type);
		}
		throw new MedorException("Unsupport namedef: " + nd);
	}


	/**
	 * Replaces the equality between a PName and an operand.
	 * @param e is the Equal, NotEqual or MemberOf object which must have a
	 * FieldOperand referencing a PNameField and an Operand.
	 * @param leaf is the leaf containg the new fields composing the PNameField
	 * @return the new expression comparing PName fields to operands.
	 * @throws MedorException if the expression is malformed or if it is not
	 * possible to fetch a new Field on the leaf.
	 */
	private Expression replacePNameOperandEquality(Operator e,
                                                    BasicRdbExpQueryLeaf leaf,
                                                    boolean add)
			throws MedorException {
		if (debug) {
			log.log(BasicLevel.DEBUG, "Equality between PName and a parameter");
		}
		Expression left = e.getExpression(0);
		Expression right = e.getExpression(1);
		FieldOperand fo = null;
		PNameField pnf = null;
		Operand o = null;
		boolean isLeft = true;
		if (left instanceof FieldOperand && right instanceof Operand) {
			fo = (FieldOperand) left;
			pnf = (PNameField) fo.getField();
			o = (Operand) right;
			isLeft = false;
		}
		else if (left instanceof Operand && right instanceof FieldOperand) {
			fo = (FieldOperand) right;
			pnf = (PNameField) fo.getField();
			o = (Operand) left;
		}
		JormExtent je = (JormExtent) pnf.getQueryTree();
        if (e instanceof MemberOf) {
            mofs.add(e);
            extents.add(je);
        }
		NameDef nd = pnf.getNameDef(je);
        VariableOperand newo;
		if (nd.isFieldName()) {
			Field f = getField(nd.getFieldName(), je, leaf, add);
			fo.setField(f);
            if (o instanceof ParameterOperand) {
                newo = new EncodePName(f.getType(), ((ParameterOperand) o).getName());
                e.setExpression((isLeft ? 0 : 1), newo);
            } else {
                newo = new BasicVariableOperand(f.getType());
                try {
                    EncodePName.assignEncodedValue((PName) o.getObject(), newo, null);
                } catch (PException ex) {
                    throw new MedorException("Impossible to encode PName", ex);
                }
                if (newo.getObject() == null) {
                    //transform to a isNull operator
                    return new IsNull(fo);
                }
                e.setExpression((isLeft ? 0 : 1), newo);
            }
			if (debug) {
				log.log(BasicLevel.DEBUG,
						"Equality between PName and a parameter: "
						+ ExpressionPrinter.e2str(e));
			}
			return e;
		}
		else if (nd.isNameRef()) {
			Expression res = null;
			Map proj1 = nd.getNameRef().getProjection();
			Iterator it = proj1.entrySet().iterator();
			while (it.hasNext()) {
				Map.Entry me = (Map.Entry) it.next();
				String cofn = (String) me.getKey();
				String clfn = (String) me.getValue();
				Field f = getField(clfn, je, leaf, add);
                if (o instanceof ParameterOperand) {
                    newo = new EncodePName(f.getType(), ((ParameterOperand) o).getName(), cofn);
                } else {
                    newo = new BasicVariableOperand(f.getType());
                    try {
                        EncodePName.assignEncodedValue((PName) o.getObject(), newo, cofn);
                    } catch (PException ex) {
                        throw new MedorException("Impossible to encode PName", ex);
                    }
                }
				if (e instanceof Equal) {
                    Expression test;
                    if (newo.getObject() == null) {
                        //transform to a isNull operator
                        test = new IsNull(fo);
                    } else {
                        test = (isLeft
                                ? new Equal(newo, new BasicFieldOperand(f))
                                : new Equal(new BasicFieldOperand(f), newo));
                    }
					if (res == null) {
						res = test;
					} else {
						res = new And(res, test);
					}
				}
				if (e instanceof MemberOf) {
					ArrayList al1 = new ArrayList(1);
					al1.add(newo);
					ArrayList al2 = new ArrayList(1);
					al2.add(new BasicFieldOperand(f));
					MemberOf memberof =
							new MemberOf(al1, al2);
					if (res == null) {
						res = memberof;
					}
					else {
						res = new And(res, memberof);
					}
				}
				else if (e instanceof NotEqual) {
					NotEqual notequal =
							(isLeft
							? new NotEqual(newo, new BasicFieldOperand(f))
							: new NotEqual(new BasicFieldOperand(f), newo));
					if (res == null) {
						res = notequal;
					}
					else {
						res = new Or(res, notequal);
					}
				}
			}
			if (debug) {
				log.log(BasicLevel.DEBUG,
						"Equality between PName and a parameter: "
						+ ExpressionPrinter.e2str(res));
			}
			return res;
		}
		else {
			throw new MalformedExpressionException("Bad PName " +
					"equality: PName structures unknwon or not same");
		}
	}


	/**
	 * Is a shorcut to the
	 * getField(String, PType, JormExtent, BasicRdbExpQueryLeaf) method.
	 * The fieldname, the type and the extent values are calculated since the
	 * old Field.
	 * @param old is the old field
	 * @throws MedorException if an error occurs during the real getField call.
	 */
	public QueryTreeField getField(Field old,
								   BasicRdbExpQueryLeaf leaf,
								   boolean add)
			throws MedorException {
		QueryTree qt = ((QueryTreeField) old).getQueryTree();
		if (qt == leaf) {
			return (QueryTreeField) old;
		}
		else if (qt instanceof JormExtent) {
			return getField(
					old.getName(), old.getType(), (JormExtent) qt, leaf, add);
		}
		else {
			throw new MedorException("Impossible to fetch the new field " +
					"corresponding to the old field: " + old.getName()
					+ " : the query tree is not a JormExtent: " + qt);
		}
	}

	/**
	 * Is a shorcut to the
	 * getField(String, PType, JormExtent, BasicRdbExpQueryLeaf) method.
	 * The field type is calculated looking for the field in the JORM meta
	 * information.
	 * @throws MedorException if the field is not found in the JORM meta
	 * information or if an error occurs during the real getField call.
	 */
	public QueryTreeField getField(String fieldname,
								   JormExtent extent,
								   BasicRdbExpQueryLeaf leaf,
								   boolean add)
			throws MedorException {
		if (debug) {
			log.log(BasicLevel.DEBUG, "(getField-4) Handling field " + fieldname);
		}
		return getField(fieldname,
				getPrimitiveElement(extent, fieldname).getType(), extent, leaf, add);
	}

	/**
	 * Adds a field in a leaf if it is not already present.
	 * @param fieldname is the field name of the expected field.
	 * @param type is the jorm type of expected field.
	 * @param extent is the extent containing the old version of the expected
	 * field.
	 * @param leaf is the leaf which must contain the expected field after the
	 * the method call
	 * @return the expected field
	 * @throws MedorException if no old field is not availlable on the extent
	 * or if no mapping is availlable for the field.
	 */
	public QueryTreeField getField(String fieldname,
								   PType type,
								   JormExtent extent,
								   BasicRdbExpQueryLeaf leaf,
								   boolean add)
			throws MedorException {
		if (debug) {
			log.log(BasicLevel.DEBUG, "(getField-5) Handling field " + fieldname);
		}
		String fqfn = qualifiedName(fieldname, extent);
		if (leaf.contains(fqfn)) {
			//The field already exists then return it
			if (debug) {
				log.log(BasicLevel.DEBUG,
						"getField:"
						+ " / field: " + fqfn
						+ " / type: " + type.getJormName()
						+ " already exists");
			}
			return (QueryTreeField) leaf.getField(fqfn);
		}
		RdbPrimitiveElementMapping pem =
				(RdbPrimitiveElementMapping) getPEM(extent, fieldname);
		RdbJormExtentMapping rjem = getMapping(extent);
		int idx = getTableIndex(rjem, pem, fieldname);
		if (debug) {
			log.log(BasicLevel.DEBUG,
					"getField:"
					+ " / field: " + fqfn
					+ " / type: " + type.getJormName()
					+ " / pem: " + pem.getName()
					+ " / rjem: " + rjem.extent.getJormName()
					+ " / table: "
					+ ((QualifiedTable) rjem.tables.get(idx)).getAliasName());
		}
		getTable(rjem, idx, leaf, add);
		RdbExpField res = leaf.addRdbField(
				fqfn,
				type, pem.getName(),
				(QualifiedTable) rjem.tables.get(idx));
		if (!add) {
			leaf.removeRdbField(res);
		}
		return res;
	}

	/**
	 * It adds (if it is not already present) a QualifiedTable instance into a
	 * leaf. The joins are automatically added in the global variable 'filter'.
	 * @param rjem is the jorm extent mapping containg the expected table
	 * @param idx is the index of the table into the list of tables of the
	 * jorm extent mapping rjem
	 * @param leaf is the BasicRdbExpQueryLeaf which must contain the expected
	 * qualified table after the call to this method.
	 * @throws MedorException if an exception occurs during the addition of
	 * join between table.
	 */
	private void getTable(RdbJormExtentMapping rjem,
						  int idx,
						  BasicRdbExpQueryLeaf leaf,
						  boolean add) throws MedorException {
		QualifiedTable qt = (QualifiedTable) rjem.tables.get(idx);
		if (debug) {
			log.log(BasicLevel.DEBUG, "Looking for the table " + qt.getAliasName());
		}
		if (leaf.containsQualifiedTable(qt))
			return;
		if (idx != 0) {
			//The external table must be added to the leaf
			getTable(rjem, 0, leaf, add);
			if (debug) {
				log.log(BasicLevel.DEBUG, "Add the external table " + qt.getAliasName());
			}
			leaf.addQualifiedTable(qt);
			addJoin(rjem, idx, leaf, add);
		}
		else {
			//The main table must be added to the leaf
			if (debug) {
				log.log(BasicLevel.DEBUG, "Add the main table " + qt.getAliasName());
			}
			leaf.addQualifiedTable(qt);
			//looking if this table is used as external table by other extent
			boolean unused = true;
			for (int i = 0; i < mappings.size(); i++) {
				RdbJormExtentMapping m = (RdbJormExtentMapping) mappings.get(i);
				//don't consider the first table which is a main table.
				int jdx = 1;
				int size = m.tables.size();
				while (jdx < size && m.tables.get(jdx) != qt) {
					jdx++;
				}
				if (jdx < size) {
					unused = false;
					// Add the main table of the extent which uses the expected
					// main table
					getTable(m, 0, leaf, add);
					//Of course add the join
					addJoin(m, jdx, leaf, add);
				}
			}
			if (unused && !joinedTables.contains(qt)) {
				joinedTables.add(qt);
			}
		}
	}

	/**
	 * Adds a join expression into the global variable 'filter". The join is
	 * between an axternal table and the main table of a jorm extent mapping.
	 * @param rjem is the jorm extent mapping containing the tables to join
	 * @param idx is the index of the external table in the table list of the
	 * jorm extent mapping
	 * @param leaf is the leaf hosting the field of the join.
	 * @throws MedorException if an error occurs during the fetching of join
	 * field.
	 */
	private void addJoin(RdbJormExtentMapping rjem,
						 int idx,
						 BasicRdbExpQueryLeaf leaf,
						 boolean add) throws MedorException {
		BasicQualifiedTable qt = (BasicQualifiedTable) rjem.tables.get(idx);
		JoinedTable.Join jtj = ((JoinedTable)
				rjem.tables.get(0)).createChildren(qt);
		Join j = (Join) rjem.joins.get(idx);
		RdbField f2;
		for (int i = 0; i < j.columns.length; i++) {
			JoinColumn jc = j.columns[i];
			if (jc.field2 == null) {
				//create a dummy field in the leaf
				f2 = leaf.addRdbField(jc.column2, jc.type, jc.column2, qt);
			}
			else {
				f2 = (RdbField) getField(jc.field2, rjem.extent, leaf, add);
			}
			RdbField f1 = (RdbField) getField(jc.field1, rjem.extent, leaf, add);
			jtj.addJoinColumn(f1.getColumnName(), f2.getColumnName());
		}
	}

	private String qualifiedName(String fieldname, JormExtent extent) {
		if (fieldname.indexOf(extent.getName()) == -1) {
			return extent.getName() + "." + fieldname;
		}
		else
			return fieldname;
	}

	/**
	 * For a given QueryLeaf, splits the leaf and adds subleaves if there is
	 * a MemberOf in the filter.
	 * @param leaf The Queryleaf to be split
	 * @param filter The Expression attached to the QueryLeaf
	 * @throws MedorException
	 */
	private void splitLeafIfMemberOf(RdbExpQueryLeaf leaf,
									 Expression filter)
			throws MedorException {
		if (mofs.isEmpty()) {
			if (debug) {
				log.log(BasicLevel.DEBUG, "No MemberOfs to update");
			}
			return;
		}
		//build new query leaves with the mof2extent
		if (debug) {
			log.log(BasicLevel.DEBUG, "Creating subleaves for filter " + ExpressionPrinter.e2str(filter));
		}
		for (int i = 0; i < mofs.size(); i++) {
			//create new RdbExpQueryLeaf
			RdbExpQueryLeaf newLeaf = new BasicRdbExpQueryLeaf(leaf.getDataStore(), "");
			subLeaves.add(newLeaf);
			MemberOf mo = (MemberOf) mofs.get(i);
			int moSize = mo.getOperandNumber() / 2;
			//iterate on all pairs of Fields
			if (debug) {
				for (int f = 0; f < moSize; f++) {
                    log.log(BasicLevel.DEBUG, "- left is "
                            + ExpressionPrinter.e2str(mo.getExpression(f)));
				}
			}
			for (int f = 0; f < moSize; f++) {
				RdbExpField fRight =
						(RdbExpField)
						((FieldOperand)
						mo.getExpression(f + moSize)).getField();
				if (debug) {
					log.log(BasicLevel.DEBUG, "- right field is " + fRight.getName());
				}
				//remove the right QTable from main leaf if it is in it
				if (leaf.containsQualifiedTable(fRight.getTable())) {
					if (debug) {
						log.log(BasicLevel.DEBUG, "- right table is " + fRight.getTable().getTableName() + " (alias " + fRight.getTable().getAliasName() + ")");
						QualifiedTable[] qtstmp = leaf.getQualifiedTables();
						log.log(BasicLevel.DEBUG, "- before remove: " + qtstmp.length);
						log.log(BasicLevel.DEBUG, "-- leaf " + leaf);
						log.log(BasicLevel.DEBUG, "-- qt " + fRight.getQueryTree());
						for (int ii = 0; ii < qtstmp.length; ii++) {
							log.log(BasicLevel.DEBUG, "--- table " + qtstmp[ii].getTableName());
						}
					}
					//TODO replace the code below with moving all tables of the extent
					leaf.removeQualifiedTable(fRight.getTable());
					//OK
					newLeaf.addQualifiedTable(fRight.getTable());
					newLeaf.addRdbField(fRight);
					fRight.setQueryLeaf(newLeaf);
					//end OK
					if (debug) {
						QualifiedTable[] qtstmp = leaf.getQualifiedTables();
						log.log(BasicLevel.DEBUG, "- after remove: " + qtstmp.length);
						for (int ii = 0; ii < qtstmp.length; ii++) {
							log.log(BasicLevel.DEBUG, "-- table " + qtstmp[ii].getTableName());
						}
					}
				}
				else {
					if (debug) {
						log.log(BasicLevel.DEBUG, "- right table aleady moved");
					}
					newLeaf.addRdbField(fRight);
					fRight.setQueryLeaf(newLeaf);
				}
				if (joinedTables.contains(fRight.getTable())) {
					if (debug) {
						log.log(BasicLevel.DEBUG, "************Removing the joined table " + fRight.getTable().getTableName() + ".");
					}
					joinedTables.remove(fRight.getTable());
				}
			}
		}
		return;
	}

	/**
	 * Scans the filter and moves expressions to subleaves in case of a
	 * MemberOf.
	 * <p>The real work is done by the recursive method tagFilterForMove
	 * @param filter The query filter of the newly created QueryLeaf
	 * @param leaf The "root" QueryLeaf
	 * @return an updated Expression for the root QueryLeaf.
	 */
	private Expression updateFilterWithNewLeaves(Expression filter,
												 RdbExpQueryLeaf leaf) {
		if (debug) {
			log.log(BasicLevel.DEBUG, "Moving part of the filter " + ExpressionPrinter.e2str(filter));
		}
		ToBeMovedToLeaf tbm = tagFilterForMove(filter, leaf);
		if (debug) {
			log.log(BasicLevel.DEBUG, "Final filter for leaf: " + ExpressionPrinter.e2str(tbm.e));
		}
		if (tbm.isModified) {
			//the expression should be moved to the leaf
			tbm.leaf.setQueryFilter(tbm.e);
			return null;
		}
		else {
			return tbm.e;
		}
	}

	/**
	 * Recursively parses a filter for moving parts of the filter to sub-
	 * query leaves in the case of MemberOf.
	 * <p>Comparisons between fieldsOperands of a Field which is in a subleaf
	 * are moved to that subleaf. The actual move is potentially performed by
	 * the upper node in the filter (namely in the case of a "Not").
	 * @param filter The portion of the filter to be tested.
	 * @param leaf The root query leaf.
	 * @return a ToBeMovedToLeaf object, where "isModified" is true if the
	 * expression "e" should be moved to leaf "leaf".
	 */
	private ToBeMovedToLeaf tagFilterForMove(Expression filter,
											 RdbExpQueryLeaf leaf) {
		if (filter instanceof Not) {
			ToBeMovedToLeaf me =
					tagFilterForMove(((Not) filter).getExpression(), leaf);
			if (me.isModified) {
				//the expression should be moved to the leaf
				Expression newFilter = null;
				if (leaf.getQueryFilter() == null) {
					newFilter = new Not(me.e);
				}
				else {
					newFilter = new And(leaf.getQueryFilter(), new Not(me.e));
				}
				me.leaf.setQueryFilter(newFilter);
				me.isModified = false;
				me.e = null;
				return me;
			}
		}
		else if (filter instanceof And) {
			if (debug) {
				log.log(BasicLevel.DEBUG, "Found AND " +
					ExpressionPrinter.e2str(filter));
				log.log(BasicLevel.DEBUG, "Left side of AND: "
					+ ExpressionPrinter.e2str(((And) filter).getLeftExpression()));
			}
			ToBeMovedToLeaf mleft = tagFilterForMove(((And) filter).getLeftExpression(), leaf);
			boolean leftMoved = false;
			boolean rightMoved = false;
			if (mleft.isModified) {
				//the expression should be moved to the leaf
				if (debug) {
					log.log(BasicLevel.DEBUG, "Moving left side to leaf "
					+ ExpressionPrinter.e2str(mleft.e));
				}
				leftMoved = true;
				mleft.leaf.setQueryFilter(
						((mleft.leaf.getQueryFilter() == null)
						? mleft.e
						: new And(
								mleft.leaf.getQueryFilter(),
								mleft.e)
						)
				);
				if (debug) {
					log.log(BasicLevel.DEBUG, "Filter for leaf "
							+ mleft.leaf + " after moving left side is "
							+ ExpressionPrinter.e2str(mleft.leaf.getQueryFilter()));
				}
			}
			if (debug) {
				log.log(BasicLevel.DEBUG, "Right side of AND: "
					+ ExpressionPrinter.e2str(((And) filter).getRightExpression()));
			}
			ToBeMovedToLeaf mright = tagFilterForMove(((And) filter).getRightExpression(), leaf);
			if (mright.isModified) {
				if (debug) {
					log.log(BasicLevel.DEBUG, "Moving right side to leaf "
					+ ExpressionPrinter.e2str(mright.e));
				}
				rightMoved = true;
				mright.leaf.setQueryFilter(
						(mright.leaf.getQueryFilter() == null
						? mright.e
						: new And(
								mright.leaf.getQueryFilter(),
								mright.e)
						)
				);
				if (debug) {
					log.log(BasicLevel.DEBUG, "Filter for leaf "
							+ mright.leaf + " after moving right side is "
							+ ExpressionPrinter.e2str(mright.leaf.getQueryFilter()));
				}
			}
			if (leftMoved) {
				if (rightMoved) {
					//left and right already moved
					mleft.clear();
					if (debug) {
						log.log(BasicLevel.DEBUG, "All moved: returning empty "
								+ ExpressionPrinter.e2str(mleft.e));
					}
					return mleft;
				}
				else {
					//only left already moved
					if (debug) {
						log.log(BasicLevel.DEBUG, "Left moved: returning right "
								+ ExpressionPrinter.e2str(mright.e));
					}
					return mright;
				}
			}
			else {
				if (rightMoved) {
					//only right already moved
					if (debug) {
						log.log(BasicLevel.DEBUG, "Right moved: returning left "
								+ ExpressionPrinter.e2str(mleft.e));
					}
					return mleft;
				}
				else {
					//nothing moved but could be modified
					Expression modifAnd =
							new And(mleft.e, mright.e);
					if (debug) {
						log.log(BasicLevel.DEBUG, "None moved: returning modified expression "
								+ ExpressionPrinter.e2str(modifAnd));
					}
					return new ToBeMovedToLeaf(modifAnd, false, null);
				}
			}
		}
		//TODO add treatment of OR
		else if (filter instanceof Comparator) {
			//get the two operands
			Expression left = ((Comparator)filter).getExpression(0);
			Expression right = ((Comparator)filter).getExpression(1);
			if (left instanceof FieldOperand) {
				//checks if the FieldOperand is for one of the subleaves
				RdbExpQueryLeaf subleaf =
						subleafOfFieldOperand((FieldOperand) left);
				if (subleaf != null) {
					//mark the subexpression to be moved
					return new ToBeMovedToLeaf(filter, true, subleaf);
				}
			}
			if (right instanceof FieldOperand) {
				//checks if the FieldOperand is for one of the subleaves
				RdbExpQueryLeaf subleaf =
						subleafOfFieldOperand((FieldOperand) right);
				if (subleaf != null) {
					//mark the subexpression to be moved
					return new ToBeMovedToLeaf(filter, true, subleaf);
				}
			}
			//TODO add treatment for arithmetic operators
		}
		//otherwise, nothing is moved
		return new ToBeMovedToLeaf(filter, false, null);
	}

	/**
	 * Returns, if it exists, the subleaf in which a given FieldOperand appears.
	 * @param fo The FieldOperand to be looked for.
	 * @return the subleaf in which the FieldOperand appears, if it exists.
	 * Otherwise, return null.
	 */
	private RdbExpQueryLeaf subleafOfFieldOperand(FieldOperand fo) {
		QualifiedTable qtab = ((RdbExpField)fo.getField()).getTable();
		for (int i = 0; i < subLeaves.size(); i++) {
			RdbExpQueryLeaf subleaf =
					(RdbExpQueryLeaf) subLeaves.get(i);
			QualifiedTable[] tables = subleaf.getQualifiedTables();
			for (int j = 0; j < tables.length; j++) {
				if (qtab == tables[j]) {
					if (debug) {
						log.log(BasicLevel.DEBUG, "Found table " + qtab.getTableName() + " in leaf " + subleaf);
					}
					return subleaf;
				}
			}
		}
		return null;
	}

	/**
	 * Extends ModifiedExpression for moving part of a filter to the subleaves.
	 * <p>Note that isModified has a different meaning here: it means that
	 * the Expression should be moved to the leaf.
	 */
	private class ToBeMovedToLeaf extends ModifiedExpression {
		/**
		 * The RdbExpQueryLeaf to which the filter portion should be moved.
		 */
		RdbExpQueryLeaf leaf;

		ToBeMovedToLeaf(Expression exp, boolean ism, RdbExpQueryLeaf leaf) {
			super(exp, ism);
			this.leaf = leaf;
		}

		public void clear() {
			e = null;
			leaf = null;
			isModified = false;
		}
	}
}
