/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.optim.jorm;

import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.filter.api.Expression;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.filter.api.Operator;
import org.objectweb.medor.optim.api.LeafRewriteRule;
import org.objectweb.medor.optim.api.LeafRewriter;
import org.objectweb.medor.optim.lib.BasicRule;
import org.objectweb.medor.query.api.CalculatedField;
import org.objectweb.medor.query.api.FilteredQueryTree;
import org.objectweb.medor.query.api.NestedField;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.jorm.lib.ClassExtent;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
/**
 * This class represents the rule to transform QueryLeaves
 * into the corresponding QueryLeaf on the data store (for example into an
 * RdbQueryLeaf).
 */
public class JormLeafRewriteRule extends BasicRule implements LeafRewriteRule  {

	ArrayList leafRewriters = null;

	public JormLeafRewriteRule() {
		super("JormLeafRewriteRule");
		leafRewriters = new ArrayList();
	}

	public JormLeafRewriteRule(LeafRewriter lr) {
		this();
		leafRewriters.add(lr);
	}

    // IMPLEMENTATION OF THE LeafRewriteRule INTERFACE //
    //-------------------------------------------------//

	public void addLeafRewriter(LeafRewriter lr) {
		leafRewriters.add(lr);
	}
	public void removeLeafRewriter(LeafRewriter lr) {
		leafRewriters.add(lr);
	}
	public Collection getLeafRewriters() {
		return leafRewriters;
	}
	public LeafRewriter getLeafRewriter(QueryLeaf ql) {
		for (Iterator it=leafRewriters.iterator(); it.hasNext();){
			LeafRewriter lr = (LeafRewriter) it.next();
			if (lr.canRewrite(ql))
				return lr;
		}
        if (debug)
            log.log(BasicLevel.DEBUG, "No LeafRewriter found (among "
                + leafRewriters.size() + ") for this QueryLeaf: " + ql);
		return null;
	}

    // IMPLEMENTATION OF THE Loggable INTERFACE //
    //------------------------------------------//

    public void setLoggerFactory(LoggerFactory loggerFactory) {
    }

    public LoggerFactory getLoggerFactory() {
        return null;
    }

    public void setLogger(Logger logger) {
        this.log = logger;
    }

    public Logger getLogger() {
        return log;
    }

    // IMPLEMENTATION OF THE RewriteRule INTERFACE //
	//---------------------------------------------//

	/**
	 */
	public QueryTree rewrite(QueryTree qt) throws MedorException {
		debug = log != null && log.isLoggable(BasicLevel.DEBUG);

        if (debug)
            log.log(BasicLevel.DEBUG, "qt=" + qt);
		if (qt==null) return null;
		ClassExtent extent = null;
		if (qt instanceof ClassExtent) {
			//Just transform it, there is no user
			extent = (ClassExtent) qt;
			return getLeafRewriter(extent).rewrite(extent);
		}
		for(Iterator it=getLeafUsers(qt).entrySet().iterator(); it.hasNext();) {
			Map.Entry entry = (Map.Entry) it.next();
			extent = (ClassExtent) entry.getKey();
			ArrayList users = (ArrayList) entry.getValue();

			//Transform the Jorm extent into a specific QueryTree
			QueryTree neo = getLeafRewriter(extent).rewrite(extent);

			//Change the reference of the user
			for(Iterator it2 = users.iterator(); it2.hasNext();) {
				Object o = it2.next();
				if (o instanceof FieldOperand) {
					FieldOperand fo = (FieldOperand) o;
					fo.setField(
						neo.getTupleStructure().getField(
							fo.getField().getName()));
				}
				else if (o instanceof PropagatedField) {
					PropagatedField pf = (PropagatedField) o;
                    Field[] fs = pf.getPreviousFields();
                    QueryTreeField[] neos = new QueryTreeField[fs.length];
                    boolean mustReplace = false;
					for(int i=0; i<fs.length; i++) {
						if (fs[i] instanceof QueryTreeField
							&& ((QueryTreeField)fs[i]).getQueryTree()==extent) {
                            mustReplace = true;
                            neos[i] = (QueryTreeField)
                                neo.getTupleStructure().getField(fs[i].getName());
						}
                        else {
                            neos[i] = (QueryTreeField) fs[i];
                        }

					}
                    if (mustReplace)
                        ((QueryNode) pf.getQueryTree())
                            .updatePropagatedField(pf.getName(), neos);
				}
				else if (o instanceof NestedField) {
					NestedField nf = (NestedField) o;
					Field[] fs = nf.getFields();
                    QueryTreeField[] neos = new QueryTreeField[fs.length];
                    boolean mustReplace = false;
					for(int i=0; i<fs.length; i++) {
						if (fs[i] instanceof QueryTreeField
							&& ((QueryTreeField)fs[i]).getQueryTree()==extent) {
                            mustReplace = true;
                            neos[i] = (QueryTreeField)
                                neo.getTupleStructure().getField(fs[i].getName());
						}
                        else {
                            neos[i] = (QueryTreeField) fs[i];
                        }
					}
                    if (mustReplace)
                        ((QueryNode) nf.getQueryTree())
                            .updatePropagatedField(nf.getName(), neos);
				}
			}
		}
		return qt;
	}

	// OTHER METHODS //
	//---------------//

	/**
	 *
	 */
	protected Map getLeafUsers(QueryTree qt) throws MedorException {
		Map res = new HashMap();
		getLeafUsers(qt, res, new ArrayList());
		return res;
	}

	/**
	 * It fetches the QueryTree nodes which are 'parents' of QueryLeaf nodes.
	 * @param qt is the queryTree to analyze.
	 * @param m is the current result of the recusrive evaluation.
	 * key = a queryTree
	 * Value = the users of the querytree key :
	 * ArrayList( FieldOperand | PropagatedField | NestedField )
	 * @param al is the list of the visited queryTree. (It does not contain the
	 * 'qt' parameter)
	 */
	protected void getLeafUsers(QueryTree qt, Map m, ArrayList al) throws MedorException {
		al.add(qt);
		//Look in the filter
		if (qt instanceof FilteredQueryTree) {
			getLeafUsers(
			((FilteredQueryTree) qt).getQueryFilter(), m, al);
		}

		//look in projected field
		if (!(qt instanceof QueryLeaf)) {
			Field[] fields = qt.getTupleStructure().getFields();
			for(int i=0; i<fields.length; i++) {
				if (fields[i] instanceof PropagatedField) {
					getLeafUsers(
					((PropagatedField) fields[i]).getPreviousFields(),
					fields[i],
					m, al);
				}
				else if (fields[i] instanceof CalculatedField) {
					getLeafUsers(
					((CalculatedField) fields[i]).getExpression(), m, al);
				}
				else if (fields[i]  instanceof NestedField) {
					getLeafUsers(
					((NestedField)fields[i]).getFields(),
					fields[i],
					m, al);
				}
			}
		}
	}

	/**
	 * It fetches the QueryTree nodes which are 'parents' of QueryLeaf nodes.
	 * @param e is the expression to analyze.
	 * @param m is the current result of the recusrive evaluation.
	 * key = a queryTree
	 * Value = the users of the querytree key :
	 * ArrayList( FieldOperand | PropagatedField | NestedField )
	 * @param al is the list of the visited queryTree.
	 */
	private void getLeafUsers(Expression e, Map m, ArrayList al) throws MedorException {
		if (e instanceof Operator) {
			for (int i = 0; i< ((Operator) e).getOperandNumber(); i++)
				getLeafUsers(((Operator)e).getExpression(i), m, al);
		}
		else if (e instanceof FieldOperand) {
			FieldOperand fo = (FieldOperand) e;
			QueryTree qt = ((QueryTreeField) fo.getField()).getQueryTree();
			if (qt instanceof ClassExtent) {
				ArrayList users = (ArrayList) m.get(qt);
				if (users==null) {
					users = new ArrayList();
					m.put(qt, users);
				}
				users.add(fo);
			}
			else if (!al.contains(qt)) {
				getLeafUsers(qt, m, al);
			}
		}
	}

	/**
	 * It fetches the QueryTree nodes which are 'parents' of QueryLeaf nodes.
	 * @param usedFields are the ancestor of a potentiel users. If one of these
	 * fields is attached to a QueryLeaf the second parameter 'user' is
	 * registered as user in the map.
	 * @param user is the field which is based on the usedFields parameter. In
	 * most case this parameter is a PropagatedField or a NestedField.
	 * @param m is the current result of the recusrive evaluation.
	 * key = a queryTree
	 * Value = the users of the querytree key :
	 * ArrayList( FieldOperand | PropagatedField | NestedField )
	 * @param al is the list of the visited queryTree.
	 */
	private void getLeafUsers(Field[] usedFields, Object user, Map m, ArrayList al) throws MedorException {
		for(int i=0; i<usedFields.length; i++) {
			if (usedFields[i] instanceof QueryTreeField) {
				QueryTree qt = ((QueryTreeField)usedFields[i]).getQueryTree();
				if (qt instanceof ClassExtent) {
					ArrayList users = (ArrayList) m.get(qt);
					if (users==null) {
						users = new ArrayList();
						m.put(qt, users);
					}
					users.add(user);
				} else if (!al.contains(qt)) {
					getLeafUsers(qt, m, al);
				}

			}
		}
	}
}
