/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.optim.lib;

import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.lib.Log;
import org.objectweb.medor.optim.api.QueryRewriter;
import org.objectweb.medor.optim.api.RewriteRule;
import org.objectweb.medor.optim.api.RuleConfiguration;
import org.objectweb.medor.optim.rdb.GroupSameDBRule;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.lib.QueryTreePrinter;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

/**
 * This class implements the QueryRewriter interface AND also the
 * RuleConfiguration interface (the aim is to optimize the number of object).
 * The addDefaultRiles permits to assign the default rules in the right order.
 *
 * @author S.Chassande-Barrioz
 */

public class BasicQueryRewriter
	implements QueryRewriter, RuleConfiguration {

	protected List rules = null;

	protected Logger log = null;

	protected boolean debug = false;

	/**
	 * It builds a BasicQueryRewriter with no rule
	 */
	public BasicQueryRewriter() {
		this(new ArrayList());
	}

	/**
	 * It builds a BasicQueryRewriter with a ArrayList of rules specified in
	 * parameter.
	 * @param rules is the list of rules for this configuration
	 */
	public BasicQueryRewriter(ArrayList rules) {
		this.rules = rules;
		log = Log.loggerFactory.getLogger(
			Log.MEDOR_PREFIX + ".optim.rewriter.BasicQueryRewriter");
		debug = log != null && log.isLoggable(BasicLevel.DEBUG);
		if (debug){
			log.log(BasicLevel.DEBUG, "Creating Query Rewriter");
		}

	}

	/**
	 * It builds a BasicQueryRewriter with a Collection of rules specified in
	 * parameter.
	 * @param rules is the list of rules for this configuration
	 */
	public BasicQueryRewriter(Collection rules) {
		this.rules = new ArrayList(rules);
	}

	/**
	 * It adds the default Medor rules in this following order
	 * <OL>
	 * <UL>PushNotInExpressionRule</UL>
	 * <UL>PushSelectionRule</UL>
	 * <UL>DropUnusedProjFieldsRule</UL>
	 * <UL>DropUselessNodeRule</UL>
	 * <UL>GroupSameStoreRule</UL>
	 * </OL>
	 */
	public List getDefaultRules() {
		ArrayList al = new ArrayList();
		al.add(new PushNotInExpressionRule());
		al.add(new PushSelectionRule());
		al.add(new DropUnusedProjFieldsRule());
		al.add(new DropUselessNodeRule());
		al.add(new GroupSameDBRule());
		al.add(new DropUnusedProjFieldsRule());
		return al;
	}

	// IMPLEMENTATION OF THE QueryRewriter INTERFACE //
	//-----------------------------------------------//

	public void setRuleConfiguration(RuleConfiguration config) {
		rules = config.getRules();
	}

	public QueryTree transform(QueryTree qt) throws MedorException {
		if (debug){
			log.log(BasicLevel.DEBUG, "Entering transformation");
		}
		QueryTree root = qt;
		for (Iterator it = rules.iterator(); it.hasNext();) {
			RewriteRule rule = (RewriteRule) it.next();
			if (debug) {
				QueryTreePrinter.printQueryTree(root, log);
				log.log(BasicLevel.DEBUG, "Run the rule: " + rule.getClass().getName());
			}
			root = rule.rewrite(root);
		}
		if (debug) QueryTreePrinter.printQueryTree(root, log);
		return root;
	}

	// IMPLEMENTATION OF THE RuleConfiguration INTERFACE //
	//-----------------------------------------------//
	public void addRule(RewriteRule rule) {
		rules.add(rule);
	}

	public void removeRule(RewriteRule rule) {
		rules.remove(rule);
	}

	public List getRules() {
		return rules;
	}
}
