/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre, S. Chassande-Barrioz
 */
package org.objectweb.medor.optim.lib;

import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.CalculatedField;
import org.objectweb.medor.query.api.FilteredQueryTree;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.Field;
import org.objectweb.medor.filter.api.Expression;
import org.objectweb.medor.filter.api.Operator;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.filter.lib.And;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.Map;
import java.util.HashMap;

/**
 *
 * @author S.Chassande-Barrioz
 */
public class FlattenQueryTreeRule extends BasicRule {

    public FlattenQueryTreeRule() {
        super("FlattenQueryTreeRule");
    }

    public QueryTree rewrite(QueryTree qt) throws MedorException {
        return flatten(null, qt);
    }

    public QueryTree flatten(QueryNode parent, QueryTree qt) throws MedorException {
        if (qt instanceof QueryLeaf) {
            return qt;
        }
        QueryNode qn = (QueryNode) qt;
        QueryTree[] children = qn.getChildren();
        Map localyReplaced = (parent == null ? null : new HashMap());
        while(containsQueryNode(children)) {
            log.log(BasicLevel.DEBUG, "node");
            if (useParent(qn.getQueryFilter())) {
                qn.setQueryFilter(qn.getQueryFilter());
            }
            Field[] fs = qt.getTupleStructure().getFields();
            for (int i=0; i < fs.length; i++) {
                if (fs[i] instanceof PropagatedField) {
                    QueryTreeField childF = (QueryTreeField)
                            ((PropagatedField) fs[i]).getPreviousFields()[0];
                    if (childF.getQueryTree() instanceof QueryLeaf) {
                        continue;
                    }
                    if (childF instanceof PropagatedField) {
                        qn.updatePropagatedField(fs[i].getName(),
                                new QueryTreeField[]{(QueryTreeField)
                                ((PropagatedField) childF).getPreviousFields()[0]});
                    } else if (childF instanceof CalculatedField) {
                        //replace the current field fs[i] by childF
                        qn.replace((QueryTreeField) fs[i], childF);
                        if (parent != null) {
                            localyReplaced.put(fs[i], childF);
                        }
                    } else {
                        throw new MedorException("Unmanaged child field: "
                                + childF);
                    }
                } else if (fs[i] instanceof CalculatedField){
                    CalculatedField cf = (CalculatedField) fs[i];
                    if (useParent(cf.getExpression())) {
                        qn.updateCalculatedField(
                                cf.getName(), cf.getExpression());
                    }
                } else {
                    throw new MedorException("Unmanaged Field: " + fs[i]);
                }
            }
            //Replace the usage of replaced fields
            replaceUsage(parent, localyReplaced);

            //fetch the filter of my children
            boolean modified = false;
            Expression e = qn.getQueryFilter();
            for(int i=0; i<children.length; i++) {
                if (children[i] instanceof FilteredQueryTree){
                    Expression filter = ((FilteredQueryTree) children[i])
                        .getQueryFilter();
                    if (filter != null) {
                        if (e == null) {
                            e = filter;
                        } else {
                            e = new And(e, filter);
                        }
                        modified = true;
                    }
                }
            }
            if (modified) {
                qn.setQueryFilter(e);
            }
            children = qn.getChildren();
        }
        return qt;
    }

    private boolean containsQueryNode(QueryTree[] qts) {
        for(int i=0; i < qts.length; i++) {
            if (qts[i] instanceof QueryNode) {
                return true;
            }
        }
        return false;
    }
    private boolean useParent(Expression e) {
        if (e instanceof Operator) {
            int size = ((Operator) e).getOperandNumber();
            boolean res = false;
            for(int i = 0; i< size; i++) {
                res |= useParent(((Operator) e).getExpression(i));
            }
            return res;
        } else if (e instanceof FieldOperand) {
            FieldOperand fo = (FieldOperand) e;
            QueryTreeField f = (QueryTreeField) fo.getField();
            if (f.getQueryTree() instanceof QueryLeaf) {
                return false;
            }
            if (f instanceof PropagatedField) {
                fo.setField(((PropagatedField) f).getPreviousFields()[0]);
                return true;
            } else {
                log.log(BasicLevel.ERROR, "Unmanaged FieldOperand: " + f);
            }
        }
        return false;
    }

    private void replaceUsage(QueryNode qn, Map old2neo) throws MedorException {
        if (qn == null) {
            return;
        }
        Field[] fs = qn.getTupleStructure().getFields();
        for(int i=0; i < fs.length; i++) {
            if (fs[i] instanceof PropagatedField) {
                PropagatedField pf = (PropagatedField) fs[i];
                QueryTreeField parent = (QueryTreeField)
                        pf.getPreviousFields()[0];
                QueryTreeField neo = (QueryTreeField) old2neo.get(parent);
                if (neo != null) {
                    qn.updatePropagatedField(fs[i].getName(),
                            new QueryTreeField[]{neo});
                }
            } else if (fs[i] instanceof CalculatedField) {
                CalculatedField cf = (CalculatedField) fs[i];
                if (replace(cf.getExpression(), old2neo)) {
                    qn.updateCalculatedField(
                            cf.getName(), cf.getExpression());
                }
            }
        }
    }

    private boolean replace(Expression e, Map old2neo) {
        if (e instanceof Operator) {
            int size = ((Operator) e).getOperandNumber();
            boolean res = false;
            for(int i = 0; i< size; i++) {
                res |= replace(((Operator) e).getExpression(i), old2neo);
            }
            return res;
        } else if (e instanceof FieldOperand) {
            FieldOperand fo = (FieldOperand) e;
            Field f = (Field) old2neo.get(fo.getField());
            if (f != null) {
                fo.setField(f);
                return true;
            }
        }
        return false;
    }
}
