/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.optim.lib;

import org.objectweb.medor.api.ExpressionException;
import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.QueryNodeException;
import org.objectweb.medor.api.TupleStructure;
import org.objectweb.medor.filter.api.BinaryOperator;
import org.objectweb.medor.filter.api.Expression;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.filter.api.UnaryOperator;
import org.objectweb.medor.filter.api.Operator;
import org.objectweb.medor.optim.api.ExecPlanGenerator;
import org.objectweb.medor.query.api.CalculatedField;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.TCQueryLeaf;
import org.objectweb.medor.tuple.lib.GeneralTupleLoader;

public class IndexesGenerator implements ExecPlanGenerator {

	public QueryTree transform(QueryTree query) throws MedorException {
		// generate indexes and TupleLoaders
		generateIndexes(query);
		return query; // We don't clone it...
	}

	private void generateIndexes(QueryTree query) throws MedorException {
		if (query instanceof QueryNode) {
			QueryNode qn = (QueryNode) query;
			setFilterIndexes(qn.getQueryFilter(), qn.getChildren());
			// setting a TupleLoader
			qn.setTupleLoader(
				new GeneralTupleLoader(
					computesFieldIndexes(qn.getTupleStructure(), qn.getChildren()),
					query.getTupleStructure()));
			QueryTree[] children = qn.getChildren();
			for (int cpt = 0; (cpt < children.length); cpt++) {
				generateIndexes(children[cpt]);
			}
			// This case appears when MedorTC queryLeaf
		}
		else if (query instanceof TCQueryLeaf) {
			// At first indexes of the Fields of the TupleStructure
			TCQueryLeaf tcLeaf = (TCQueryLeaf) query;
			TupleStructure ts = tcLeaf.getTupleStructure();
			boolean stop = false;
			int cpt = 0;
			int[] indexes = new int[ts.getSize()];
			while (cpt < ts.getSize() && !stop) {
				indexes[cpt] =
					tcLeaf.getTupleCollection().getMetaData().getFieldRank(
						tcLeaf.getTupleCollection().getMetaData().getField(
							ts.getField(cpt + 1).getName()
						)
					);
				cpt++;
			}
			tcLeaf.setTupleLoader(
				new GeneralTupleLoader(indexes, query.getTupleStructure()));

			// Then the expression filter fieldOperand indexes
			setTCQLFilterIndexes(tcLeaf.getQueryFilter(), tcLeaf);
		}
	}

	public int[] computesFieldIndexes(TupleStructure ts, QueryTree[] nodes)
		throws MedorException {
		Field[] ancestor = null;
		Field current = null;
		QueryTree qt = null;

		//indexs will contain ranges off attribute to project in the
		// concatenated tuple
		int[] indexes = new int[ts.getSize()];
		for (int i = 1; i <= ts.getSize(); i++) {
			try {
				current = ts.getField(i);
			}
			catch (MedorException me) {
				// Impossible case
			}
			if (current instanceof PropagatedField) {
				ancestor = ((PropagatedField) current).getPreviousFields();
				for (int j = 0; j < ancestor.length; j++) {
					int rank = 0;
					int cpt = 0;
					// We will take the first QueryTree
					for (cpt = 0;
						 ((cpt < nodes.length) &&
						!containsField(nodes[cpt], ancestor[j]));
						 cpt++) {
						rank = rank + nodes[cpt].getTupleStructure().getSize();
					}
					// The last node witch matches the field i
					if (cpt < nodes.length) {
						indexes[i - 1] = rank + nodes[cpt].getTupleStructure().getFieldRank(nodes[cpt].getTupleStructure().getField(ancestor[j].getName()));
					}
					else {
						throw new QueryNodeException(
							"The previous field " + ancestor[j].getName()
							+ " must have a QueryTree which matches with one of children");
					}
				}
			}
			else if (current instanceof CalculatedField) {
				// Setting the field operand indexes
				setFilterIndexes(((CalculatedField) current).getExpression(), nodes);
				indexes[i - 1] = -1; // means a calculated field
			}
			else
				throw new QueryNodeException(
					"QueryNode fields error");
		}
//        for (int cpt = 0; (cpt < indexes.length); cpt++) {
//            System.out.println("les indexes " + indexes[cpt]);
//        }
		return indexes;
	}

	/**
	 * This method inherited from QueryIndexGenerator.
	 * It checks that all Fields in the Expression are present in the array
	 * of QueryTree and sets yhe indexs of the fieldOperand.
	 * @param e Expression filter
	 * @param qts array of QueryTree
	 */
	public void setFilterIndexes(Expression e, QueryTree[] qts)
		throws MedorException {
		if (e == null)
			return;
		e.compileExpression();// Compiling the QueryFilter Expression
		if (e instanceof Operator) {
			for (int i = 0; i<((Operator) e).getOperandNumber(); i++) {
				setFilterIndexes(((Operator) e).getExpression(i), qts);
			}
		}
		else if (e instanceof FieldOperand) {
			int i;
			/**
			 * Stops if the QueryTree of the Field is found in the array.
			 * In this case, it stops with i equals to the rank of the
			 * QueryTree in the array.
			 */
			int rank = 0;
			for (i = 0;
				 i < qts.length &&
				!containsField(qts[i], ((FieldOperand) e).getField());
				 rank = rank + qts[i].getTupleStructure().getSize(), i++)
				;

			/**
			 * All qts have been looked at.
			 * If the QueryTree has been found, i != length.
			 * Otherwise, the QueryTree of the Field has not been found and
			 * i equals length.
			 */

			if (i != qts.length) {
				((FieldOperand) e).setIndex(
					rank + qts[i].getTupleStructure().getFieldRank(qts[i].getTupleStructure().getField(((FieldOperand) e).getField().getName())));
			}
			else {
				throw new ExpressionException(
					"Field not found in the subQueryTree: "
					+ ((FieldOperand) e).getField().getName());
			}
		}
	}

	private void setTCQLFilterIndexes(Expression e, TCQueryLeaf tcLeaf) throws MedorException {
		if (e == null)
			return;

		e.compileExpression();// Compiling QueryFilter Expression
		if (e instanceof Operator) {
			for (int i = 0; i<((Operator) e).getOperandNumber(); i++) {
				setTCQLFilterIndexes(((Operator) e).getExpression(i), tcLeaf);
			}
		}
		else if (e instanceof FieldOperand) {
			((FieldOperand) e).setIndex(
				tcLeaf.getTupleStructure().getFieldRank(
					tcLeaf.getTupleStructure().getField(
						((FieldOperand) e).getField().getName()
					)
				)
			);
		}
	}

	private boolean containsField(QueryTree qt, Field field) {
		return qt.getTupleStructure().contains(field.getName());
	}
}
