/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */

package org.objectweb.medor.optim.rdb;

import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TupleStructure;
import org.objectweb.medor.datasource.api.DataStore;
import org.objectweb.medor.filter.api.BinaryOperator;
import org.objectweb.medor.filter.api.Expression;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.filter.api.Operand;
import org.objectweb.medor.filter.api.Operator;
import org.objectweb.medor.filter.api.UnaryOperator;
import org.objectweb.medor.filter.jorm.lib.CompositePName;
import org.objectweb.medor.filter.jorm.lib.SinglePName;
import org.objectweb.medor.filter.lib.And;
import org.objectweb.medor.filter.lib.MemberOf;
import org.objectweb.medor.optim.api.RewriteRule;
import org.objectweb.medor.optim.jorm.JormRule;
import org.objectweb.medor.query.api.CalculatedField;
import org.objectweb.medor.query.api.FilteredQueryTree;
import org.objectweb.medor.query.api.NestedField;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.rdb.api.QualifiedTable;
import org.objectweb.medor.query.rdb.api.RdbExpField;
import org.objectweb.medor.query.rdb.api.RdbExpQueryLeaf;
import org.objectweb.medor.query.rdb.lib.BasicQualifiedTable;
import org.objectweb.medor.query.rdb.lib.BasicRdbExpQueryLeaf;

import java.util.ArrayList;

/**
 * This class groups together as a single RbdExpQueryLeaf the RdbQueryLeaves
 * working on the same data store.
 * TODO: Known limitation : This rule does not reorganize a QueryTree
 * Another rule should be developed to put close together leaves working on the
 * same data store.
 */
public class GroupSameDBRule implements RewriteRule {
	/**
	 * The RdbQuery class represents the result of rewriting a QueryTree.
	 * It has an array of RdbExpFields (its fields), an array of QualifiedTables
	 * (its tables), an Expression working on those Fields, an associated
	 * DataStore and a boolean indicating whether it is a query leaf, or whether
	 * it is the result of merging several QueryLeaves.
	 */
	protected class RdbQuery {
		public ArrayList fields;
		public ArrayList tables;
		public Expression exp;
		public DataStore ds;
		/**
		 * tableIds is the current list of table names (without alias) and
		 * table alias names.
		 */
		public ArrayList tableIds;
		public boolean hasSubQuery;
		public ArrayList subQueries;
	}

	/**
	 * This class represents the result of splitting the children of
	 * a given QueryTree qt in presence of MemberOf operators
	 * in qt's query filter.
	 * <p>The splitting process is performed by the splitQueries and
	 * recurseSplit methods.
	 */
	private class SplitQueries {
		/**
		 * An ArrayList of QueryTrees corresponding to the left part of the
		 * MemberOf.
		 */
		public ArrayList outerQts;
		/**
		 * Inner QueryTree: it is an ArrayList of ArrayLists.
		 * <p>There is one such ArrayList for each subquery. Indeed, if there
		 * are several MemberOf operators in the QueryTree's filter, MEDOR has
		 * to generate as many subqueries as there are MemberOf operators. Each
		 * such ArrayList may contain several QueryTrees, namely in case of a
		 * multiple MemberOf. For example, if we have (a, b) MemberOf (c, d),
		 * and c and d are FieldOperands of different QueryTrees, then there
		 * will be an ArrayList containing the two QueryTrees for c and d.s
		 */
		public ArrayList subQts;
	}

	/**
	 * Implementation of the RewriteRule interface.
	 * <p>It groups together the RdbQueryLeaves if they are on the same data
	 * store.
	 * @param qt the QueryTree to be rewritten by this rule.
	 * @return the rewritten QueryTree
	 * @throws MedorException if all RdbQueryLeaves are not on the same data
	 * store.
	 * @see RewriteRule
	 */
	public QueryTree rewrite(QueryTree qt) throws MedorException {
		//TODO propagate the distinct and order by clauses
		RdbQuery rq = recurseRewrite(qt);
		debug("end of recurse");
		if (rq != null) {
			debug("building RdbLeaf");
			return buildRdbLeaf(rq, qt);
		}
		else
			return qt;
	}

	/**
	 * Performs the transformation of a QueryTree.
	 * <p>It performs recursive calls on the QueryTree's schildren.
	 * @param qt is the QueryTree to rewrite.
	 * @return an RdbQuery corresponding to the rewriting.
	 */
	private RdbQuery recurseRewrite(QueryTree qt) throws MedorException {
		//debug(qt.getName());
		if (qt instanceof RdbExpQueryLeaf) {
			//builds the RdbQuery associated to the Queryleaf
			debug("RdbExpQueryLeaf: START " + qt.getName());
			RdbQuery rq = new RdbQuery();
			Field[] lf = qt.getTupleStructure().getFields();
			rq.fields = new ArrayList(lf.length);
			for (int i = 0; i < lf.length; i++) {
				rq.fields.add(lf[i]);
			}
			QualifiedTable[] lqt = ((RdbExpQueryLeaf) qt).getQualifiedTables();
			rq.tables = new ArrayList(lqt.length);
            rq.tableIds = new ArrayList(lqt.length);
			for (int i = 0; i < lqt.length; i++) {
				rq.tables.add(lqt[i]);
				if (lqt[i].getAliasName() != null)
					rq.tableIds.add(lqt[i].getAliasName());
				else
					rq.tableIds.add(lqt[i].getTableName());
			}
			rq.exp = ((RdbExpQueryLeaf) qt).getQueryFilter();
			rq.ds = ((RdbExpQueryLeaf) qt).getDataStore();
			debug("RdbExpQueryLeaf: END " + qt.getName());
			return rq;
		}
		else if (qt instanceof QueryLeaf) {
			debug("QueryLeaf: " + qt);
			return null;
		}
		//else, it is a QueryNode
		else if (memberOfInExpression(((QueryNode) qt).getQueryFilter())) {
			//qt is a QueryNode with memberOf
			debug("QueryNode with MemberOf: " + qt);
			SplitQueries split = splitQueries(qt);
			debug("splitQueries done");
			debug("Structure of split: "
				+ "Left: " + split.outerQts.size()
				+ " Right: " + split.subQts.size()
			);
			//deal with the top QueryTree
			RdbQuery workRq =
				exploreQueryTrees((QueryTree[]) split.outerQts.toArray(
					new QueryTree[0])
				);
			debug("recursive exploreQueryTrees done");
			workRq.hasSubQuery = true;
			if (noCalculation(qt)) {
				debug("noCalculation: => continue");
				//add exp of current node to the workRq
				if (workRq.exp == null) {
					workRq.exp = ((QueryNode) qt).getQueryFilter();
				}
				else {
					workRq.exp =
						new And(workRq.exp, ((QueryNode) qt).getQueryFilter());
				}
				//construct subqueries
				workRq.subQueries = new ArrayList();
				debug("number of subqueries: " + split.subQts.size());
				for (int i = 0; i < split.subQts.size(); i++) {
					RdbQuery innerRq =
						exploreQueryTrees((QueryTree[])
						((ArrayList) split.subQts.get(i)).toArray(
							new QueryTree[0])
						);
					workRq.subQueries.add(innerRq);
				}

			}
			else {
				createLeafAndUpdateLinks(workRq, qt);
				workRq = null;
			}
			return workRq;
		}
		else {
			//qt is a QueryNode without memberOf
			debug("QueryNode without MemberOf: " + qt);
			QueryTree[] children = ((QueryNode) qt).getChildren();
			//recurse on the Children
			RdbQuery workRq = exploreQueryTrees(children);
			if (noCalculation(qt)) {
				debug("noCalculation: => continue");
				//add exp of current node to the workRq
				if (workRq.exp == null) {
					workRq.exp = ((QueryNode) qt).getQueryFilter();
				}
				else {
					workRq.exp =
						new And(workRq.exp, ((QueryNode) qt).getQueryFilter());
				}
			}
			else {
				createLeafAndUpdateLinks(workRq, qt);
				workRq = null;
			}
			debug("QueryNode: END");
			return workRq;
		}
	}

	/**
	 * Breaks the recursion when the QueryTree has calculated fields.
	 * <p>In this case, build a RdbExpQueryLeaf for the children, but not for
	 * the top QueryTree which remains as it is, and for which the links to the
	 * fields are updated to that of the newly created QueryLeaf.
	 * @param workRq the RdbQuery structure containing the children
	 * @param qt the top QueryTree
	 * @throws MedorException in the case of a NestedField.
	 */
	private void createLeafAndUpdateLinks(RdbQuery workRq, QueryTree qt)
		throws MedorException {
		debug("Has calculated field: => stop, create RdbExpQL and link");
		RdbExpQueryLeaf ql = buildRdbLeaf(workRq, qt);
		// Upgrade the link between qt and the workRq

		//update fields
		Field[] fs = qt.getTupleStructure().getFields();
		TupleStructure ts = ql.getTupleStructure();
		for (int i = 0; i < fs.length; i++) {
			if (fs[i] instanceof PropagatedField) {
				if (ts.contains(fs[i].getName())) {
					debug("Replace in the PropagatedField " + fs[i].getName());
					((QueryNode) qt).updatePropagatedField(
						fs[i].getName(),
						new QueryTreeField[]{
							(QueryTreeField) ts.getField(
								fs[i].getName())});
				}
				else
					debug("Do not replace in the PropagatedField " + fs[i].getName());
			}
			else if (fs[i] instanceof CalculatedField) {
				debug("Try to replace in the CalculatedField " + fs[i].getName());
				Expression e =
					((CalculatedField) fs[i]).getExpression();
				if (replaceInFilter(e, ts))
					((QueryNode) qt).updateCalculatedField(
						fs[i].getName(), e);
			}
			else if (fs[i] instanceof NestedField) {
				debug("Replace in the NestedField " + fs[i].getName());
				throw new MedorException("Unmanaged Nested Field");
			}
		}
		//update filter
		if (qt instanceof FilteredQueryTree) {
			debug("Try to replace in the filter ");
			Expression e = ((FilteredQueryTree) qt).getQueryFilter();
			if (replaceInFilter(e, ts))
				((FilteredQueryTree) qt).setQueryFilter(e);
		}
	}

	/**
	 * Constructs an RdbQuery corresponding to the recursive exploration and
	 * merging of a given list of QueryTrees.
	 * @param qts the input list of QueryTrees to be explored
	 * @return the RdbQuery constructed by exploring and merging the QueryTrees.
	 * @throws MedorException if some QueryTrees are not on the same data store.
	 */
	private RdbQuery exploreQueryTrees(QueryTree[] qts) throws MedorException {
		boolean first = true;
		RdbQuery workRq = null;
		for (int i = 0; i < qts.length; i++) {
			RdbQuery currentRq = recurseRewrite(qts[i]);
			if (first) {
				debug("first");
				workRq = currentRq;
				first = false;
			}
			else if (workRq.ds != null &&
				workRq.ds.isSameAs(currentRq.ds)) { //same datastore
				debug("Same store => merging");
				mergeRdbQueries(workRq, currentRq);
			}
			else {
				throw new MedorException(
					"Impossible to group two nodes with different data stores");
			}
		}
		return workRq;
	}

	/**
	 * Replaces the fields referenced in the expression by the fields of the
	 * tuple structure. This the field name that is used to match the fields
	 * @param e is the expression where the Field referenced by FieldOperand
	 * must be replaced.
	 * @param ts is the tuple structure that contains the fields that replace
	 * the expression fields.
	 * @return true if the expression was modified, otherwise false.
	 * @throws MedorException when the data structure is wrong.
	 */
	private boolean replaceInFilter(Expression e, TupleStructure ts)
		throws MedorException {
		ArrayList al = new ArrayList();
		al.add(e);
		boolean isModified = false;
		while (!al.isEmpty()) {
			Expression ex = (Expression) al.remove(0);
			if (ex instanceof Operator) {
				for (int i = 0; i < ((Operator) ex).getOperandNumber(); i++) {
					al.add(((Operator) ex).getExpression(i));
				}
			}
			else if (ex instanceof FieldOperand) {
				FieldOperand fo = (FieldOperand) ex;
				if (ts.contains(fo.getField().getName())) {
					debug("Replace in the FieldOperand: " + fo.getField().getName());
					fo.setField(ts.getField(fo.getField().getName()));
					isModified = true;
				}
				else
					debug("Do not replace in the FieldOperand: " + fo.getField().getName());
			}
		}
		return isModified;
	}

	/**
	 * Checks whether a QueryTree contains calculated fields.
	 * <p>
	 * A QueryTree is elligible for being rewritten as an RdbQueryLeaf if all
	 * its children have been rewritten, and if it does not have any calculated
	 * fields.
	 * @param qt the QueryTree for verification
	 * @return true if the QueryTree does not contain calculated fields.
	 */
	private boolean noCalculation(QueryTree qt) {
		debug("Testing for calculated fields: ");
		Field[] qtf = qt.getTupleStructure().getFields();
		for (int i = 0; i < qtf.length; i++) {
			if (qtf[i] instanceof CalculatedField) {
				debug("found");
				return false;
			}
		}
		debug("not found");
		return true;
	}

	/**
	 * Merges two RdbQueries.
	 * <p>It has to rename the QualifiedTables if they have the same name.
	 * @param into the RdbQuery into which to merge the other RdbQuery
	 * @param other the RdbQuery to be merged into the first RdbQuery (into
	 * parameter).
	 */
	private void mergeRdbQueries(RdbQuery into, RdbQuery other) {
		for (int i = 0; i < other.fields.size(); i++) {
			into.fields.add(other.fields.get(i));
		}
		QualifiedTable qt;
		for (int i = 0; i < other.tables.size(); i++) {
			qt = (QualifiedTable) other.tables.get(i);
			if (into.tables.contains(qt)) {
				//The actual same object is used in another query tree. It needs
				//to be replaced, and all references to qt in the fields
				QualifiedTable qt1 =
					new BasicQualifiedTable(qt.getTableName(),
						qt.getAliasName());
				for (int j = 0; j < other.fields.size(); j++) {
					if (((RdbExpField) other.fields.get(j)).getTable() == qt)
						((RdbExpField) other.fields.get(j)).setTable(qt1);
				}
				qt = qt1;
			}
			//checks aliases: if it has an alias, that alias must not already
			//exist
			if (qt.getAliasName() != null) {
				if (into.tableIds.contains(qt.getAliasName()))
					qt.setAliasName(generateAlias(into.tableIds,
						qt.getAliasName()));
			}
			else if (into.tableIds.contains(qt.getTableName()))
				qt.setAliasName(generateAlias(into.tableIds,
					qt.getTableName()));
			into.tables.add(qt);
		}
		if (into.exp == null) //into.exp is null. Take other.exp
			into.exp = other.exp;
		else if (other.exp != null) //into.ext and other.exp are not null
			into.exp = new And(into.exp, other.exp);
	}

	/**
	 * Generates an alias from a root name given existing table identifiers.
	 * <p>The generated alias is new, i.e., it should not already be in the
	 * existing table identifiers.
	 * <p>The new alias is added to the existing table identifiers.
	 * @param tableIds the existing table identifiers.
	 * @param root the String root for building the new identifier.
	 * @return an new alias based on root.
	 */
	private String generateAlias(ArrayList tableIds, String root) {
		root += "_";
		int i = 1;
		while (tableIds.contains(root + i)) {
			i++;
		}
		root += i;
		tableIds.add(root);
		return root;
	}

	/**
	 * Parses an Expression to detect the presence of a MemberOf operator.
	 * @param e is the Expression to be parsed.
	 * @return true if the Expression contains a MemberOf, NotMemberof, IsEmpty
	 * or IsNotEmpty operator.
	 */
	private boolean memberOfInExpression(Expression e) {
		debug("entering memberOfInExpression");
		if (e instanceof MemberOf)
			return true;
		else if (e instanceof Operator) {
			Operator op = (Operator) e;
			boolean found = false;
			for (int i = 0; i < op.getOperandNumber(); i++) {
				if (memberOfInExpression(op.getExpression(i))) {
					found = true;
					break;
				}
			}
			return found;
		}
		return false;
	}

	/**
	 * Constructs an RdbExpQueryLeaf from the result of grouping
	 * the RdbExpQueryLeaves of a QueryTree, expressed as an RdbQuery.
	 * @param rq is the RdbQuery representation of the QueryTree for which the
	 * RdbExpQueryLeaves have to be grouped.
	 * @return the RdbExpQueryLeaf obtained when grouping the RdbExpQueryLeaves
	 * of the original QueryTree
	 */
	private RdbExpQueryLeaf buildRdbLeaf(RdbQuery rq, QueryTree qt)
		throws MedorException {
		RdbExpQueryLeaf ql;
		if (rq.hasSubQuery) {
			ql = new BasicRdbExpQueryLeaf(
				rq.ds,
				(QualifiedTable[]) rq.tables.toArray(new QualifiedTable[0]),
				qt.getName()
			);
			ql.setQueryFilter(rq.exp);
			for (int i = 0; i < rq.fields.size(); i++) {
				ql.addRdbField((RdbExpField) rq.fields.get(i));
			}
			for (int i = 0; i < rq.subQueries.size(); i++) {
				BasicRdbExpQueryLeaf subql = new BasicRdbExpQueryLeaf(
					rq.ds,
					(QualifiedTable[])
					((RdbQuery) rq.subQueries.get(i)).tables.toArray(
						new QualifiedTable[0]),
					qt.getName());
				replaceInFilter(ql.getQueryFilter(), subql.getTupleStructure());
			}
		}
		else {
			ql = new BasicRdbExpQueryLeaf(
				rq.ds,
				(QualifiedTable[]) rq.tables.toArray(new QualifiedTable[0]),
				qt.getName());
			//System.out.println(ExpressionPrinter.e2str(rq.exp));
			ql.setQueryFilter(rq.exp);
			for (int i = 0; i < rq.fields.size(); i++) {
				ql.addRdbField((RdbExpField) rq.fields.get(i));
			}
		}

		return ql;
	}

	/**
	 * For a given QueryTree with an identified MemberOf operator in its query
	 * filter, constructs a SplitQueries.
	 * @param qt the QueryTree to be split
	 * @return the new SplitQueries
	 * @see SplitQueries
	 */
	private SplitQueries splitQueries(QueryTree qt) {
		SplitQueries split = new SplitQueries();
		split.outerQts = new ArrayList();
		split.subQts = new ArrayList();
		Expression e = ((QueryNode) qt).getQueryFilter();
		return recurseSplit(split, e);
	}

	/**
	 * Performs the actual work of splitting a QueryTree starting from its
	 * expression.
	 * @param split the input SplitQueries object
	 * @param e the Expression from which to build the SplitQueries
	 * @return the SplitQueries object constructed from the Expression
	 * @see SplitQueries
	 */
	private SplitQueries recurseSplit(SplitQueries split, Expression e) {
		if (e instanceof FieldOperand) {
			QueryTreeField f = (QueryTreeField) ((FieldOperand) e).getField();
			QueryTree qtf = f.getQueryTree();
			if (!split.outerQts.contains(qtf))
				split.outerQts.add(qtf);
		}
		else if ((e instanceof CompositePName) ||
			(e instanceof SinglePName) ||
			(e instanceof Operand)) {
			return split;
		}
		else if (e instanceof MemberOf) {
			debug("splitting MemberOf");
			MemberOf mo = (MemberOf) e;
			int moSize = mo.getOperandNumber() / 2;
			//iterate on all pairs of Fields
			for (int i = 0; i < moSize; i++) {
				if (((MemberOf) e).getExpression(i) instanceof FieldOperand) {
					QueryTreeField fLeft =
						(QueryTreeField)
						((FieldOperand)
						((MemberOf) e).getExpression(i)).getField();
					QueryTree qtfLeft = fLeft.getQueryTree();
					debug("Left node is " + qtfLeft.getName());
					if (!split.outerQts.contains(qtfLeft))
						split.outerQts.add(qtfLeft);
				}
			}
			ArrayList newSub = null;
			for (int i = 0; i < moSize; i++) {
				QueryTreeField fRight =
					(QueryTreeField)
					((FieldOperand)
					((MemberOf) e).getExpression(i + moSize)).getField();
				QueryTree qtfRight = fRight.getQueryTree();
				debug("Right node is " + qtfRight.getName());
				//remove the right QT from outer QT if it is in it
				if (split.outerQts.contains(qtfRight))
					split.outerQts.remove(qtfRight);
				//search whether right QT already in subQts or not
				//if not, add it
				boolean isInSub = false;
				for (int j = 0; j < split.subQts.size(); j++) {
					if (((ArrayList) split.subQts.get(j)).contains(qtfRight)) {
						isInSub = true;
						break;
					}
				}
				debug("found in subqueries:" + isInSub);
				if (!isInSub) {
					if (newSub == null)
						newSub = new ArrayList(1);
					newSub.add(qtfRight);
				}
			}
			if (newSub != null)
				split.subQts.add(newSub);
			return split;
		}
		else if (e instanceof BinaryOperator) {
			BinaryOperator bo = (BinaryOperator) e;
			return recurseSplit(
				recurseSplit(split, bo.getLeftExpression()),
				bo.getRightExpression()
			);
		}
		else if (e instanceof UnaryOperator) {
			UnaryOperator uo = (UnaryOperator) e;
			return recurseSplit(split, uo.getExpression());
		}
		else if (e instanceof Operator) {
			Operator op = (Operator) e;
			for (int i = 0; i < op.getOperandNumber(); i++) {
				split = recurseSplit(split, op.getExpression(i));
			}
			return split;
		}

		return split;
	}

	private static final void debug(String msg) {
		//System.out.println(msg);
	}
}
