/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.medor.query.jorm.lib;

import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.CalculatedField;
import org.objectweb.medor.query.api.FilteredQueryTree;
import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.jorm.api.JormExtent;
import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.filter.api.Expression;
import org.objectweb.medor.filter.api.Operator;
import org.objectweb.medor.filter.api.ParameterOperand;
import org.objectweb.medor.type.lib.QType;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.ClassMapping;
import org.objectweb.jorm.metainfo.api.PrimitiveElementMapping;
import org.objectweb.jorm.metainfo.api.PrimitiveElement;

import java.util.Collection;
import java.util.ArrayList;
import java.util.List;

/**
 * This helper class provides methods to helper the Jorm/Medor users.
 *
 * @author S.Chassande-Barrioz
 */
public class JormQueryTreeHelper {

    /**
     * Retrieves a Collection of JormExtent instance which are used in a given
     * query tree. This permits to known all JormExtent used in a queries and to
     * assign some parameters in order to replace a JormExtent by a specific
     * query leaf during the optimization step. A JormExtent can be a
     * ClassExtent or a GenClassExtent.
     * @param qt is the querytree instance containing (or not) JormExtent nodes
     * @return a collection of JormExtent instances
     */
    public static Collection getJormExtents(QueryTree qt) {
        if (qt == null)
            return null;
        ArrayList res = new ArrayList();
        ArrayList toTreat = new ArrayList();
        toTreat.add(qt);
        while (!toTreat.isEmpty()) {
            QueryTree current = (QueryTree) toTreat.remove(0);
            if (current instanceof JormExtent) {
                res.add(current);
            } else if (current instanceof QueryNode) {
                QueryTree[] children = ((QueryNode) current).getChildren();
                for (int i = 0; i < children.length; i++) {
                    toTreat.add(children[i]);
                }
            }
        }
        return res;
    }

    /**
     * Retrieves a Collection of ParameterOperand reuiqring in a given query
     * tree which the type is NAMING_CONTEXT.
     * It permits to know all parameters needed at evaluation
     * time for the navigation over classes persitent with Jorm.
     * @param qt is the query tree requiring PNamingContext parameters
     * @return a collection of ParameterOperand instance
     */
    public static Collection getRequiredPNameManagers(QueryTree qt) {
        if (qt == null)
            return null;
        ArrayList res = new ArrayList();
        ArrayList toTreat = new ArrayList();
        toTreat.add(qt);
        while (!toTreat.isEmpty()) {
            QueryTree current = (QueryTree) toTreat.remove(0);
            Field[] fs = qt.getTupleStructure().getFields();
            for (int i = 0; i < fs.length; i++) {
                if (fs[i] instanceof CalculatedField) {
                    getRequiredPNameManagers(
                        ((CalculatedField) fs[i]).getExpression(), res);
                }
            }
            if (current instanceof FilteredQueryTree) {
                getRequiredPNameManagers(
                    ((FilteredQueryTree) current).getQueryFilter(), res);
            }
            if (current instanceof QueryNode) {
                QueryTree[] children = ((QueryNode) qt).getChildren();
                for (int i = 0; i < children.length; i++) {
                    toTreat.add(children[i]);
                }
            }
        }
        return res;
    }

    /**
     * Adds the required ParameterOperand which the type is NAMING_CONTEXT.
     * @param e is the expression requiring PNamingContext parameters
     * @param c is a non null Collection inside which the required
     * PNamingContext parameters must be added.
     */
    private static void getRequiredPNameManagers(Expression e, Collection c) {
        if (e instanceof ParameterOperand) {
            ParameterOperand po = (ParameterOperand) e;
            if (po.getType().getTypeCode() == QType.TYPECODE_NAMING_CONTEXT) {
                c.add(po);
            }
        } else if (e instanceof Operator) {
            Operator op = (Operator) e;
            int nb = op.getOperandNumber();
            for (int i = 0; i < nb; i++) {
                getRequiredPNameManagers(op.getExpression(i), c);
            }
        }
    }

    public static void addPrefetchFields(ClassExtent ext,
                                         QueryTree oldroot,
                                         QueryNode newroot) throws MedorException {
        ClassMapping cm = ((Class) ext.getMetaObject())
            .getClassProject(ext.getProjectName())
            .getMapping(getMapperName(ext.getPMapper().getMapperName()))
            .getClassMapping();
        List pems = new ArrayList(cm.getPrimitiveElementMappings());
        List fields = recurseAPF(ext, oldroot, pems);
        for(int i = 0; i<fields.size(); i++) {
            QueryTreeField qtf = (QueryTreeField) fields.get(i);
            newroot.addPropagatedField(
                qtf.getName(), qtf.getType(), new QueryTreeField[]{qtf});
        }
    }
    /**
     *
     */
    public static void addPrefetchFields(ClassExtent ext,
                                         QueryTree qt) throws MedorException {
        ClassMapping cm = ((Class) ext.getMetaObject())
            .getClassProject(ext.getProjectName())
            .getMapping(getMapperName(ext.getPMapper().getMapperName()))
            .getClassMapping();
        List fields = new ArrayList(cm.getPrimitiveElementMappings());
        recurseAPF(ext, qt, fields);
    }

    private static List recurseAPF(ClassExtent ext,
                                   QueryTree qt,
                                   List pems) throws MedorException {
        if (qt == ext) {
            //found the extent
            //check that the extent has the required fields
            ArrayList result = new ArrayList(pems.size());
            for (int i = 0; i < pems.size(); i++) {
                PrimitiveElement pe = (PrimitiveElement) ((PrimitiveElementMapping) pems.get(i)).getLinkedMO();
                String fn = pe.getName();
                Field f = null;
                if (ext.contains(ext.getFieldName(ext.getName(), fn))) {
                    f = ext.getField(ext.getFieldName(ext.getName(), fn));
                } else {
                    f = ext.addField(pe);
                }
                result.add(f);
            }
            return result;
        } else if (qt instanceof QueryLeaf) {
            return null;
        }
        QueryTree[] children = ((QueryNode) qt).getChildren();
        int i = 0;
        List childfs = null;
        while (i < children.length
            && ((childfs = recurseAPF(ext, children[i], pems)) == null)) {
            i++;
        }
        if (childfs == null) {
            //ClassExtent is not in the current QueryTree
            return null;
        }
        //ClassExtent found and childfs contains the Fields to be propagated
        Field[] qtfs = qt.getTupleStructure().getFields();
        ArrayList result = new ArrayList(pems.size());
        for (int j = 0; j < childfs.size(); j++) {
            QueryTreeField previous = (QueryTreeField) childfs.get(j);
            for (i = 0; i < qtfs.length; i++) {
                if ((qtfs[i] instanceof PropagatedField)
                    && ((PropagatedField) qtfs[i]).getPreviousFields()[0] == previous) {
                    result.add(qtfs[i]);
                    break;
                }
            }
            if (i == qtfs.length) {
                //field not found
                try {
                    result.add(
                        ((QueryNode) qt).addPropagatedField(previous.getName(),
                            previous.getType(),
                            new QueryTreeField[]{previous}
                        )
                    );
                } catch (MedorException e) {
                    //bad luck: name collision
                }
            }
        }
        return result;
    }

    private static String getMapperName(String mn) {
        if (mn == null) {
            return null;
        }
        int idx = mn.indexOf('.');
        if (idx == -1) {
            return mn;
        } else {
            return mn.substring(0, idx);
        }
    }
}
