/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.query.jorm.lib;

import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.metainfo.api.MetaObject;
import org.objectweb.jorm.metainfo.api.Reference;
import org.objectweb.jorm.metainfo.api.TypedElement;
import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.jorm.metainfo.api.NameDef;
import org.objectweb.jorm.metainfo.api.ClassRef;
import org.objectweb.jorm.metainfo.api.Mapping;
import org.objectweb.jorm.metainfo.api.ClassProject;
import org.objectweb.jorm.metainfo.api.ReferenceMapping;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.runtime.JormPathHelper;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.jorm.api.JormField;
import org.objectweb.medor.query.jorm.api.JormExtent;
import org.objectweb.medor.api.MedorException;

import java.util.Hashtable;
import java.util.Iterator;

/**
 * This class represents Fields corresponding to the PName of a class or a
 * Genclass(the NameDef of the class itself) or the the PName associated to a
 * reference field or to a genclass element.
 * The name of such a Field is a constant.
 *
 * @author Alexandre Lefebvre, Sebastien Chassande-Barrioz
 */
public class PNameField implements QueryTreeField, JormField {

    /**
     * In case where this PNameField represents the PName of a Jorm class, this
     * field is not null and references the Jorm Meta object Class.
     */
    protected Class clazz;

    /**
     * In case where this PNameField represents the a reference field, this
     * field is not null and references the Jorm Meta object Reference.
     */
    protected Reference reference;

    /**
     * In case where this PNameField represents GenClass identifier or the
     * genclass element this field is not null and references the Jorm meta
     * object GenClassRef.
     */
    //protected GenClassRef gcr;

    /**
     * If the current PNameField represents a PName of a Jorm class or genclass,
     * this field is equals to 'true'. Otherwise the current PNameField
     * represents a reference field or the genclass element then this field is
     * equals to 'false'.
     */
    protected boolean isClassPName;

    /**
     * This fiedl returns 'true' if the current PNameField has been built to
     * describe the PName identifier of a GenClass or the PName of the GenClass
     * element.
     */
    protected boolean isInGenClass;


    /**
     * Type of the field
     */
    protected PType type;
    protected short valueState;

    /**
     * Name of the field.
     */
    protected String fieldName = null;
    protected QueryTree queryTree = null;

    /**
     * This constructor must be used to built a PNameField associated to a
     * GenClass identifier or the genclass element.
     * @param fieldName is the name which must be given to the field.
     * @param gcr is the jorm meta object which describes the genclass.
     * @param isGenClassPName indicates if this PNameField will represent a
     * genclass identifier(value='true'). Otherwise the this PNameField will
     * represent a genclass element(value='false').
     * @param qt is the QueryTree which is associated to this QueryTree field.
     */
    public PNameField(String fieldName,
                      GenClassRef gcr,
                      boolean isGenClassPName,
                      QueryTree qt) {
        this.fieldName = fieldName;
        queryTree = qt;
        this.reference = gcr;
        isInGenClass = true;
        isClassPName = isGenClassPName;
        type = gcr.getType();
    }

    /**
     * This constructor must be used to built a PNameField associated to a
     * classRef.
     * @param cr is the jorm meta object ClassRef associated to this PNameField.
     * @param qt is the QueryTree which is associated to this QueryTree field.
     */
    public PNameField(String fieldName, ClassRef cr, QueryTree qt) throws MedorException {
        isInGenClass = cr.getParent() instanceof GenClassRef;
        queryTree = qt;
        this.fieldName = fieldName;
        type = cr.getType();
        isClassPName = false;
        reference = cr;
    }

    /**
     * This constructor must be used to built a PNameField associated to a Class
     * identifier or a reference field.
     * @param fn is the name which must be given to the field.
     * @param cl is the jorm meta object Class associated to this PNameField. In
     * case of the class contains a field which the name is equals to @fn then
     * this PNameField represents a reference field.
     * @param qt is the QueryTree which is associated to this QueryTree field.
     */
    public PNameField(String fn, Class cl, QueryTree qt) throws MedorException {
        isInGenClass = false;
        queryTree = qt;
        fieldName = fn;
        TypedElement te = getTypedElement(cl, fieldName);
        //System.out.println("PNameField("+fn + ") te=" + te);
        if (te!=null && te instanceof Reference) {
            type = te.getType();
            isClassPName = false;
            reference = (Reference) te;
        }
        else {
            clazz = cl;
            isClassPName = true;
            MetaObject tmp = clazz.getParent();
            MetaObject par = null;
            while (tmp != null) {
                par = tmp;
                tmp = tmp.getParent();
            }

            try {
                if (clazz.getInheritedClassNumber() == 0) {
                    // no super class, just use the name of the class.
                    type = ((Manager) par)
                        .getPTypeSpace().createPType(clazz.getFQName());
                }
                else {
                    // minimum 1 super class
                    Hashtable table = new Hashtable();
                    clazz.addAllInheritedClass(table);
                    String allNames [][] = new String[table.size()][];
                    int count = 0;

                    Iterator iter = table.values().iterator();
                    while (iter.hasNext()) {
                        String[] names = (String[]) iter.next();
                        allNames[count][0] = names[0];
                        allNames[count] = names;
                    }
                    type = ((Manager) par)
                        .getPTypeSpace().createPType(clazz.getFQName(), allNames);
                }
            }
            catch (Exception e) {
                throw new MedorException(
                    "Impossible to built the type of the class "
                    + clazz.getName(), e);
            }
        }
    }

    /**
     * It returns the type of the field. The type has been calculated a building
     * time.
     */
    public PType getType() {
        return type;
    }

    public short getNullStatus() {
        return valueState;
    }

    public Object clone() {
        return this.clone();
    }

    /**
     * It returns the field name of the PNameField
     */
    public String getName() {
        return fieldName;
    }

    /**
     * It returns 'true' when this PNameField represents the identifier of a jorm
     * class or a jorm genclass.
     */
    public boolean isClassPName() {
        return isClassPName;
    }

    /**
     * It returns the jorm meta object Reference in case of this PNameField
     * represents a reference field in a jorm class.
     */
    public Reference getReference() {
        return reference;
    }

    /**
     * It returns the jorm meta object Class in case of this PNameField
     * represents an identifier of a jorm class.
     */
    public Class getMetaObjectClass() {
        return clazz;
    }

    /**
     * It returns the jorm meta object GenClassRef in case of this PNameField
     * represents a GenClass identifier or a GenClass element.
     */
    public GenClassRef getGenClassRef() {
        return (GenClassRef)reference;
    }

    /**
     * It returns the QueryTree whith which this QueryTreeField is associated.
     */
    public QueryTree getQueryTree() {
        return queryTree;
    }

    /**
     * It returns 'true' if the current PNameField has been built to
     * describe the PName identifier of a GenClass or the PName of the GenClass
     * element.
     */
    public boolean isInGenClass() {
        return isInGenClass;
    }

    public TypedElement getTypedElement() {
        return reference;
    }

    private TypedElement getTypedElement(Class clazz, String fn) {
        int idx = fn.length();
        TypedElement te = null;
        te = clazz.getTypedElement(fn);
        while ((idx = fn.lastIndexOf(".", idx)) != -1 && te == null) {
            te = clazz.getTypedElement(fn.substring(idx + 1, fn.length()));
            idx--;
        }
        return te;
    }

    /**
     * Retrieves the NameDef corresponding to this PNameField
     * @param extent attached to this PNameField
     * @throws MedorException if it is not possible to fetch the Mapping for
     * the project name and the mapper attached to the JormExtent.
     */
    public NameDef getNameDef(JormExtent extent) throws MedorException {
        if (isClassPName) {
            if (isInGenClass) {
                //Pname of the gen class identifier
                return getRefNameDef((GenClassRef) extent.getMetaObject(), extent);
            } else {
                //PName of the class identifier
                return getClassNameDef(extent);
            }
        } else {
            if (isInGenClass) {
                return getElemNameDef(reference, extent);
            } else {
                //PName of the class reference
                return getRefNameDef(reference, extent);
            }
        }
    }

    /**
     * Retrieves the name of the ParameterOperand containing the PNamingContext.
     */
    public String getPNamingContextParameter() {
        if (isClassPName) {
            if (isInGenClass) {
                //Pname of the gen class identifier
                return JormPathHelper.getPath((GenClassRef) reference, false);
            } else {
                //PName of the class identifier
                return JormPathHelper.getPath(clazz);
            }
        } else {
			if (reference instanceof GenClassRef) {
				//reference to a GenClass from a Class, or a GenClass
				return JormPathHelper.getPath((GenClassRef) reference,
						!isInGenClass);
			}
			else
				//reference to a Class from a Class, or a GenClass
				return JormPathHelper.getPath((ClassRef) reference);
         }
    }

    /**
     * Retrieves the Jorm meta object of the Class represented by this
     * PNameField. If the parameter is
     * a Class then the same value is returned. If the parameter is a ClassRef
     * or a GenClassRef then referenced Class is returned. Otherwise a null
     * value is returned.
     */
    public MetaObject getTargetedMetaObject() {
        if (isInGenClass) {
            if (isClassPName) {
                return reference;
            } else {
                GenClassRef gcr = (GenClassRef) reference;
                if (gcr.isClassRef()) {
                    return gcr.getClassRef().getMOClass();
                } else if (gcr.isGenClassRef()) {
                    return gcr.getGenClassRef();
                } else {
                    return null;
                }
            }
        } else {
            if (isClassPName) {
                return clazz;
            } else {
                return reference;
            }
        }
    }
    /**
     * Retrieves the NameDef representing the identifier of a GenClass/
     * @param gcr is the jorm meta object representing the GenClass
     * @param extent is the GenClassExtent corresponding to the 'gcr'
     * @throws MedorException if it is not possible to fetch the Mapping for
     * the project name and the mapper attached to the JormExtent.
     */
    private static NameDef getRefNameDef(GenClassRef gcr,
                                 JormExtent extent) throws MedorException {
        return (NameDef) getMapping(extent)
                .getGenClassMapping(gcr.getGenClassId())
                .getIdentifierMapping().getLinkedMO();
    }


    /**
     * Retrieves the NameDef representing the element of a GenClass/
     * @param r is the jorm meta object representing the reference element of
     * the GenClass
     * @param extent is the GenClassExtent corresponding to the 'gcr'
     * @throws MedorException if it is not possible to fetch the Mapping for
     * the project name and the mapper attached to the JormExtent.
     */
    private static NameDef getElemNameDef(Reference r,
                                 JormExtent extent) throws MedorException {
        return (NameDef) getMapping(extent)
                .getGenClassMapping(((GenClassRef) r.getParent()).getGenClassId())
                .getReferenceMapping().getLinkedMO();
    }

    /**
     * Retrieves the NameDef representing the reference to a Class. This
     * reference is hosted by a Class (not a GenClass see getElemNameDef)
     * @param r is the jorm meta object representing the class reference
     * @param extent is the ClassExtent corresponding to the 'cr'
     * @throws MedorException if it is not possible to fetch the Mapping for
     * the project name and the mapper attached to the JormExtent.
     */
    private static NameDef getRefNameDef(Reference r,
                                 JormExtent extent) throws MedorException {
/*
		System.out.println("Extent class mapping " + getMapping(extent).getClassMapping());
		Iterator it = getMapping(extent).getClassMapping().getReferenceMappings().iterator();
		while (it.hasNext()) {
			System.out.println("Reference rule name " + ((ReferenceMapping)it.next()).getRuleName());
		}
		System.out.println("+ reference mapping " + getMapping(extent).getClassMapping().getReferenceMapping(r.getName()));
		System.out.println("+ rule name " + getMapping(extent).getClassMapping().getReferenceMapping(r.getName()).getRuleName());
		System.out.println("+ linkedMO " + getMapping(extent).getClassMapping().getReferenceMapping(r.getName()).getLinkedMO());
*/
		return (NameDef) getMapping(extent).getClassMapping()
                .getReferenceMapping(r.getName()).getLinkedMO();
    }

    /**
     * Retrieves the Jorm meta object representing a Class. If the parameter is
     * a Class then the same value is returned. If the parameter is a ClassRef
     * or a GenClassRef then the hosting Class is returned. Otherwise a null
     * value is returned.
     */
    private static Class getClass(MetaObject mo) {
        if (mo instanceof Class) {
            return (Class) mo;
        } else if (mo instanceof Package || mo instanceof Manager) {
            return null;
        } else {
            return getClass(mo.getParent());
        }
    }

    /**
     * It retrieves the Mapping of a JormExtent.
     * @param extent represents a persitent object which the Mapping is expected
     * @throws MedorException if no mapping exists for the project name and the
     * mapper specified by the extent.
     */
    private static Mapping getMapping(JormExtent extent) throws MedorException {
        Class c = getClass(extent.getMetaObject());
        ClassProject cp = c.getClassProject(extent.getProjectName());
        if (cp == null) {
            throw new MedorException("No classproject found for the class "
                    + c.getFQName() + " and  the project "
                    + extent.getProjectName());
        }
        String mapperName = extent.getPMapper().getMapperName();
        int idx =  mapperName.indexOf('.');
        Mapping m = cp.getMapping(idx == -1
                ? mapperName
                : mapperName.substring(0, idx));
        if (m == null) {
            throw new MedorException("No mapping found for the class "
                    + c.getFQName() + ",  the project " + extent.getProjectName()
                    + " and  the mapper " + mapperName);
        }
        return m;
    }

    /**
     * Retrieves the NameDef representing the identifier of a Class.
     * @param extent is the ClassExtent representing the class
     * @throws MedorException if it is not possible to fetch the Mapping for
     * the project name and the mapper attached to the JormExtent.
     */
    private static NameDef getClassNameDef(JormExtent extent) throws MedorException {
        return (NameDef) getMapping(extent).getClassMapping()
                .getIdentifierMapping().getLinkedMO();
    }

}
