/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 * 
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.query.lib;

import org.objectweb.jorm.type.api.PType;
import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TupleStructure;
import org.objectweb.medor.filter.api.Expression;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.filter.api.Operator;
import org.objectweb.medor.lib.BasicTupleStructure;
import org.objectweb.medor.query.api.CalculatedField;
import org.objectweb.medor.query.api.NestedField;
import org.objectweb.medor.query.api.OptimizationMetaData;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.api.OrderField;
import org.objectweb.medor.tuple.api.TupleLoader;

import java.util.ArrayList;
import java.util.List;

/**
 * A BasicQueryNode is an implementation of QueryNode.
 * It also implements QueryIndexGenerator and RWTupleStructure (it is its own
 * TupleStructure).
 * @author Sebastien Chassande-Barrioz
 */
public abstract class BasicQueryNode extends BasicQueryTree
	implements QueryNode {

	protected List children = new ArrayList();
	protected TupleLoader tupleLoader = null;
	protected Expression filter = null;
	/**
	 * Contains ranges of attribute to be projected from the tuple resulting
	 * of the concatenation of the tuples from the children QueryTrees.
	 */
	protected int[] indexes;

	public BasicQueryNode(String _name, OptimizationMetaData optimizationMD) {
        super(_name, optimizationMD);
	}

	public Object clone() throws CloneNotSupportedException {
		return super.clone();
	}

    public String getName() {
        return name;
    }

    /**
	 * Creates and adds a PropagatedField to the TupleStructure of the
	 * QueryNode.
	 * @param name is the name of the Field to be added.
	 * @param type is the PType of the Field to be added.
	 * @param anc are the origin Fields for the PropagatedField to be created.
	 * @return the newly created PropagatedField
	 * @throws MedorException when a Field with the same name already exists.
	 */
	public PropagatedField addPropagatedField(String name, PType type,
											  QueryTreeField[] anc)
		throws MedorException {
        String fn = getFieldName(this.name, name);
        if (name2field.containsKey(fn))
            throw new MedorException("Two fields with the same name: " + fn);
		PropagatedField newField = new BasicPropagatedField(fn, type, this, anc);
		//adds a Field to the TupleStructure
		this.addField(newField);
		//add the QueryTree of the ancestors to the set of Children
		for (int i = 0; i < anc.length; i++) {
			addChild(anc[i].getQueryTree());
		}
		return newField;
	}

	/**
	 * Creates and adds a CalculatedField to the TupleStructure of the
	 * QueryNode.
	 * @param name is the name of the Field to be added.
	 * @param type is the PType of the Field to be added.
	 * @param exp is the Expression for calculating the new Field.
	 * @return the newly created CalculatedField
	 * @throws MedorException when a Field with the same name already exists.
	 */
	public CalculatedField addCalculatedField(String name, PType type,
											  Expression exp)
		throws MedorException {
        String fn = getFieldName(this.name, name);
        if (name2field.containsKey(fn))
            throw new MedorException("Two fields with the same name: " + fn);
		CalculatedField newField = new BasicCalculatedField(fn, type, this, exp);
		this.addField(newField);
		//add the QueryTree of the FieldOperand present in the Expression to
		//the set of Children
		addChildrenFromExpression(exp);
		return newField;
	}

	/**
	 * Removes a Field.
	 * @param name is the name of the Field to be removed.
	 * @return the Field which has been removed, or null if no such Field could
	 * be removed.
	 */
	public QueryTreeField removeField(String name) throws MedorException {
		Field toRemove = this.getField(name);
		if (removeField(toRemove)) {
			//updates the children list from the other Fields and from the
			//query filter
			updateChildren();
			return (QueryTreeField)toRemove;
		}
		else return null;
	}

	public QueryTree[] getChildren() {
		return (QueryTree[]) children.toArray(new QueryTree[0]);
	}

    public QueryTreeField replace(QueryTreeField old, QueryTreeField neo) {
        int idx = this.fields.indexOf(old);
        if (idx == -1) {
            return null;
        }
        fields.set(idx, neo);
        updateChildren();
        return old;
    }

	public OptimizationMetaData getOptimizationMetaData() {
		if (optimizationMD == null) {
			optimizationMD = new BasicOptimizationMetaData();
		}
		return optimizationMD;
	}

	/**
	 * The setQueryFilter method will be overwritten for implementations
	 * which do not support the operation.
	 */
	public void setQueryFilter(Expression filter)
		throws UnsupportedOperationException {
		this.filter = filter;
		//add the QueryTree of the FieldOperands present in the Expression to
		//the Children
		updateChildren();
	}

	public Expression getQueryFilter() {
		return filter;
	}

	public abstract short getType();

	public TupleStructure getTupleStructure() {
		return this;
	}

	public TupleLoader getTupleLoader() {
		return tupleLoader;
	}

	public void setTupleLoader(TupleLoader loader) {
		tupleLoader = loader;
	}

    public void updatePropagatedField(String name, QueryTreeField[] previous)
        throws MedorException {
        Field f = (Field) name2field.get(name);
        if (f ==null)
            throw new MedorException("No field found");
        if (!(f instanceof PropagatedField))
            throw new MedorException("Bad type: " + f.getClass().getName());
        ((PropagatedField) f).replacePreviousField(previous);
        updateChildren();
    }

    public void updateCalculatedField(String name, Expression e)
        throws MedorException {
        Field f = (Field) name2field.get(name);
        if (f ==null)
            throw new MedorException("No field found");
        if (!(f instanceof CalculatedField))
            throw new MedorException("Bad type: " + f.getClass().getName());
        ((CalculatedField) f).setExpression(e);
        updateChildren();
    }

	public void setDistinct(boolean distinct) {
		this.distinct = distinct;
	}

	public boolean getDistinct() {
		return distinct;
	}

    /**
	 * Adds a Field to the current RWTupleStructure
	 * @param f The Field to be added.
	 */
	protected synchronized void addField(Field f) {
		name2field.put(f.getName(), f);
		fields.add(f);
	}

	/**
	 * Removes a Field from the current RWTupleStructure.
	 * @param f The Field to be removed.
	 * @return false if the Field could not be removed (was not part of the
	 * TupleStructure).
	 */
	private boolean removeField(Field f) {
		if (name2field.remove(f.getName()) == null)
			return false;
		else {
			fields.remove(f);
			return true;
		}
	}

	/**
	 * Updates the list of children by screening all Fields and the
	 * query filter.
	 */
	protected void updateChildren() {
		children.clear();
		Field[] fs = this.getFields();
		//add children from Fields
		for (int i = 0; i < fs.length; i++) {
			if (fs[i] instanceof PropagatedField) {
				Field[] anc = ((PropagatedField)fs[i]).getPreviousFields();
				for (int j = 0; j < anc.length; j++) {
					addChild(((QueryTreeField)anc[j]).getQueryTree());
				}
			}
			else if (fs[i] instanceof NestedField) {
				Field[] grouped = ((NestedField)fs[i]).getFields();
				for (int j = 0; j < grouped.length; j++) {
					addChild(((QueryTreeField)grouped[j]).getQueryTree());
				}
			}
			else if (fs[i] instanceof CalculatedField) {
				addChildrenFromExpression(((CalculatedField)fs[i]).getExpression());
			}
		}
		addChildrenFromExpression(this.getQueryFilter());
	}

	/**
	 * Parses an Expression and adds the QueryTrees of the FieldOperands to
	 * the children of the current QueryNode.
	 * @param exp is the Expression to be parsed.
	 */
	private void addChildrenFromExpression(Expression exp) {
		if (exp instanceof FieldOperand) {
			addChild(((QueryTreeField)((FieldOperand)exp).getField()).getQueryTree());
		}
		else if (exp instanceof Operator) {
            Operator op = (Operator) exp;
            for (int i=0; i<op.getOperandNumber(); i++) {
                addChildrenFromExpression(op.getExpression(i));
            }
		}
	}

    protected String getFieldName(String nodeName, String fieldName) {
        if (nodeName==null || nodeName.length()==0)
            return fieldName;
        else
            return nodeName + "." + fieldName;
    }

	protected void addChild(QueryTree qt) {
		if (!children.contains(qt)) {
			children.add(qt);
		}
	}
}
