/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 * 
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.query.lib;

import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.OptimizationMetaData;
import org.objectweb.medor.query.api.OperationType;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.TupleStructure;
import org.objectweb.medor.filter.lib.Equal;
import org.objectweb.medor.filter.lib.BasicOperand;
import org.objectweb.medor.filter.lib.And;
import org.objectweb.medor.filter.lib.BasicFieldOperand;
import org.objectweb.medor.filter.api.Operand;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.filter.api.Expression;

import java.util.ArrayList;

/**
 *
 * @author Sebastien Chassande-Barrioz
 */
public class Intersection extends BasicQueryNode {


	/**
	 * It represents filter associated to the Intersection QueryNode
	 * All elements of this stack have the link to its children. This implies
	 * that the top element stack[stack.length-1] is also the root of the tree which describes the
	 * filter.
	 */
	private Expression[] stack;

	public Intersection(QueryTree leftQT,
						QueryTree rightQT,
                        String name,
						OptimizationMetaData optimizationMD)
		throws MedorException {
		super(name, optimizationMD);

		// we test the attributes types and names before performing
		// the intersection opertation
		if (leftQT.getTupleStructure().getSize()
			!=rightQT.getTupleStructure().getSize()) {
			throw new MedorException(
				"The intersection must be with the same Tuple Collection");
		}
		for(int i=1; i<leftQT.getTupleStructure().getSize(); i++) {
			if (//fields must have the same name
				(!leftQT.getTupleStructure().getField(i).getName()
					.equalsIgnoreCase(
						rightQT.getTupleStructure().getField(i).getName()))
				||
					//the types must be comparable
				(!
					(
					(leftQT.getTupleStructure().getField(i).getType().isa(
						rightQT.getTupleStructure().getField(i).getType()))
					||
					(rightQT.getTupleStructure().getField(i).getType().isa(
						leftQT.getTupleStructure().getField(i).getType()))
					)
				)
			)
			{
				throw new MedorException(
					"The intersection must be with the same Tuple Collection");
			}
		}

		//adds the fields from the underlying query trees
		Field[] leftFields = leftQT.getTupleStructure().getFields();
		Field[] rightFields = rightQT.getTupleStructure().getFields();
		for (int i = 0; i < leftFields.length; i++) {
			this.addPropagatedField(
				leftFields[i].getName(),
				leftFields[i].getType(),
				new QueryTreeField[] {(QueryTreeField)leftFields[i],
									  (QueryTreeField)rightFields[i]});
		}

		indexes = new int[leftQT.getTupleStructure().getSize()];
		stack = new Expression[4 * leftQT.getTupleStructure().getSize() - 1];
		TupleStructure leftTS = leftQT.getTupleStructure();
		TupleStructure rightTS = rightQT.getTupleStructure();
		// For each element :
		// - build a stack which represents the filter
		// - Fill the fields list
		// - fix the index
		for (int i=1; i<leftTS.getSize(); i++) {
			stack[4*(i-1)] = new BasicFieldOperand(leftTS.getField(i));
			stack[4*(i-1) + 1] = new BasicFieldOperand(rightTS.getField(i));
			stack[4*(i-1) + 2] = new Equal(stack[4*i-1], stack[4*i]);
            if (i>1) {
			    stack[4*(i-1) + 3] = new And(stack[4*i+1], stack[4*i-2]);
            }
			indexes[i] = i;
		}
    }

	public void setQueryFilter(Expression f) {
		throw new UnsupportedOperationException("Intersections cannot be assigned a filter.");
	}

	public Expression getQueryFilter() {
		return stack[stack.length-1];
	}

	public short getType() {
		return OperationType.INTERSECTION;
	}
}
