/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 * 
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.query.lib;

import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.filter.api.Expression;
import org.objectweb.medor.query.api.NestQueryNode;
import org.objectweb.medor.query.api.NestedField;
import org.objectweb.medor.query.api.OperationType;
import org.objectweb.medor.query.api.OptimizationMetaData;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.api.QueryTreeField;

/**
 *
 * @author Sebastien Chassande-Barrioz
 */
public class Nest extends BasicQueryNode implements NestQueryNode {

    /**
     * It lists the fields which have been grouped.
     */
    protected PropagatedField[] groupByField = null;
    protected NestedField nestedField = null;

	/**
	 * Creates a QueryNode corresponding to a nest operation (group by).
	 * The Fields that are grouped and those that are used for the grouping
	 * must come from one single (child) QueryTree.
	 * @param groupedFields is the array of Fields which are nested by the
	 * operator. These are Fields of the child QueryTree.
	 * @param groupedFieldName is the name of the Field corresponding to the
	 * nesting.
	 * @param groupByField is the array of Fields which are used for defining
	 * the group to be nested (equivalent to SQL GROUP BY). These are Fields of
	 * the child QueryTree. They are also called NestingFields.
	 * @param optimizationMD are meta data for optimization.
	 */
    public Nest(QueryTreeField[] groupedFields,
                String groupedFieldName,
                QueryTreeField[] groupByField,
                String nodeName,
                OptimizationMetaData optimizationMD) throws MedorException{
        super(nodeName, optimizationMD);
		//adding all groupBy fields as PropagatedFields
 		this.groupByField = new PropagatedField[groupByField.length];
		for (int i = 0; i < groupByField.length; i++) {
			QueryTreeField[] anc = {groupByField[i]};
			this.groupByField[i] = this.addPropagatedField(
				groupByField[i].getName(),
				groupByField[i].getType(),
				anc);
		}
		//creating and adding the NestedField
        nestedField = this.addNestedField(groupedFieldName, groupedFields);
     }

	/**
	 * Creates and adds a NestedField to the TupleStructure of the
	 * QueryNode.
	 * @param name is the name of the Field to be added.
	 * @param grouped are the grouped Fields for the NestedField to be created.
	 * @return the newly created NestedField
	 * @throws MedorException when a Field with the same name already exists.
	 */
	public NestedField addNestedField(String name,
									  QueryTreeField[] grouped)
		throws MedorException {
		NestedField newField =
			new BasicNestedField(name, this, grouped);
		this.addField(newField);
		//add the QueryTree of the grouped fields to the set of Children
		for (int i = 0; i < grouped.length; i++) {
			addChild(grouped[i].getQueryTree());
		}
		return newField;
	}

	public void setQueryFilter(Expression e) {
		throw new UnsupportedOperationException("Nest query nodes cannot be assigned a filter.");
	}

    public PropagatedField[] getNestingFields() {
        return groupByField;
    }

    public NestedField getNestedField() {
        return nestedField;
    }

    public short getType() {
        return OperationType.NEST;
    }
}
