/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre, S. Chassande-Barrioz
 */

package org.objectweb.medor.query.rdb.lib;

import org.objectweb.jorm.runtime.mapper.rdb.adapter.api.RdbAdapter;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.jorm.api.PException;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.QueryLeafException;
import org.objectweb.medor.api.TupleStructure;
import org.objectweb.medor.api.DataSourceException;
import org.objectweb.medor.datasource.api.DataStore;
import org.objectweb.medor.filter.api.BinaryOperator;
import org.objectweb.medor.filter.api.Expression;
import org.objectweb.medor.filter.api.FieldOperand;
import org.objectweb.medor.filter.api.Operand;
import org.objectweb.medor.filter.api.Operator;
import org.objectweb.medor.filter.api.ParameterOperand;
import org.objectweb.medor.filter.api.UnaryOperator;
import org.objectweb.medor.filter.lib.BasicOperand;
import org.objectweb.medor.filter.lib.InCollection;
import org.objectweb.medor.filter.lib.Like;
import org.objectweb.medor.filter.lib.MemberOf;
import org.objectweb.medor.filter.lib.IndexedLocate;
import org.objectweb.medor.filter.lib.Substring;
import org.objectweb.medor.filter.lib.IsNull;
import org.objectweb.medor.filter.lib.FirstLocate;
import org.objectweb.medor.filter.lib.ExpressionPrinter;
import org.objectweb.medor.filter.lib.Concat;
import org.objectweb.medor.lib.BasicTupleStructure;
import org.objectweb.medor.lib.Log;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.api.OrderField;
import org.objectweb.medor.query.rdb.api.QualifiedTable;
import org.objectweb.medor.query.rdb.api.RdbExpField;
import org.objectweb.medor.query.rdb.api.RdbExpQueryLeaf;
import org.objectweb.medor.query.lib.BasicQueryTree;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.Collection;
import java.util.Iterator;
import java.util.Vector;
import java.util.List;

/**
 * This class represents a QueryLeaf that maps onto a relational database store.
 * It implements the case where the relational query is provided as a set of
 * QualifiedTables and an Expression representing the filter.
 */
public class BasicRdbExpQueryLeaf extends BasicQueryTree
		implements RdbExpQueryLeaf, TupleStructure {

	private final static String LAND = "&";
	private final static String LCOND_AND = "&&";
	private final static String AND = "AND";
	private final static String EQUAL = "=";
	private final static String LEQUAL = "==";
	private final static String LOR = "|";
	private final static String LCOND_OR = "||";
	private final static String OR = "OR";
	private final static String NOT = "NOT";
	private boolean distinct = false;

	Expression sqlFilter;
	DataStore ds;
	QualifiedTable[] from;
	String nodeName;
	RdbExpField groupByField;
	String rdbAdapterName = null;
	RdbAdapter rdbAdapter = null;
	List rootJoinedTables;
	String query = null;

	protected Logger log = null;
	protected boolean debug = false;

	/**
	 * Constructs a BasicRdbExpQueryLeaf from an Expression and a set of
	 * QualifiedTables.
	 * <p>The SQL request (string) will be calculated.
	 * @param ds is the DataStore associated to the QueryLeaf to be created
	 * @param from is the array of QualifiedTables associated to the QueryLeaf
	 * to be created
	 * @throws QueryLeafException if the QualifiedTables have all the same
	 * names, and in this case also the same aliases.
	 */
	public BasicRdbExpQueryLeaf(DataStore ds,
								QualifiedTable[] from, String nodeName)
			throws QueryLeafException {
		super(nodeName, null);
		log = Log.loggerFactory.getLogger(
				Log.MEDOR_PREFIX + ".query.rdb");
		debug = log != null && log.isLoggable(BasicLevel.DEBUG);
		this.nodeName = nodeName;
		this.ds = ds;
		//all QualifiedTables should have different names, and if not,
		//different aliases
		//TODO throw an exception if all QualifiedTables have the same name
		//and the same alias name
		this.from = from;
	}

	public BasicRdbExpQueryLeaf(DataStore ds, String nodeName) {
		super(nodeName, null);
		log = Log.loggerFactory.getLogger(
				Log.MEDOR_PREFIX + ".query.rdb");
		debug = log != null && log.isLoggable(BasicLevel.DEBUG);
		this.nodeName = nodeName;
		this.ds = ds;
	}

	public void setRdbAdapterName(String an) {
		this.rdbAdapterName = an;
	}

	public String getRdbAdapterName() {
		return rdbAdapterName;
	}

	public void setRdbAdapter(RdbAdapter adapter) {
		rdbAdapter = adapter;
	}

	public RdbAdapter getRdbAdapter() {
		return rdbAdapter;
	}

	public String getName() {
		return nodeName;
	}

	public void setRootJoinedTables(List rootjts) {
		rootJoinedTables = rootjts;
	}

	/**
	 * Adds a Field to the RdbExpQueryLeaf
	 * @param name is the name of the Field to be added.
	 * @param type is the PTyme of the Field to be added.
	 * @param colName is the name of the column in the associated
	 * QualifiedTable.
	 * @param table is the QualifiedTable from which the Field is to be created.
	 * @return the created RdbExpField.
	 */
	public RdbExpField addRdbField(String name,
								   PType type,
								   String colName,
								   QualifiedTable table)
			throws QueryLeafException {
		//check that the parameter QualifiedTable is a valid QualifiedTable
		//in this RdbExpQueryLeaf
		QualifiedTable[] qts = this.getQualifiedTables();
		int i = 0;
		while (i < qts.length) {
			if (table == qts[i])
				break;
			else
				i++;
		}
		if (i == qts.length)
			throw new QueryLeafException("The parameter QualifiedTable is not a QualifiedTable of the current RdbExpQueryLeaf");
		else {
			RdbExpField f =
					new BasicRdbExpField(name, type, colName, table, this);
			name2field.put(f.getName(), f);
			fields.add(f);
			return f;
		}
	}

	public void addQualifiedTable(QualifiedTable qt) {
		if (from == null)
			from = new QualifiedTable[]{qt};
		else {
			QualifiedTable[] qts = new QualifiedTable[from.length + 1];
			System.arraycopy(from, 0, qts, 0, from.length);
			qts[from.length] = qt;
			from = qts;
		}
	}

	public void removeQualifiedTable(QualifiedTable qt) throws MedorException {
		if (from == null)
			throw new MedorException("QualifiedTable " + qt + " is not in leaf " + this);
		for (int i = 0; i < from.length; i++) {
			if (from[i] == qt) {
				//System.out.println("Found table " + qt.getTableName());
				QualifiedTable[] qts = new QualifiedTable[from.length - 1];
				//TODO use System.arraycopy instead
				for (int j = 0; j < i; j++) {
					qts[j] = from[j];
				}
				for (int j = i + 1; j < from.length; j++) {
					qts[j - 1] = from[j];
				}
				from = qts;
				return;
			}
		}
		throw new MedorException("QualifiedTable " + qt + " is not in leaf " + this);
	}

	public boolean containsQualifiedTable(QualifiedTable qt) {
		if (from == null)
			return false;
		for (int i = 0; i < from.length; i++) {
			if (from[i] == qt)
				return true;
		}
		return false;
	}

	/**
	 * Adds an RdbExpField to the RdbExpQueryLeaf
	 * @param rdbField is the RdbField to be added.
	 */
	public RdbExpField addRdbField(RdbExpField rdbField)
			throws QueryLeafException {
		rdbField.setQueryLeaf(this);
		name2field.put(rdbField.getName(), rdbField);
		fields.add(rdbField);
		return rdbField;
	}

	public void removeRdbField(RdbExpField rdbField)
			throws QueryLeafException {
		name2field.remove(rdbField.getName());
		fields.remove(rdbField);
	}

	/**
	 * Builds and returns the SQL query as a String corresponding to the
	 * current RdbExpQueryLeaf.
	 * <p>
	 * The SQL query is computed from the provided RdbExpFields (SELECT clause),
	 * QualifiedTables (FROM clause) and Expression (WHERE clause).
	 * @return the SQL query as a String.
	 */
	public String getSqlRequest(ParameterOperand[] pos) throws MedorException {
		if (query != null) {
			return query;
		}
		log.log(BasicLevel.DEBUG, "Entering getSqlRequest for " + this);
		String where = (sqlFilter == null
				? ""
				: parseExpression(sqlFilter, pos));
		log.log(BasicLevel.DEBUG, "WHERE clause for " + this + " : " + where);

		if (rootJoinedTables != null && rootJoinedTables.size() > 0) {
			log.log(BasicLevel.DEBUG, "Query with rootJoinedTables");
			query = rdbAdapter.getQuery(
					getSelectList(""), rootJoinedTables, where);
		}
		else {
			log.log(BasicLevel.DEBUG, "Query without rootJoinedTables");
			query = getSelectList("SELECT ")
					+ getFromList()
					+ ((where == "") ? "" : " WHERE " + where);
		}
		OrderField[] orderByFields = this.getOrderBy();
		if ((orderByFields != null) && (orderByFields.length > 0)) {
			String orderByString = "";
			for (int cpt = 0; (cpt < orderByFields.length); cpt++) {
				if (cpt > 0) orderByString = orderByString + ", ";
				orderByString = orderByString +
						getQualifiedFieldName((RdbExpField) (orderByFields[cpt].getField()));
				orderByString +=
						(orderByFields[cpt].getDesc() ? " DESC" : " ASC");
			}
			query = query + " " + "ORDER BY " +
					orderByString;
		}
		if (groupByField != null) {
			query = query + " " + "GROUP BY " +
					getQualifiedFieldName(groupByField);
		}
		//In the case of OrderBy, and for Oracle, embed the query inside
		//a select *
		if ((orderByFields != null)
				&& (orderByFields.length > 0)
				&& rdbAdapter != null) {
			query = rdbAdapter.handleOrderBy(query);
		}
		return query;
	}

	/**
	 * Sets the filter associated to the RbdQueryLeaf.
	 * <p>This filter is
	 * represented by an Expression. Unlike the filter of a QueryNode, is
	 * applies to its own Fields, instead of the Fields of its children.
	 * @param exp is the Medor Expression representing the filter to be applied
	 * to the relational source tables.
	 */
	public void setQueryFilter(Expression exp) {
		this.sqlFilter = exp;
	}

	/**
	 * Retrieves the filter associated to the RbdQueryLeaf.
	 * @return the Medor Expression representing the filter to be applied to
	 * the relational source tables.
	 */
	public Expression getQueryFilter() {
		return this.sqlFilter;
	}

	/**
	 * Returns the associated QualifiedTables.
	 * @return the array of associated QualifiedTables.
	 */
	public QualifiedTable[] getQualifiedTables() {
		return this.from;
	}

	public DataStore getDataStore() {
		return ds;
	}

	public TupleStructure getTupleStructure() {
		return this;
	}

	public boolean getDistinct() {
		return distinct;
	}

	public void setDistinct(boolean d) {
		distinct = d;
	}

	public void setGroupBy(RdbExpField groupBy) throws QueryLeafException {
		if (fields.contains(groupBy)) {
			groupByField = groupBy;
		}
		else
			throw new QueryLeafException("This RdbExpField does not exist in the TupleStructure of this RdbQueryLeaf");
	}

	public RdbExpField getGroupByField() {
		return groupByField;
	}

	/**
	 * Builds the SELECT clause of the query (without "SELECT").
	 * @return the list of qualified columns for the SELECT clause
	 * @param selectList is the start of the SELECT clause to which the list
	 * of qualified columns is appended.
	 */
	protected String getSelectList(String selectList) {
		selectList += (getDistinct() ? "DISTINCT " : "");
		for (int i = 0; i < fields.size(); i++) {
			selectList = selectList + getQualifiedFieldName(
					((RdbExpField) fields.get(i)));
			if (i < fields.size() - 1)
				selectList += ", ";
		}
		return selectList;
	}

	/**
	 * Builds the qualified name of a RdbExpField.
	 * @return the qualified name of an RdbExpField.
	 * @param field is the RdbExpField for which the qualified name is built.
	 */
	protected String getQualifiedFieldName(RdbExpField field) {
		String res = null;
		if (field.getTable().getAliasName() == null)
			res = field.getTable().getTableName();
		else
			res = field.getTable().getAliasName();
		return res + "." + field.getColumnName();
	}

	/**
	 * Builds the FROM clause of the query (without "FROM").
	 * @return the list of qualified table names for the FROM clause
	 * tables is appended.
	 */
	protected String getFromList() {
		StringBuffer sb = new StringBuffer();
		sb.append(" FROM ");
		for (int i = 0; i < from.length; i++) {
			sb.append(from[i].getTableName());
			if (from[i].getAliasName() != null) {
				if (rdbAdapter != null) {
					rdbAdapter.writeTableAlias(from[i].getAliasName(), sb);
				}
				else {
					sb.append(" as ");
					sb.append(from[i].getAliasName());
				}
			}
			if (i < from.length - 1)
				sb.append(", ");
		}
		log.log(BasicLevel.DEBUG, "FROM clause: " + sb.toString());
		return sb.toString();
	}

	/**
	 * Computes a portion of SQL statement for the WHERE clause from a Medor
	 * Expression.
	 * @param exp is the Medor Expression from which to compute the portion of
	 * the SQL WHERE clause.
	 * @return the portion of the SQL WHERE clause corresponding to the
	 * Expression.
	 */
	private String parseExpression(Expression exp, ParameterOperand[] pos)
			throws MedorException {
		log.log(BasicLevel.DEBUG, "Parsing " + ExpressionPrinter.e2str(exp));
		if (exp instanceof FieldOperand) {
			log.log(BasicLevel.DEBUG, "FieldOperand for Field " + ((FieldOperand) exp).getField());
			return getQualifiedFieldName(
					((RdbExpField) ((FieldOperand) exp).getField()));
		}
		else if (exp instanceof ParameterOperand)
			return "?";

		else if (exp instanceof BasicOperand) {
			BasicOperand o = (BasicOperand) exp;
			if (rdbAdapter != null) {
				switch (o.getType().getTypeCode()) {
				case PType.TYPECODE_BOOLEAN:
					return rdbAdapter.getValueAsSQLString(o.getBoolean());
				case PType.TYPECODE_CHAR:
					return rdbAdapter.getValueAsSQLString(o.getChar());
				case PType.TYPECODE_BYTE:
					return rdbAdapter.getValueAsSQLString(o.getByte());
				case PType.TYPECODE_SHORT:
					return rdbAdapter.getValueAsSQLString(o.getShort());
				case PType.TYPECODE_INT:
					return rdbAdapter.getValueAsSQLString(o.getInt());
				case PType.TYPECODE_LONG:
					return rdbAdapter.getValueAsSQLString(o.getLong());
				case PType.TYPECODE_FLOAT:
					return rdbAdapter.getValueAsSQLString(o.getFloat());
				case PType.TYPECODE_DOUBLE:
					return rdbAdapter.getValueAsSQLString(o.getDouble());
				default:
					return rdbAdapter.getValueAsSQLString(
							o.getObject(),
							o.getType());
				}
			}
			else {
				switch (o.getType().getTypeCode()) {
				case PType.TYPECODE_OBJBOOLEAN:
				case PType.TYPECODE_BOOLEAN:
					return (o.getBoolean() ? "1" : "0");
				case PType.TYPECODE_BYTE:
				case PType.TYPECODE_OBJBYTE:
				case PType.TYPECODE_SHORT:
				case PType.TYPECODE_OBJSHORT:
				case PType.TYPECODE_INT:
				case PType.TYPECODE_OBJINT:
				case PType.TYPECODE_LONG:
				case PType.TYPECODE_OBJLONG:
				case PType.TYPECODE_FLOAT:
				case PType.TYPECODE_OBJFLOAT:
				case PType.TYPECODE_DOUBLE:
				case PType.TYPECODE_OBJDOUBLE:
					return o.getValueAsString();
				default:
					return "'" + o.getValueAsString() + "'";
				}
			}
		}
		else if (exp instanceof Like) {
			Like l = (Like) exp;
			return
					"("
					+ parseExpression(l.getExpression(0), pos)
					+ " LIKE "
					+ parseExpression(l.getExpression(1), pos)
					+ (l.getOperandNumber() == 3
					? " ESCAPE " + parseExpression(l.getExpression(2), pos)
					+ ")"
					: ")");
		}
		else if (exp instanceof InCollection) {
			InCollection bo = (InCollection) exp;
			return
					"("
					+ parseExpression(bo.getLeftExpression(), pos)
					+ " "
					+ toSqlOperator(bo.getOperatorString())
					+ " "
					+ getSQLEnumeration(bo.getRightExpression(),
							bo.getElemType(),
							pos)
					+ ")";
		}
		else if (exp instanceof MemberOf) {
			log.log(BasicLevel.DEBUG, "Found MemberOf: " + ExpressionPrinter.e2str(exp));
			MemberOf mo = (MemberOf) exp;
			int size = mo.getOperandNumber() / 2;
			String subquery = "(";
			for (int i = 0; i < size; i++) {
				subquery += parseExpression(mo.getExpression(i), pos);
				subquery += (i < (size - 1) ? ", " : "");
			}
			subquery += " IN (";
			log.log(BasicLevel.DEBUG, "subquery: " + subquery);
			FieldOperand fop = (FieldOperand) mo.getExpression(size);
			QueryTreeField qtf = (QueryTreeField) fop.getField();
			BasicRdbExpQueryLeaf qLeaf =
					(BasicRdbExpQueryLeaf) qtf.getQueryTree();
			if (qLeaf == this) {
				throw new MedorException("The field '" + qtf.getName()
						+ "' is not attched to a sub queryleaf (qtf= "
						+ qtf + ")");
			}
			qLeaf.setRdbAdapter(rdbAdapter);
			subquery += qLeaf.getSqlRequest(pos);
			subquery += "))";
			return subquery;
		}
		else if (exp instanceof FirstLocate) {
			if (rdbAdapter == null) {
				return
						toSqlOperator(((Operator) exp).getOperatorString())
						+ "("
						+ parseExpression(
								((Operator) exp).getExpression(0), pos)
						+ " in "
						+ parseExpression(
								((Operator) exp).getExpression(1), pos)
						+ ")";
			}
			else {
				try {
					return rdbAdapter.getFirstLocateExpression(
							parseExpression(
									((Operator) exp).getExpression(0), pos),
							parseExpression(
									((Operator) exp).getExpression(1), pos)
					);
				}
				catch (PException pe) {
					throw new DataSourceException(pe.getMessage());
				}
			}
		}
		else if (exp instanceof Concat) {
			if (rdbAdapter == null) {
				return toSqlOperator(((Operator) exp).getOperatorString())
						+ "("
						+ parseExpression(
								((Operator) exp).getExpression(0), pos)
						+ ", "
						+ parseExpression(
								((Operator) exp).getExpression(1), pos)
						+ ")";
			}
			else {
				return rdbAdapter.getConcatExpression(
						parseExpression(((Operator) exp).getExpression(0), pos),
						parseExpression(((Operator) exp).getExpression(1), pos)
				);
			}
		}
		else if (exp instanceof BinaryOperator) {
			BinaryOperator bo = (BinaryOperator) exp;
			return
					"("
					+ parseExpression(bo.getLeftExpression(), pos)
					+ " "
					+ toSqlOperator(bo.getOperatorString())
					+ " "
					+ parseExpression(bo.getRightExpression(), pos)
					+ ")";
		}
		//TODO: in the case of NOT xyz, it may sometimes be better to
		//produce something else, e.g. NOT (toto IS NULL) -> toto IS NOT NULL
		else if (exp instanceof IsNull) {
			return
					parseExpression(((UnaryOperator) exp).getExpression(), pos)
					+ " "
					+ toSqlOperator(((Operator) exp).getOperatorString());
		}
		else if (exp instanceof UnaryOperator) {
			return
					toSqlOperator(((Operator) exp).getOperatorString())
					+ "("
					+ parseExpression(
							((UnaryOperator) exp).getExpression(), pos)
					+ ")";
		}
		else if (exp instanceof Substring) {
			return
					((rdbAdapter == null)
					? toSqlOperator(((Operator) exp).getOperatorString())
					: rdbAdapter.getSubstringName()
					)
					+ "("
					+ parseExpression(((Operator) exp).getExpression(0), pos)
					+ ", "
					+ parseExpression(((Operator) exp).getExpression(1), pos)
					+ ", "
					+ parseExpression(((Operator) exp).getExpression(2), pos)
					+ ")";
		}
		else if (exp instanceof IndexedLocate) {
			if (rdbAdapter == null) {
				return
						toSqlOperator(((Operator) exp).getOperatorString())
						+ "("
						+ parseExpression(
								((Operator) exp).getExpression(0), pos)
						+ ", "
						+ parseExpression(
								((Operator) exp).getExpression(1), pos)
						+ ", "
						+ parseExpression(
								((Operator) exp).getExpression(2), pos)
						+ ")";
			}
			else {
				try {
					return rdbAdapter.getIndexedLocateExpression(
							parseExpression(
									((Operator) exp).getExpression(0), pos),
							parseExpression(
									((Operator) exp).getExpression(1), pos),
							parseExpression(
									((Operator) exp).getExpression(2), pos)
					);
				}
				catch (PException pe) {
					throw new DataSourceException(pe.getMessage());
				}
			}
		}
		else
			throw new MedorException("Umanaged expression element: " + exp);
	}

	private String getSQLEnumeration(
			Expression e,
			PType elemType,
			ParameterOperand[] pos)
			throws MedorException {
		Collection c = null;
		String result = "(";
		if (e instanceof ParameterOperand) {
			String pn = ((ParameterOperand) e).getName();
			if (pos == null || pos.length == 0)
				throw new MedorException("No Parameter");
			int i = 0;
			while (i < pos.length && !pn.equals(pos[i].getName()))
				i++;
			if (i < pos.length)
				c = (Collection) pos[i].getObject();
			else
				throw new MedorException("Parameter " + pn + " not found");
		}
		else
			c = (Collection) ((Operand) e).getObject();
		Iterator it = c.iterator();
		if (!it.hasNext())
			throw new MedorException("Empty collection for InCollection operator");
		while (it.hasNext()) {

			if (elemType.equals(PTypeSpace.STRING)) {
				result += "'";
				result += it.next();
				result += "'";
			}
			else
				result += it.next();
			if (it.hasNext())
				result += ", ";
		}
		return result += ")";
	}

	/**
	 * Returns the SQL equivalent of the Java operator String.
	 */
	private static String toSqlOperator(String in) {
		if (LAND.equals(in) || LCOND_AND.equals(in)) {
			return BasicRdbExpQueryLeaf.AND;
		}
		else if (LOR.equals(in) || LCOND_OR.equals(in)) {
			return BasicRdbExpQueryLeaf.OR;
		}
		else if (in.equals(BasicRdbExpQueryLeaf.LEQUAL)) {
			return BasicRdbExpQueryLeaf.EQUAL;
		}
		else if (in.equals(Operator.NOT)) {
			return BasicRdbExpQueryLeaf.NOT;
		}
		else
			return in;
	}
}
