/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 * 
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.tuple.api;

import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.filter.api.Operand;

import java.util.Date;

/**
 * This interface represents a tuple like in database terminology.
 * <p>
 * It is a list of attributes. Each getXXX(attributeNum) method retrieves the
 * value of
 * an attribute into an XXX type if possible.  Columns are numbered starting
 * from 1.
 * <p>Tuple extends the Cloneable
 * interface to indicate to the Object.clone() method that it is legal for
 * that method to make a field-for-field copy of instances of a tuple class.
 */
public interface Tuple extends Cloneable {

	/**
	 * Give the number of the attribute in this Tuple object.
	 * @return an int
	 */
	public int getSize();

	/**
	 * Give the array of Operand object representation of the Tuple.
	 * @return an array of Operand type.
	 */
	public Operand[] toOperandArray();

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple
	 * as a Boolean of java programming language
	 */
	public boolean getBoolean(int i) throws MedorException;

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple as a Java byte
	 */
	public byte getByte(int i) throws MedorException;

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple as a Java byte
	 */
	public byte[] getByteArray(int i) throws MedorException;

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple as a Java char
	 */
	public char getChar(int i) throws MedorException;

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple as a Java char[]
	 */
	public char[] getCharArray(int i) throws MedorException;

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple as a java.sql.Data
	 */
	public Date getDate(int i) throws MedorException;

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple as a Java double
	 */
	public double getDouble(int i) throws MedorException;

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple as a Java float
	 */
	public float getFloat(int i) throws MedorException;

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple as a java int
	 */
	public int getInt(int i) throws MedorException;

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple
	 * as an org.objectweb.medor.filter.Operand Object
	 */
	public Operand getLikeOperand(int i) throws MedorException;

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple as a Java long
	 */
	public long getLong(int i) throws MedorException;

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple as a Java Object
	 */
	public Object getObject(int i) throws MedorException;

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple
	 * as a Short in the java programming language
	 */
	public short getShort(int i) throws MedorException;

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple
	 * as a String in java programming language
	 */
	public String getString(int i) throws MedorException;

	/**
	 * Returns the value of the designated column
	 * in the current row of this Tuple
	 * as a org.objectweb.medor object
	 */
	public TupleCollection getTupleCollection(int i) throws MedorException;

    public Object clone();

    boolean isDefined(int i);
}
