/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.tuple.api;

import org.objectweb.medor.api.TupleStructure;
import org.objectweb.medor.api.MedorException;

/**
 * This interface represent collections of tuples wich are manipulated by MEDOR.
 * <p>A TupleCollection is a set of data typically representing the result of
 * querying a data source result.
 * Like JDBC ResultSets, a TupleCollection  is scrollable over
 * iteration methods. The number, types and properties of a TupleCollection
 * object's
 * attribute are provided by the TupleStructure object returned by the
 * TupleCollection.getMetaData() method.
 * <p>
 * For example, to print all values of the first attribute of a
 * TupleCollection tc :
 * <tt>while (!tci.isLast()){<br>
 * System.out.println(tci.getJavaObject(1));
 * tci.next();<br>
 * }</tt>
 * Methods are provided to enumerate through the attributes of the TupleCollection
 * This Interface represent also the data flow in the Query Tree wich also represents
 * the result of a query.
 */
public interface TupleCollection {
	/**
	 * Retrieves the number,types and stors properties of the data sources of this TupleCollection Object.
	 * @return a TupleCollectionMetaData Object defining teh TupleCollection.
	 * @throws MedorException if a data source access error occurs
	 */
	public TupleStructure getMetaData() throws MedorException;

	/**
	 * Indicates whether the cursor is on the last row of this TupleCollection object.
	 * @return true if the current Tuple is the last one or if the TupleCollection is Empty.
	 * @throws MedorException if data source access error
	 */
	public boolean isLast() throws MedorException;

	/**
	 * Moves the cursor down one row from its current position to the next tuple of this
	 * TupleCollection object.
	 * @return true if move is impossible and false if move is not performed.
	 * @throws MedorException if data source access error
	 */
	public boolean next() throws MedorException;

	/**
	 * Moves the cursor to the first Tuple, row number 1.
	 * @throws MedorException if data source access error
	 */
	public void first() throws MedorException;

	/**
	 * Retrieves the current row Number.
	 * @return -1 if the iterator is out of bounds of the TupleCollection or if it is empty.
	 * @throws MedorException if data source access error.
	 */
	public int getRow() throws MedorException;

	/**
	 * Returns the value of the designated column in the current row of this TupleCollection object as an
	 * Object in the Java programming language.
	 * @return the value of the designeted attribute like an Object.
	 * @throws MedorException if data source access error or invald attribute index.
	 */
	public Object getObject(int column) throws MedorException;

	/**
	 * Returns the Tuple value of the current row of this TupleCollection object.
	 * @return the value of the current Tuple.
	 * @throws MedorException if data source access error.
	 */
	public Tuple getTuple() throws MedorException;

	/**
	 * Returns the Tuple value of the designeted row of this TupleCollection object.
	 * @param row :int
	 * @return the a Tuple Number row.
	 * @throws MedorException if data source access error.
	 */
	public Tuple getTuple(int row) throws MedorException;

	/**
	 * Moves the cursor to the given row number in this TupleCollection object.
	 * The first row is row 1, the second is row 2, and so on.
	 * @return true if move is possible
	 * @throws MedorException if row number is invelid or data source access error
	 */
	public boolean row(int row) throws MedorException;

	/**
	 * Tell whether the current TupleCollection Object is empty or no.
	 * @return true if the TupleCollection is empty.
	 * @throws MedorException if data source access error.
	 */
	public boolean isEmpty() throws MedorException;
}
