/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 * 
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.tuple.lib;

import org.objectweb.jorm.type.api.PType;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.api.TypingException;
import org.objectweb.medor.filter.api.Operand;
import org.objectweb.medor.filter.api.VariableOperand;
import org.objectweb.medor.filter.lib.BasicVariableOperand;
import org.objectweb.medor.tuple.api.Tuple;
import org.objectweb.medor.tuple.api.TupleCollection;
import org.objectweb.medor.type.lib.QTypeTuple;

import java.util.Date;


/**
 * This class represent a tuple interface in a memory. Java array is used to stores fienlds Values.
 * Rather than getXXX() methods, setXXX() allow updating the content of the Tuple.
 */
public class MemoryTuple implements Tuple {
    private int size;
    private VariableOperand[] uplet;

    public MemoryTuple(PType types[]) {
        size = types.length;
        uplet = new BasicVariableOperand[size];
        for (int cpt = 0; (cpt < types.length); cpt++) {
            uplet[cpt] = new BasicVariableOperand(types[cpt]);
        }
    }

    public MemoryTuple(VariableOperand fields[]) {
        size = fields.length;
        uplet = fields;
    }

    public MemoryTuple(Operand fields[]) {
        try {
            PType type;
            size = fields.length;
            uplet = new BasicVariableOperand[size];
            for (int i = 0; (i < size); i++) {
                type = fields[i].getType();
                uplet[i] = new BasicVariableOperand(type);
                switch (type.getTypeCode()) {
                    case QTypeTuple.TYPECODE_INT:
                        uplet[i].setValue(fields[i].getInt());
                        break;
                    case QTypeTuple.TYPECODE_SHORT:
                        uplet[i].setValue(fields[i].getShort());
                        break;
                    case QTypeTuple.TYPECODE_BYTE:
                        uplet[i].setValue(fields[i].getByte());
                        break;
                    case QTypeTuple.TYPECODE_LONG:
                        uplet[i].setValue(fields[i].getLong());
                        break;
                    case QTypeTuple.TYPECODE_DOUBLE:
                        uplet[i].setValue(fields[i].getDouble());
                        break;
                    case QTypeTuple.TYPECODE_BOOLEAN:
                        uplet[i].setValue(fields[i].getBoolean());
                        break;
                    case QTypeTuple.TYPECODE_FLOAT:
                        uplet[i].setValue(fields[i].getFloat());
                        break;
                    case QTypeTuple.TYPECODE_STRING:
                        uplet[i].setValue(fields[i].getString());
                        break;
                    case QTypeTuple.TYPECODE_CHAR:
                        uplet[i].setValue(fields[i].getChar());
                        break;
                    case QTypeTuple.TYPECODE_DATE:
                        uplet[i].setValue(fields[i].getDate());
                        break;
				case QTypeTuple.TYPECODE_OBJBOOLEAN:
				case QTypeTuple.TYPECODE_OBJBYTE:
				case QTypeTuple.TYPECODE_OBJCHAR:
				case QTypeTuple.TYPECODE_OBJDOUBLE:
				case QTypeTuple.TYPECODE_OBJFLOAT:
				case QTypeTuple.TYPECODE_OBJINT:
				case QTypeTuple.TYPECODE_OBJLONG:
				case QTypeTuple.TYPECODE_OBJSHORT:
					uplet[i].setValue(fields[i].getObject());
					break;
                        // TO DO: Add TupleCollection type and Tuple Types
                }
            }
        } catch (TypingException texp) {
            // impossible case
        }
    }

    public boolean isDefined(int i) {
        return uplet[i - 1].isDefined();
    }

    public int getSize() {
        return size;
    }

    public Operand[] toOperandArray() {
        return uplet;
    }

    public int getInt(int i) throws MedorException {
        if ((i <= 0) || (i > size)) {
            throw new MedorException("Invalid Field index: " + i);
        } else {
            return uplet[i - 1].getInt();
        }
    }

    public void setInt(int x, int i) throws MedorException {
        if ((i <= 0) || (i > size)) {
            throw new MedorException("Invalid Field index: " + i);
        } else {
            uplet[i - 1].setValue(x);
        }
    }

	public byte getByte(int i) throws MedorException {
		 if ((i <= 0) || (i > size)) {
			 throw new MedorException("Invalid Field index: " + i);
		 } else {
			 return uplet[i - 1].getByte();
		 }
	 }

	public byte[] getByteArray(int i) throws MedorException {
		 if ((i <= 0) || (i > size)) {
			 throw new MedorException("Invalid Field index: " + i);
		 } else {
			 return uplet[i - 1].getByteArray();
		 }
	 }

     public void setByte(byte x, int i) throws MedorException {
        if ((i <= 0) || (i > size)) {
            throw new MedorException("Invalid Field index: " + i);
        } else
            uplet[i - 1].setValue(x);
    }

    public double getDouble(int i) throws MedorException {
        if ((i <= 0) || (i > size)) {
            throw new MedorException("Invalid Field index: " + i);
        } else {
            return uplet[i - 1].getDouble();
        }
    }

    public void setDouble(double x, int i) throws MedorException {
        if ((i <= 0) || (i > size)) {
            throw new MedorException("Invalid Field index: " + i);
        } else
            uplet[i - 1].setValue(x);
    }

    public float getFloat(int i) throws MedorException {
        if ((i <= 0) || (i > size)) {
            throw new MedorException("Invalid Field index: " + i);
        } else {
            return uplet[i - 1].getFloat();
        }
    }

    public void setFloat(float x, int i) throws MedorException {
        if ((i <= 0) || (i > size)) {
            throw new MedorException("Invalid Field index: " + i);
        } else
            uplet[i - 1].setValue(x);
    }


    public short getShort(int i) throws MedorException {
        if ((i <= 0) || (i > size)) {
            throw new MedorException("Invalid Field index: " + i);
        } else
            return uplet[i - 1].getShort();
    }

    public void setShort(short x, int i) throws MedorException {
        if ((i <= 0) || (i > size)) {
            throw new MedorException("Invalid Field index: " + i);
        } else
            uplet[i - 1].setValue(x);
    }

    public String getString(int i) throws MedorException {
        if ((i <= 0) || (i > size)) {
            throw new MedorException("Invalid Field index: " + i);
        } else
            return uplet[i - 1].getString();
    }

    public void setString(String x, int i) throws MedorException {
        if ((i <= 0) || (i > size)) {
            throw new MedorException("Invalid Field index: " + i);
        } else
            uplet[i - 1].setValue(x);
    }

    public long getLong(int i) throws MedorException {
        if ((i <= 0) || (i > size)) {
            throw new MedorException("Invalid Field index: " + i);
        } else
            return uplet[i - 1].getLong();
    }

    public void setLong(long x, int i) throws MedorException {
        if ((i <= 0) || (i > size)) {
            throw new MedorException("Invalid Field index: " + i);
        } else
            uplet[i - 1].setValue(x);
    }

    public Object getObject(int i) throws MedorException {
        Object result = null;
        if ((i <= 0) || (i > size)) {
            throw new MedorException("Invalid Field index: " + i);
        } else {
            switch (uplet[i - 1].getType().getTypeCode()) {
                case QTypeTuple.TYPECODE_INT:
                    result = new Integer(uplet[i - 1].getInt());
                    break;
                case QTypeTuple.TYPECODE_SHORT:
                    result = new Short(uplet[i - 1].getShort());
                    break;
                case QTypeTuple.TYPECODE_BYTE:
                    result = new Byte(uplet[i - 1].getByte());
                    break;
                case QTypeTuple.TYPECODE_LONG:
                    result = new Long(uplet[i - 1].getLong());
                    break;
                case QTypeTuple.TYPECODE_DOUBLE:
                    result = new Double(uplet[i - 1].getDouble());
                    break;
                case QTypeTuple.TYPECODE_BOOLEAN:
                    result = new Boolean(uplet[i - 1].getBoolean());
                    break;
                case QTypeTuple.TYPECODE_FLOAT:
                    result = new Float(uplet[i - 1].getFloat());
                    break;
                case QTypeTuple.TYPECODE_STRING:
                    result = new String(uplet[i - 1].getString());
                    break;
                case QTypeTuple.TYPECODE_CHAR:
                    result = new Character(uplet[i - 1].getChar());
                    break;
                case QTypeTuple.TYPECODE_DATE:
                    result = uplet[i - 1].getDate();
                    break;

                default :
                    if (uplet[i - 1].getType().getJormName().equalsIgnoreCase("collection")) {
                        result = uplet[i - 1].getTupleCollection();
                        break;
                    } else
						result = uplet[i - 1].getObject();
            }
        }
        return result;
    }

    public boolean getBoolean(int i) throws MedorException {
        if ((i <= 0) || (i > size)) {
            throw new MedorException("Invalid Field index: " + i);
        } else
            return uplet[i - 1].getBoolean();
    }

    public void setBoolean(boolean x, int i) throws MedorException {
        if ((i <= 0) || (i > size)) {
            throw new MedorException("Invalid Field index: " + i);
        } else
            uplet[i - 1].setValue(x);
    }

    public char getChar(int i) throws MedorException {
        if ((i <= 0) || (i > size)) {
            throw new MedorException("Invalid Field index: " + i);
        } else
            return uplet[i - 1].getChar();
    }

	public char[] getCharArray(int i) throws MedorException {
		if ((i <= 0) || (i > size)) {
			throw new MedorException("Invalid Field index: " + i);
		} else
			return uplet[i - 1].getCharArray();
	}

     public void setChar(char x, int i) throws MedorException {
        if ((i <= 0) || (i > size)) {
            throw new MedorException("Invalid Field index: " + i);
        } else
            uplet[i - 1].setValue(x);
    }

    public Date getDate(int i) throws MedorException {
        if ((i <= 0) || (i > size)) {
            throw new MedorException("Invalid Field index: " + i);
        } else
            return uplet[i - 1].getDate();
    }

    public void setDate(Date x, int i) throws MedorException {
        if ((i <= 0) || (i > size)) {
            throw new MedorException("Invalid Field index: " + i);
        } else
            uplet[i - 1].setValue(x);
    }

    public TupleCollection getTupleCollection(int i) throws MedorException {
        if ((i <= 0) || (i > size)) {
            throw new MedorException("Invalid Field index: " + i);
        } else
            return uplet[i - 1].getTupleCollection();
    }

    public Operand getLikeOperand(int i) throws MedorException {
        if ((i <= 0) || (i > size)) {
            throw new MedorException("Invalid Field index: " + i);
        } else
            return uplet[i - 1];
    }

    public Object clone() {
        MemoryTuple o = null;
        try {
            o = (MemoryTuple) super.clone();
        } catch (CloneNotSupportedException e) {
            System.out.println("we can't have a copy of this Tuple ");
        }
        o.uplet = (VariableOperand[]) o.uplet.clone();
        boolean stop = false;
        for (int cpt = 0; ((cpt < size) && (!stop)); cpt++) {
            try {
                o.uplet[cpt] = (VariableOperand) o.uplet[cpt].clone();
            } catch (CloneNotSupportedException e) {
                stop = true;
                System.out.println("we can't have a copy of this Tuple ");
            }
        }
        if (!stop)
            return o;
        else
            return null;
    }
}
